Package: rdflib
Title: Tools to Manipulate and Query Semantic Data
Version: 0.2.9
Authors@R: c(person("Carl", "Boettiger", 
                  email = "cboettig@gmail.com", 
                  role = c("aut", "cre", "cph"),
                  comment=c(ORCID = "0000-0002-1642-628X")),
             person("Bryce", "Mecum", 
                  role = "rev",
                  comment=c(ORCID = "0000-0002-0381-3766")),
             person("Anna", "Krystalli", 
                  role = "rev",
                  comment=c(ORCID = "0000-0002-2378-4915")),
             person("Viktor", "Senderov",
                  email = "vsenderov@gmail.com",
                  role = "ctb",
                  comment = c(ORCID = "0000-0003-3340-5963")))
Description: The Resource Description Framework, or 'RDF' is a widely used
             data representation model that forms the cornerstone of the 
             Semantic Web. 'RDF' represents data as a graph rather than 
             the familiar data table or rectangle of relational databases.
             The 'rdflib' package provides a friendly and concise user interface
             for performing common tasks on 'RDF' data, such as reading, writing
             and converting between the various serializations of 'RDF' data,
             including 'rdfxml', 'turtle', 'nquads', 'ntriples', and 'json-ld';
             creating new 'RDF' graphs, and performing graph queries using 'SPARQL'.
             This package wraps the low level 'redland' R package which
             provides direct bindings to the 'redland' C library.  Additionally,
             the package supports the newer and more developer friendly
             'JSON-LD' format through the 'jsonld' package. The package
             interface takes inspiration from the Python 'rdflib' library.
License: MIT + file LICENSE
Encoding: UTF-8
URL: https://docs.ropensci.org/rdflib/,
        https://github.com/ropensci/rdflib
BugReports: https://github.com/ropensci/rdflib/issues
Imports: redland, methods, utils, stringi, readr, dplyr, tidyr
RoxygenNote: 7.2.3
Suggests: magrittr, covr, testthat, knitr, rmarkdown, jqr, DT, tibble,
        purrr, lubridate, httr, xml2, jsonlite, repurrrsive,
        nycflights13, spelling, jsonld
VignetteBuilder: knitr
Language: en-US
NeedsCompilation: no
Packaged: 2024-08-05 18:53:50 UTC; root
Author: Carl Boettiger [aut, cre, cph]
    (<https://orcid.org/0000-0002-1642-628X>),
  Bryce Mecum [rev] (<https://orcid.org/0000-0002-0381-3766>),
  Anna Krystalli [rev] (<https://orcid.org/0000-0002-2378-4915>),
  Viktor Senderov [ctb] (<https://orcid.org/0000-0003-3340-5963>)
Maintainer: Carl Boettiger <cboettig@gmail.com>
Repository: CRAN
Date/Publication: 2024-08-17 06:00:05 UTC
Built: R 4.6.0; ; 2026-04-25 12:41:18 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(as_rdf,data.frame)
S3method(as_rdf,list)
S3method(c,rdf)
S3method(format,rdf)
S3method(length,rdf)
S3method(print,rdf)
S3method(write_nquads,data.frame)
S3method(write_nquads,rdf)
export(as_rdf)
export(rdf)
export(rdf_add)
export(rdf_free)
export(rdf_has_bdb)
export(rdf_parse)
export(rdf_query)
export(rdf_serialize)
export(read_nquads)
export(write_nquads)
importClassesFrom(redland,Model)
importClassesFrom(redland,Parser)
importClassesFrom(redland,Query)
importClassesFrom(redland,Serializer)
importClassesFrom(redland,Statement)
importClassesFrom(redland,Storage)
importClassesFrom(redland,World)
importFrom(dplyr,left_join)
importFrom(methods,new)
importFrom(readr,read_csv)
importFrom(redland,is.null.externalptr)
importFrom(redland,librdf_model_size)
importFrom(redland,librdf_query_results_to_string2)
importFrom(stringi,stri_unescape_unicode)
importFrom(tidyr,gather)
importFrom(utils,capture.output)
importFrom(utils,download.file)
importFrom(utils,write.table)
importMethodsFrom(redland,addStatement)
importMethodsFrom(redland,executeQuery)
importMethodsFrom(redland,freeModel)
importMethodsFrom(redland,freeParser)
importMethodsFrom(redland,freeQuery)
importMethodsFrom(redland,freeQueryResults)
importMethodsFrom(redland,freeSerializer)
importMethodsFrom(redland,freeStatement)
importMethodsFrom(redland,freeStorage)
importMethodsFrom(redland,freeWorld)
importMethodsFrom(redland,parseFileIntoModel)
importMethodsFrom(redland,serializeToFile)
importMethodsFrom(redland,setNameSpace)

RDFLib is a Python library for working with RDF, a simple yet 
powerful language for representing information.

The library contains parsers and serializers for RDF/XML, N3,
NTriples, Turtle, TriX and RDFa.

The library presents a Graph interface which can be backed by
any one of a number of Store implementations. 

This core RDFLib package includes store implementations for 
in memory storage and persistent storage on top of the Berkeley DB. 

RDFLib is open source and is maintained in a 
`GitHub <http://github.com/RDFLib/rdflib/>`_ repository. RDFLib releases, current and previous 
are listed on `PyPi <pypi.python.org/pypi/rdflib/>`_

RDFLib may be easy_installed::

    $ easy_install rdflib

More information is available on the project webpage:

https://github.com/RDFLib/rdflib/

The RDFExtras project offers several additional stores as well as a 
SPARQL engine for use with RDFLib: 

https://github.com/RDFLib/rdfextras/

The documentation can be built by doing::

    $ python setup.py build_sphinx

And is also available from ReadTheDocs: 

http://readthedocs.org/docs/rdflib/

Support is available through the rdflib-dev group: 

http://groups.google.com/group/rdflib-dev

and on the IRC channel #redfoot on the freenode.net server

Some tools built using rdflib. Eventually these will become part of a
release separate from rdflib, but until such a separate release exists
we'll start gathering some here.

These rdflib_tool are being released under than same license as
rdflib.

# rdflib <img src="man/figures/logo.svg" align="right" alt="" width="120" />

<!-- badges: start -->
[![R-CMD-check](https://github.com/ropensci/rdflib/workflows/R-CMD-check/badge.svg)](https://github.com/ropensci/rdflib/actions)
[![Project Status: Active – The project has reached a stable, usable
state and is being actively
developed.](http://www.repostatus.org/badges/latest/active.svg)](https://www.repostatus.org/)
[![CRAN_Status_Badge](http://www.r-pkg.org/badges/version/rdflib)](https://cran.r-project.org/package=rdflib)
[![](http://badges.ropensci.org/169_status.svg)](https://github.com/ropensci/software-review/issues/169)
[![CRAN RStudio mirror
downloads](http://cranlogs.r-pkg.org/badges/rdflib)](https://CRAN.R-project.org/package=rdflib)
[![DOI](https://zenodo.org/badge/100521776.svg)](https://zenodo.org/badge/latestdoi/100521776)
<!-- badges: end -->

<!-- README.md is generated from README.Rmd. Please edit that file -->

A friendly and consise user interface for performing common tasks on rdf
data, such as parsing and converting between formats including rdfxml,
turtle, nquads, ntriples, and trig, creating rdf graphs, and performing
SPARQL queries. This package wraps the redland R package which provides
direct bindings to the redland C library. Additionally, the package
supports parsing and serialization of rdf into json-ld through the
json-ld package, which binds the official json-ld javascript API. The
package interface takes inspiration from the Python rdflib library.

## Installation

You can install rdflib from GitHub with:

``` r
# install.packages("devtools")
devtools::install_github("ropensci/rdflib")
```

## Basic use

While not required, `rdflib` is designed to play nicely with `%>%`
pipes, so we will load the `magrittr` package as well:

``` r
library(magrittr)
library(rdflib)
```

Parse a file and serialize into a different format:

``` r
system.file("extdata/dc.rdf", package="redland") %>%
  rdf_parse() %>%
  rdf_serialize("test.nquads", "nquads")
```

Perform SPARQL queries:

``` r
sparql <-
 'PREFIX dc: <http://purl.org/dc/elements/1.1/>
  SELECT ?a ?c
  WHERE { ?a dc:creator ?c . }'

system.file("extdata/dc.rdf", package="redland") %>%
rdf_parse() %>%
rdf_query(sparql)
#> # A tibble: 1 × 2
#>   a                      c           
#>   <chr>                  <chr>       
#> 1 http://www.dajobe.org/ Dave Beckett
```

Initialize graph a new object or add triples statements to an existing
graph:

``` r
x <- rdf()
x <- rdf_add(x, 
    subject="http://www.dajobe.org/",
    predicate="http://purl.org/dc/elements/1.1/language",
    object="en")
x
#> Total of 1 triples, stored in hashes
#> -------------------------------
#> <http://www.dajobe.org/> <http://purl.org/dc/elements/1.1/language> "en" .
```

Change the default display format (`nquads`) for graph objects:

``` r
options(rdf_print_format = "jsonld")
x
#> Total of 1 triples, stored in hashes
#> -------------------------------
#> {
#>   "@id": "http://www.dajobe.org/",
#>   "http://purl.org/dc/elements/1.1/language": "en"
#> }
```

## JSON-LD

We can also work with the JSON-LD format through additional functions
provided in the R package, `jsonld`.

``` r
out <- tempfile()
rdf_serialize(x, out, "jsonld")
rdf_parse(out, format = "jsonld")
#> Total of 1 triples, stored in hashes
#> -------------------------------
#> {
#>   "@id": "http://www.dajobe.org/",
#>   "http://purl.org/dc/elements/1.1/language": "en"
#> }
```

For more information on the JSON-LD RDF API, see
<https://json-ld.org/spec/latest/json-ld-rdf/>.

## Advanced Use

See [articles](https://docs.ropensci.org/rdflib/articles/) from the
documentation for advanced use including applications to large
triplestores, example SPARQL queries, and information about additional
database backends.

------------------------------------------------------------------------

## Citing rdflib

Please also cite the underlying `redland` library when citing `rdflib`

Carl Boettiger. (2018). rdflib: A high level wrapper around the redland
package for common rdf applications (Version 0.1.0). Zenodo.
<https://doi.org/10.5281/zenodo.1098478>

Jones M, Slaughter P, Ooms J, Boettiger C, Chamberlain S (2022).
*redland: RDF Library Bindings in R*. <doi:10.5063/F1VM496B>
<https://doi.org/10.5063/F1VM496B>, R package version 1.0.17-16,
<https://github.com/ropensci/redland-bindings/tree/master/R/redland>.

[![rofooter](https://ropensci.org//public_images/github_footer.png)](https://ropensci.org/)

![](docs/_static/RDFlib.png)    

RDFLib
======
[![Build Status](https://github.com/RDFLib/rdflib/actions/workflows/validate.yaml/badge.svg?branch=main)](https://github.com/RDFLib/rdflib/actions?query=branch%3Amain)
[![Documentation Status](https://readthedocs.org/projects/rdflib/badge/?version=latest)](https://rdflib.readthedocs.io/en/latest/?badge=latest)
[![Coveralls branch](https://img.shields.io/coveralls/RDFLib/rdflib/main.svg)](https://coveralls.io/r/RDFLib/rdflib?branch=main)

[![GitHub stars](https://img.shields.io/github/stars/RDFLib/rdflib.svg)](https://github.com/RDFLib/rdflib/stargazers)
[![Downloads](https://pepy.tech/badge/rdflib/week)](https://pepy.tech/project/rdflib)
[![PyPI](https://img.shields.io/pypi/v/rdflib.svg)](https://pypi.python.org/pypi/rdflib)
[![PyPI](https://img.shields.io/pypi/pyversions/rdflib.svg)](https://pypi.python.org/pypi/rdflib)
[![DOI](https://zenodo.org/badge/DOI/10.5281/zenodo.6845245.svg)](https://doi.org/10.5281/zenodo.6845245)

[![Contribute with Gitpod](https://img.shields.io/badge/Contribute%20with-Gitpod-908a85?logo=gitpod)](https://gitpod.io/#https://github.com/RDFLib/rdflib)
[![Gitter](https://badges.gitter.im/RDFLib/rdflib.svg)](https://gitter.im/RDFLib/rdflib?utm_source=badge&utm_medium=badge&utm_campaign=pr-badge)
[![Matrix](https://img.shields.io/matrix/rdflib:matrix.org?label=matrix.org%20chat)](https://matrix.to/#/#RDFLib_rdflib:gitter.im)

RDFLib is a pure Python package for working with [RDF](http://www.w3.org/RDF/). RDFLib contains most things you need to work with RDF, including:

* parsers and serializers for RDF/XML, N3, NTriples, N-Quads, Turtle, TriX, Trig, JSON-LD and even HexTuples
* a Graph interface which can be backed by any one of a number of Store implementations
* Store implementations for in-memory, persistent on disk (Berkeley DB) and remote SPARQL endpoints
    * additional Stores can be supplied via plugins 
* a SPARQL 1.1 implementation - supporting SPARQL 1.1 Queries and Update statements
* SPARQL function extension mechanisms

## RDFlib Family of packages
The RDFlib community maintains many RDF-related Python code repositories with different purposes. For example:

* [rdflib](https://github.com/RDFLib/rdflib) - the RDFLib core
* [sparqlwrapper](https://github.com/RDFLib/sparqlwrapper) - a simple Python wrapper around a SPARQL service to remotely execute your queries
* [pyLODE](https://github.com/RDFLib/pyLODE) - An OWL ontology documentation tool using Python and templating, based on LODE
* [pySHACL](https://github.com/RDFLib/pySHACL) - A pure Python module which allows for the validation of RDF graphs against SHACL graphs
* [OWL-RL](https://github.com/RDFLib/OWL-RL) - A simple implementation of the OWL2 RL Profile which expands the graph with all possible triples that OWL RL defines.

Please see the list for all packages/repositories here:

* <https://github.com/RDFLib>

Help with maintenance of all of the RDFLib family of packages is always welcome and appreciated.

## Versions & Releases

* `main` branch in this repository is the current unstable release
* `7.1.3` current stable release, small improvements to 7.1.1
* `7.1.2` previously deleted release
* `7.1.1` previous stable release
    * see <https://github.com/RDFLib/rdflib/releases/tag/7.1.1>
* `7.0.0` previous stable release, supports Python 3.8.1+ only.
    * see [Releases](https://github.com/RDFLib/rdflib/releases)
* `6.x.y` supports Python 3.7+ only. Many improvements over 5.0.0
    * see [Releases](https://github.com/RDFLib/rdflib/releases)
* `5.x.y` supports Python 2.7 and 3.4+ and is [mostly backwards compatible with 4.2.2](https://rdflib.readthedocs.io/en/stable/upgrade4to5.html).

See <https://github.com/RDFLib/rdflib/releases/> for the release details.

## Documentation
See <https://rdflib.readthedocs.io> for our documentation built from the code. Note that there are `latest`, `stable` and versioned builds, such as `5.0.0`, matching releases.

## Installation
The stable release of RDFLib may be installed with Python's package management tool *pip*:

    $ pip install rdflib

Some features of RDFLib require optional dependencies which may be installed using *pip* extras:

    $ pip install rdflib[berkeleydb,networkx,html,lxml,orjson]

Alternatively manually download the package from the Python Package
Index (PyPI) at https://pypi.python.org/pypi/rdflib

The current version of RDFLib is 7.1.3, see the ``CHANGELOG.md`` file for what's new in this release.

### Installation of the current main branch (for developers)

With *pip* you can also install rdflib from the git repository with one of the following options:

    $ pip install git+https://github.com/rdflib/rdflib@main

or

    $ pip install -e git+https://github.com/rdflib/rdflib@main#egg=rdflib

or from your locally cloned repository you can install it with one of the following options:

    $ poetry install  # installs into a poetry-managed venv

or

    $ pip install -e .

## Getting Started
RDFLib aims to be a pythonic RDF API. RDFLib's main data object is a `Graph` which is a Python collection
of RDF *Subject, Predicate, Object* Triples:

To create graph and load it with RDF data from DBPedia then print the results:

```python
from rdflib import Graph
g = Graph()
g.parse('http://dbpedia.org/resource/Semantic_Web')

for s, p, o in g:
    print(s, p, o)
```
The components of the triples are URIs (resources) or Literals
(values).

URIs are grouped together by *namespace*, common namespaces are included in RDFLib:

```python
from rdflib.namespace import DC, DCTERMS, DOAP, FOAF, SKOS, OWL, RDF, RDFS, VOID, XMLNS, XSD
```

You can use them like this:

```python
from rdflib import Graph, URIRef, Literal
from rdflib.namespace import RDFS, XSD

g = Graph()
semweb = URIRef('http://dbpedia.org/resource/Semantic_Web')
type = g.value(semweb, RDFS.label)
```
Where `RDFS` is the RDFS namespace, `XSD` the XML Schema Datatypes namespace and `g.value` returns an object of the triple-pattern given (or an arbitrary one if multiple exist).

Or like this, adding a triple to a graph `g`:

```python
g.add((
    URIRef("http://example.com/person/nick"),
    FOAF.givenName,
    Literal("Nick", datatype=XSD.string)
))
```
The triple (in n-triples notation) `<http://example.com/person/nick> <http://xmlns.com/foaf/0.1/givenName> "Nick"^^<http://www.w3.org/2001/XMLSchema#string> .`
is created where the property `FOAF.givenName` is the URI `<http://xmlns.com/foaf/0.1/givenName>` and `XSD.string` is the
URI `<http://www.w3.org/2001/XMLSchema#string>`.

You can bind namespaces to prefixes to shorten the URIs for RDF/XML, Turtle, N3, TriG, TriX & JSON-LD serializations:

 ```python
g.bind("foaf", FOAF)
g.bind("xsd", XSD)
```
This will allow the n-triples triple above to be serialised like this:
 ```python
print(g.serialize(format="turtle"))
```

With these results:
```turtle
PREFIX foaf: <http://xmlns.com/foaf/0.1/>
PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>

<http://example.com/person/nick> foaf:givenName "Nick"^^xsd:string .
```

New Namespaces can also be defined:

```python
dbpedia = Namespace('http://dbpedia.org/ontology/')

abstracts = list(x for x in g.objects(semweb, dbpedia['abstract']) if x.language=='en')
```

See also [./examples](./examples)


## Features
The library contains parsers and serializers for RDF/XML, N3,
NTriples, N-Quads, Turtle, TriX, JSON-LD, RDFa and Microdata.

The library presents a Graph interface which can be backed by
any one of a number of Store implementations.

This core RDFLib package includes store implementations for
in-memory storage and persistent storage on top of the Berkeley DB.

A SPARQL 1.1 implementation is included - supporting SPARQL 1.1 Queries and Update statements.

RDFLib is open source and is maintained on [GitHub](https://github.com/RDFLib/rdflib/). RDFLib releases, current and previous
are listed on [PyPI](https://pypi.python.org/pypi/rdflib/)

Multiple other projects are contained within the RDFlib "family", see <https://github.com/RDFLib/>.

## Running tests

### Running the tests on the host

Run the test suite with `pytest`.
```shell
poetry install
poetry run pytest
```

### Running test coverage on the host with coverage report

Run the test suite and generate a HTML coverage report with `pytest` and `pytest-cov`.
```shell
poetry run pytest --cov
```

### Viewing test coverage

Once tests have produced HTML output of the coverage report, view it by running:
```shell
poetry run pytest --cov --cov-report term --cov-report html
python -m http.server --directory=htmlcov
```

## Contributing

RDFLib survives and grows via user contributions!
Please read our [contributing guide](https://rdflib.readthedocs.io/en/latest/CONTRIBUTING.html) and [developers guide](https://rdflib.readthedocs.io/en/latest/developers.html) to get started.
Please consider lodging Pull Requests here:

* <https://github.com/RDFLib/rdflib/pulls>

To get a development environment consider using Gitpod or Google Cloud Shell.

[![Open in Gitpod](https://gitpod.io/button/open-in-gitpod.svg)](https://gitpod.io/#https://github.com/RDFLib/rdflib)
[![Open in Cloud Shell](https://gstatic.com/cloudssh/images/open-btn.svg)](https://shell.cloud.google.com/cloudshell/editor?cloudshell_git_repo=https%3A%2F%2Fgithub.com%2FRDFLib%2Frdflib&cloudshell_git_branch=main&cloudshell_open_in_editor=README.md)

You can also raise issues here:

* <https://github.com/RDFLib/rdflib/issues>

## Support & Contacts
For general "how do I..." queries, please use https://stackoverflow.com and tag your question with `rdflib`.
Existing questions:

* <https://stackoverflow.com/questions/tagged/rdflib>

If you want to contact the rdflib maintainers, please do so via:

* the rdflib-dev mailing list: <https://groups.google.com/group/rdflib-dev>
* the chat, which is available at [gitter](https://gitter.im/RDFLib/rdflib) or via matrix [#RDFLib_rdflib:gitter.im](https://matrix.to/#/#RDFLib_rdflib:gitter.im)

This parser is implemented using the BisonGen format (see references at bottom).  SPARQLParser.c is generated from all the 
.bgen and .bgen.frag files.  The command-line invokation for doing this is:

BisonGen --mode=c SPARQL.bgen

NOTE: The latest version of BisonGen (from CVS) may be required instead of the most package ftp://ftp.4suite.org/pub/BisonGen/

## Bison Gen Resources ##
- Copia article on BisonGen (with links): http://copia.ogbuji.net/blog/2005-04-27/Of_BisonGe
- BisonGen CVS Tree: cvs.4suite.org/viewcvs/BisonGen/

# Admin Tools

Tools to assist with RDFlib releases, like extracting all merged PRs from GitHub since last release and printing them into MArkdown lists.

To make a release of RDFLib, see the [Developer's Guide](https://rdflib.readthedocs.io/en/latest/developers.html).

## PR Changelog Summary

An alternative to the `get_merged_prs.py` script is to use the GitHub CL to get the list of PRs and pipe it into the `pr_markdown.py` script. The following command retrieves the list of PRs merged since the last release (`2025-09-19`) from a particular branch (`7.x`).

```bash
gh api '/search/issues?q=repo:rdflib/rdflib+is:pr+is:merged+base:7.x+merged:>2025-10-24&per_page=100' | jq '{total_count, incomplete_results, items: [.items[] | {number, title, pull_request_merged_at: .pull_request.merged_at, pull_request_url: .pull_request.html_url, username: .user.login}]}' | poetry run python admin/pr_markdown.py
```

This harness is meant to work against the DAWG test cases:

http://www.w3.org/2001/sw/DataAccess/tests/

The contents should be uncompressed to this directory and test.py should be run from the command line.
Currently it only checks SPARQL parsing.  I.e., it doesn't check against the results of evaluating the parsed
expression, just checks that it parses it.  Any exceptions that are thrown by the generated Bison parser
will be thrown.

# Test Data

This directory contains data for use inside tests, ideally the data in this
directory should be constant and should not change, and in general non-original
data that is widely known is preferred to original data as well known data has
well known attributes and qualities that can make it easier to reason about.


## File origins

- `rdfs.ttl`: `http://www.w3.org/2000/01/rdf-schema#`

## Fetcher

Files that originate from the internet should be downloaded using `fetcher.py`,
so we can easily verify the integrity of the files by re-running `fetcher.py`.

```bash
# run in repo root

# fetch everything
.venv/bin/python3 test/data/fetcher.py

# only fetch single file
.venv/bin/python3 test/data/fetcher.py test/data/rdfs.ttl

# only fetch files below path:
.venv/bin/python3 test/data/fetcher.py test/data/suites
```

For archives, if there is a `.patch` directory corresponding to the local path
name (e.g. `test/data/suites/w3c/rdf-xml.patch` for
`test/data/suites/w3c/rdf-xml`) then the content of the `.patch` directory will
be copied into the local path after fetching the remote content. Patch
directories should contain a README.md file that explains the reason for the
patch content.

The `test-suite` directory is a partial *copy* of the official test suite available at <http://www.w3.org/2013/json-ld-tests/>. Do *not* add tests directly there, but follow the instructions provided at the preceding location.

To update this copy, first obtain a local copy of the json-ld-tests test-suite. Then copy the relevants parts of that directory into this test directory.

You can do so by first defining a `$JSONLD_TESTSUITE` environment variable set to the local copy of the test-suite directory. Then run:

    $ cp $JSONLD_TESTSUITE/*.jsonld test/test-suite/
    $ cp $JSONLD_TESTSUITE/tests/{toRdf,fromRdf}-*.* test/test-suite/tests/

This README is for the W3C RDF Working Group's Turtle test suite.
This test suite contains four kinds of tests:

  132 Evaluation (rdft:TestTurtleEval) - a pair of an input turtle
  file and reference ntriples file.

  77 Positive syntax (rdft:TestTurtlePositiveSyntax) - an input turtle
  file with no syntax errors.

  78 Negative syntax (rdft:TestTurtleNegativeSyntax) - an input turtle
  file with at least one syntax error.

  4 Negative Evaluation (rdft:TestTurtleNegativeEval) - a pair of an
  input turtle file and reference ntriples file. These tests have the
  same properties as rdft:TestTurtleNegativeSyntax.

The manifest.ttl file in this directory lists all of the tests in the
RDF WG's Turtle test suite. Each test is one of the above tests. All
tests have a name (mf:name) and an input (mf:action). The Evaluation
tests have an expected result (mf:result).

• An implementation passes an Evaluation test if it parses the input
  into a graph, parses the expecte result into another graph, and
  those two graphs are isomorphic (see
  <http://www.w3.org/TR/rdf11-concepts/#graph-isomorphism>).

• An implementation passes a positive syntax test if it parses the
  input.

• An implementation passes a negative syntax test if it fails to parse
  the input.


RELATIVE IRI RESOLUTION:

The home of the test suite is <https://w3c.github.io/N3/tests/TurtleTests/>.
Per RFC 3986 section 5.1.3, the base IRI for parsing each file is the
retrieval IRI for that file. For example, the tests turtle-subm-01 and
turtle-subm-27 require relative IRI resolution against a base of
<https://w3c.github.io/N3/tests/TurtleTests/turtle-subm-01.ttl> and
<https://w3c.github.io/N3/tests/TurtleTests/turtle-subm-27.ttl> respectively.


CHARACTER ENCODING:

The Turtle language uses UTF-8 encoding. The following tests include
non-ascii characters:
  localName_with_assigned_nfc_bmp_PN_CHARS_BASE_character_boundaries
  localName_with_assigned_nfc_PN_CHARS_BASE_character_boundaries *
  localName_with_nfc_PN_CHARS_BASE_character_boundaries *
  labeled_blank_node_with_PN_CHARS_BASE_character_boundaries *
  LITERAL1_with_UTF8_boundaries *
  LITERAL_LONG1_with_UTF8_boundaries *
  LITERAL2_with_UTF8_boundaries *
  LITERAL_LONG2_with_UTF8_boundaries *

Those marked with a * include characters with codepoints greater than
U+FFFD and are thus expressed as a pair of surrogate characters when
represented in UCS2.


See http://www.w3.org/2011/rdf-wg/wiki/Turtle_Test_Suite for more details.


Eric Prud'hommeaux <eric+turtle@w3.org> - 11 June 2013.

Introduction
------------

The JSON-LD Test Suite is a set of tests that can
be used to verify JSON-LD Processor conformance to the set of specifications
that constitute JSON-LD. The goal of the suite is to provide an easy and
comprehensive JSON-LD testing solution for developers creating JSON-LD Processors.

Design
------

Tests are defined into _compact_, _expand_, _frame_, _normalize_, and _rdf_ sections:
* _compact_ tests have _input_, _expected_ and _context_ documents. The _expected_ results
  can be compared using JSON object comparison with the processor output.
* _expand_ tests have _input_ and _expected_ documents. The _expected_ results
  can be compared using JSON object comparison with the processor output.
* _frame_ tests have _input_, _frame_ and _expected_ documents. The _expected_ results
  can be compared using JSON object comparison with the processor output.
* _normalize_ tests have _input_ and _expected_ documents. The _expected_ results
  can be compared using string comparison with the processor output.
* _rdf_ tests have _input_ and _sparql_ documents. The results are tested
  by performing the RDF conversion and using this as the default document for an `ASK` query
  contained within the _sparql_ document using a SPARQL endpoint. The end result is a
  yes/no on whether the expected triples were extracted by the JSON-LD processor.


Contributing
------------

If you would like to contribute a new test or a fix to an existing test,
please follow these steps:

1. Notify the JSON-LD mailing list, public-linked-json@w3.org,
   that you will be creating a new test or fix and the purpose of the
   change.
2. Clone the git repository: git://github.com/json-ld/json-ld.org.git
3. Make your changes and submit them via github, or via a 'git format-patch'
   to the [JSON-LD mailing list](mailto:public-linked-json@w3.org).

Optionally, you can ask for direct access to the repository and may make
changes directly to the JSON-LD Test Suite source code. All updates to the test
suite go live on Digital Bazaar's JSON-LD Test Suite site within seconds of
committing changes to github via a WebHook call.

How to Add a Unit Test
----------------------

In order to add a unit test, you must follow these steps:

1. Pick a new unit test number. For example - 250. To be consistent, please use
   the next available unit test number.
2. Create a markup file in the tests/ directory with a .jsonld extension.
   For example: tests/rdf-250.jsonld
3. Create a SPARQL query file in the tests/ directory with a .jsonld or .sparql extension.
   For example: tests/rdf-250.sparql
4. Add your test to manifest.jsonld.

The test suite is designed to empower JSON-LD processor maintainers to create
and add tests as they see fit. This may mean that the test suite may become
unstable from time to time, but this approach has been taken so that the
long-term goal of having a comprehensive test suite for JSON-LD can be achieved
by the JSON-LD community.

# multi variant graphs

This directory containts variants of the same graph encoded in different
formats, or differently in the same format.

The graph that a specific file is a variant of is determined by it's filename.
Files that differ only in file extention but have the same basename are
considered variants of the same graph. Additionally, any suffix that matches
`-variant-[^/]*` is excluded when determening the graph key, so the following
files are all considered variants of the same graph:

```
test/variants/literal_with_lang-variant-control.ttl
test/variants/literal_with_lang.nt
test/variants/literal_with_lang.rdf
test/variants/literal_with_lang.ttl
```

Some additional assertions on graphs can be specified in file names that end
with `-asserts.json`, for details on supported asserts see
`test/test_graph/test_variants.py`.

This README is adapted from the README of the W3C RDF Working Group's
TriG test suite and modified for use with RDFLib's informally-assembled
TrIX test suite. This test suite contains three kinds of tests:

  Evaluation (rdft:TestTrixEval) - a pair of an input trig file
  and reference ntriples file.

  Positive syntax (rdft:TestTrixPositiveSyntax) - an input trig
  file with no syntax errors.

  Negative syntax (rdft:TestTrixNegativeSyntax) - an input trig
  file with at least one syntax error.

The manifest.ttl file in this directory lists all of the tests in the
RDFLib TriX test suite. Each test is one of the above tests. All
tests have a name (mf:name) and an input (mf:action). The Evaluation
tests have an expected result (mf:result).

• An implementation passes an Evaluation test if it parses the input
  into a graph, parses the expecte result into another graph, and
  those two graphs are isomorphic (see
  <http://www.w3.org/TR/rdf11-concepts/#graph-isomorphism>).

• An implementation passes a positive syntax test if it parses the
  input.

• An implementation passes a negative syntax test if it fails to parse
  the input.

The IRI for the test suite is <https://rdflib.github.io/tests/trix/>.
Per RFC 3986 section 5.1.3, the base IRI for parsing each file is the
retrieval IRI for that file. For example, the tests trig-subm-01 and
trig-subm-27 require relative IRI resolution against a base of
<https://rdflib.github.io/tests/trix/trig-subm-01.trix> and
<https://rdflib.github.io/tests/trix/strig-subm-27.trix> respectively.


Adapted from https://github.com/w3c/rdf-tests/blob/main/trig/README,
authors:

Eric Prud'hommeaux <eric+turtle@w3.org> - 11 June 2013.
Gregg Kellogg <gregg@greggkellogg.net> - 12 June 2013.

# Notation3 Tests

Tests exist for [N3 grammar](N3Tests/manifest.ttl), [N3 reasoning](N3Tests/manifest-reasoner.ttl), [extended N3 grammar](N3Tests/manifest-extended.ttl) and appropriate [Turtle](TurtleTests/manifest.ttl) tests.

# Design

In general, tests are described as Positive or Negative Syntax tests, Evaluation Test, and Reason tests. Syntax tests simply check that the data can be parsed properly; negative syntax tests should generate an error.

Evaluation tests check whether the input (`action`) can be parsed into a dataset which is isomorphic to those described in `results`.

Reason tests invoke the Notation3 reasoner, with various options, and check that the results are isomorphic to those described in `results`.

Tests should be run with an assumed base URI of their associated manifest, either `https://w3c.github.io/N3/tests/N3Tests` or `https://w3c.github.io/N3/tests/TurtleTests`.

# Contributing

If you would like to contribute a new test or a fix to an existing test,
please file an [issue](https://github.com/w3c/N3/issues) and/or create a [pull request](https://github.com/w3c/N3/pulls).

This README is for the W3C RDF Working Group's RDF/XML test suite.
This test suite contains two kinds of tests:

  Evaluation (rdft:TestXMLEval) - a pair of an input RDF/XML file
  and reference ntriples file.

  Negative syntax (rdft:TestXMLNegativeSyntax) - an input RDF/XML
  file with at least one syntax error.

The manifest.ttl file in this directory lists all of the tests in the
RDF WG's RDF/XML test suite. Each test is one of the above tests. All
tests have a name (mf:name) and an input (mf:action). The Evaluation
tests have an expected result (mf:result).

• An implementation passes an Evaluation test if it parses the input
  into a graph, parses the expected result into another graph, and
  those two graphs are isomorphic (see
  <http://www.w3.org/TR/rdf11-concepts/#graph-isomorphism>).

• An implementation passes a positive syntax test if it parses the
  input.

• An implementation passes a negative syntax test if it fails to parse
  the input.

The home of the test suite is <http://www.w3.org/2013/RDFXMLTests/>.
Per RFC 3986 section 5.1.3, the base IRI for parsing each file is the
retrieval IRI for that file. For example, the tests rdf-containers-syntax-vs-schema-test004 and
xmlbase-test014 require relative IRI resolution against a base of
<http://www.w3.org/2013/RDFXMLTests/rdf-containers-syntax-vs-schema/test004.rdf> and
<http://www.w3.org/2013/RDFXMLTests/xmlbase/test014.rdf> respectively.

See http://www.w3.org/2011/rdf-wg/wiki/RDFXML_Test_Suite for more details.

Gregg Kellogg <gregg@greggkellogg.net> - 23 December 2013.

This README is for the W3C RDF Working Group's N-Triples test suite.
This test suite contains two kinds of tests:

  Positive syntax (rdft:TestNTriplesPositiveSyntax) - an input N-Triples
  file with no syntax errors.

  Negative syntax (rdft:TestNTriplesNegativeSyntax) - an input N-Triples
  file with at least one syntax error.

The manifest.ttl file in this directory lists tests in the
RDF WG's N-Triples test suite. All
tests have a name (mf:name) and an input (mf:action).

• An implementation passes a positive syntax test if it parses the
  input.

• An implementation passes a negative syntax test if it fails to parse
  the input.

The home of the test suite is <http://www.w3.org/2013/N-TriplesTests/>.

See http://www.w3.org/2011/rdf-wg/wiki/RDF_Test_Suites for more details.

Eric Prud'hommeaux <eric+turtle@w3.org> - 11 June 2013.
Gregg Kellogg <gregg@greggkellogg.net> - 26 June 2013.

# Non-normative tests W3C tests for RDF/XML

This test suite contains non-normative tests for RDF/XML. These tests originate
from tests commented out in the [W3C RDF/XML test
suite](http://www.w3.org/2013/RDFXMLTests/).

These tests are not added in an RDFLib specific test suite as they are authored
under W3C efforts.

The original home of the tests in this suite is
<http://www.w3.org/2013/RDFXMLTests/>. Per RFC 3986 section 5.1.3, the base IRI
for parsing each file is the retrieval IRI for that file. For example, the
tests rdfms-xml-literal-namespaces-test001 require relative IRI resolution
against a base of
<http://www.w3.org/2013/RDFXMLTests/rdfms-xml-literal-namespaces/test001.rdf>.

For more information see:
- <http://www.w3.org/2013/RDFXMLTests/README>.
- <https://github.com/w3c/rdf-tests/issues/80>.

#' Tidy Schema
#' 
#' Construct a SPARQL query in which properties (predicates) become column headings
#' and values (objects) become entries of individual cells.  
#' 
#' @param ... names of the properties that should make up columns in the table
#' @param columns Alternatively, supply a vector of property names. 
#' @param prefix the URI string to prefix before the property names to give
#'  fully-resolved properties.  
#' @param na.rm logical, default TRUE. Will not return a row for any 
#'  subject that does not have a object value matching the schema.
#' @examples
#' 
#' sparql <- table_schema("Species",  "Sepal.Length", prefix = "iris")
#' rdf_query(rdf, sparql)
#' 
#' ## use columns arg for an existing vector of names.
#' columns <- names(iris)
#' sparql <- tidy_schema(columns = columns, prefix = "iris")
#' rdf_query(rdf, sparql)
#' 
#' ## use na.rm = FALSE to include NA if variable is not defined for some observations
#' sparql <- tidy_schema("Species",  "Sepal.Length", "Sepal.Color",
#'                        prefix = "iris", na.rm=FALSE)
#' rdf_query(rdf, sparql)
tidy_schema <- function(..., columns = NULL, prefix=NULL, na.rm = TRUE){
  if(is.null(columns)){
    columns <- c(...)
  }
  
  if(!is.null(prefix)){
    attributes <- paste0("<",prefix, ":", columns, ">")
  } else {
    attributes <- paste0("<", columns, ">")
  }
  ## Replace forbidden characters with "_", 
  ## See spec: https://www.w3.org/TR/rdf-sparql-query/#rVARNAME
  vars <- gsub("[^a-zA-Z1-9_]","_", basename(columns))
  select <- paste0("?", vars)
  
  where <- paste("?s", attributes, select)
  if(!na.rm)
    where <- paste("OPTIONAL {", where, "}")
  else
    where <- paste(where, ".")
  where <- paste(where, collapse = "\n")
  query <- paste("SELECT", paste0(select, collapse = " "),
                 "\nWHERE {\n", where, "\n}")
  query
}


#sparql <- table_schema("Sepal.Length", "Sepal.Width", 
#                       "Petal.Length", "Petal.Width",  
#                       "Species", prefix="iris")
#sparql <- table_schema(columns=names(iris), prefix="iris")
#rdf_query(rdf, sparql) %>% as_tibble()

## Experimental testing of different storage backends in redland:

library(redland)
world <- new("World")

## BDB
bdb_storage <- new("Storage", world, "hashes", name = "db1", 
                      options = "new='yes',hash-type='bdb',dir='.'")
model <- new("Model", world = world, storage = bdb_storage, options = "")

## SQLITE 
sqlite_storage <- new("Storage", world, "sqlite", name = "sqlite1", options = "new='yes'")


## POSTGRES
## Needs postgres backend running
postgres_storage <- new("Storage", world, "postgresql", name = "postgres1", 
    options = "new='yes',host='postgres',user='postgres','password='rdflib'")

## MYSQL
## Needs mysql backend running
mysql_store <- new("Storage", world, "mysql", name = "mysql1", 
                        options = "new='yes',host='mariadb',user='root',password='rdflib',database='db'")

## VIRTUOSO
## Needs virtuoso backend running
virtuoso_storage <- new("Storage", world, "virtuoso", name = "db1",
      "dsn='Local Virtuoso',user='demo',password='demo'")


## Works, in memory, serializes to an rdf/xml file called thing.rdf when freed.
## Not indexed, so will be slow. Suitable for small models.
file_storage <- new("Storage", world, "file", "thing.rdf", "")
storage <- file_storage
model <- new("Model", world = world, storage = storage, options = "")

## Works, fast write, not indexed, good for only small models, 
## no reason to use this instead of hash-based memory (which is indexed)
memory_storage <- new("Storage", world, "memory", "", "")
storage <- memory_storage
model <- new("Model", world = world, storage = storage, options = "")


## Testing
library(rdflib)
rdf <- structure(list(world = world, model = model, storage = storage),
          class = "rdf")
rdf_add(rdf, 
        subject="http://www.dajobe.org/",
        predicate="http://purl.org/dc/elements/1.1/language",
        object="en")
rdf

library(DBI)
library(odbc)

virtuoso <- dbConnect(odbc::odbc(),
                      driver = "Virtuoso",
                      uid = "dba",
                      pwd = "dba",
                      host = "virtuoso",
                      port = "1111")

## It's alive!
virtuoso

## Bulk import -- fast!
dbGetQuery(virtuoso, "ld_dir('/var/data/', '*.nq', 'rdflib')" )
dbGetQuery(virtuoso, "rdf_loader_run()" )


## List all graphs -- look, we have one called 'rdflib' now!
dbGetQuery(virtuoso,
           "SPARQL SELECT  DISTINCT ?g WHERE { GRAPH ?g {?s ?p ?o} } ORDER BY ?g"
)


## Select FROM GRAPH
dbGetQuery(virtuoso, "SPARQL SELECT * FROM <rdflib> WHERE { {?s ?p ?o} } LIMIT 10")

dbGetQuery(virtuoso, "SPARQL SELECT * FROM <rdflib> WHERE { {?s <flights:dep_delay> ?o} } LIMIT 10")

## Wow, we can write standard tables into Virtuoso too...
#dbWriteTable(virtuoso, "iris", iris) # Dots not allowed in column names
dbWriteTable(virtuoso, "mtcars", mtcars)
dbListTables(virtuoso, table_name = "mt%")
dbReadTable(virtuoso, "mtcars")


## Clear graph <rdflib>
dbGetQuery(virtuoso, "SPARQL CLEAR GRAPH <rdflib>")


## Counting triples in each different graph...
q <- "SPARQL SELECT ?g ?s ?p ?o  WHERE { GRAPH ?g {?s ?p ?o} }"
df <- dbGetQuery(virtuoso, q)
library(dplyr)
df %>% count(g)



#remotes::install_github("ropensci/rdflib")
library(rdflib)
library(nycflights13)
library(dplyr)
## prefix foreign keys
uri_flights <- flights %>% 
  mutate(tailnum = paste0("planes:", tailnum),
         carrier = paste0("airlines:", carrier))

h <- here::here("data/rdflib/inst/extdata")
write_nquads(airlines, file.path(h, "airlines.nq"),  key = "carrier", prefix = "airlines:")
write_nquads(planes, file.path(h, "planes.nq"),   key = "tailnum", prefix = "planes:")
write_nquads(uri_flights,  file.path(h, "uri_flights.nq"), prefix = "flights:")

## Bulk import -- fast!
dbGetQuery(virtuoso, "ld_dir('/var/data/', '*.nq', 'rdflib')" )
dbGetQuery(virtuoso, "rdf_loader_run()" )





remotes::install_github("ropensci/rdflib")
library(rdflib)
triplestore <- rdf(storage = "virtuoso", 
                   user = "dba", 
                   password = "dba",
                   host = "virtuoso:1111"
)
triplestore
#write_nquads(iris, "iris.nq", prefix = "iris")
#read_nquads("iris.nq", rdf = triplestore)
## Works!
triplestore

## We can query with rdflib too!
query <- "SELECT ?s ?p ?o WHERE {?s ?p ?o } LIMIT 10"
out <- rdf_query(triplestore, query)
rdf_free(triplestore)

library(nycflights13)
library(dplyr)

dim(flights) # 336,776 x 19 

as_uri <- function(x, base_uri = "x:") paste0(base_uri, x)
uri_flights <- flights %>% 
  mutate(tailnum = as_uri(tailnum),
         carrier = as_uri(carrier))

  
df1 <- rdflib:::normalize_table(airlines, key = "carrier", prefix = "x:")
df2 <- rdflib:::normalize_table(planes,   key = "tailnum", prefix = "x:")
df3 <- rdflib:::normalize_table(uri_flights, key = NULL, prefix = "x:")

df <- bind_rows(df1,df2,df3)

dim(df3) # 6,398,744 x 4
  
df %>% 
  filter(predicate %in% c("carrier", "name", "manufacturer", "model", "dep_delay")) %>% 
  count(predicate)

## We can recover an individual table
df %>% 
  filter(predicate %in% c("manufacturer", "model")) %>% 
  select(subject, predicate, object) %>% 
  tidyr::spread(predicate, object) %>% 
  rename(tailnum = subject) ## Also need to apply datatype...
## ... and join manually....


## What about recovering the "joined" table?
df %>% 
  filter(predicate %in% c("carrier", "name", "manufacturer", "model", "dep_delay")) %>%
  select(subject, predicate, object) %>% tidyr::spread(predicate, object) %>% 
  filter(!is.na(carrier))


## Compare to pure approach on original tables
flights %>% 
  left_join(airlines, by = "carrier") %>%
  left_join(planes, by = "tailnum") %>% 
  select(carrier, name, manufacturer, model, dep_delay) %>% 
  distinct()

#### All fits in memory anyway
#library(MonetDBLite)
#library(DBI)
#triplestore <- rappdirs::user_data_dir("rdflib")
#con <- dbConnect(MonetDBLite::MonetDBLite(), triplestore)

# Could use append=TRUE instead to extend triplestore later
#DBI::dbWriteTable(con, "flights", df, overwrite = TRUE)  


## size as flat files on disk:
readr::write_tsv(flights, "flights.tsv")    # 29.6 MB
readr::write_tsv(df, "triplestore.tsv")     # 319 MB
readr::write_tsv(df, "triplestore.tsv.bz2") # 17 MB

## Deprecate this strategy in favor of direct result
## also means we can deprecate type_by_datauri and other utilities
iter_getNextResult <- function(queryResult){
  out <- list()
  result <- redland::getNextResult(queryResult)
  out <- c(out, result)
  while(!is.null(result)){
    result <- redland::getNextResult(queryResult)
    out <- c(out, result)
  }
  out
}

rectangularize_query_results <- function(out){
  vars <- unique(names(out))
  
  X <- lapply(vars, function(v){ 
    contents <- as.character(out[names(out) == v ])
    values <- type_by_datauri(contents)
    
    ## use "character" if mixed type column
    types <- vapply(values, function(x) class(x)[[1]], character(1))
    u <- unique(types)
    if(length(u) == 1){
      values <- unlist(values)
      if(u %in% c("Date", "POSIXct"))
        class(values) <- unique(types) # Restore date class
    } else {
      values <- vapply(values, as.character, character(1))
    }
    values
  })
  
  names(X) <- vars
  ## Or we could use tibble::as_data_frame for list columns w/ mixed type..
  as.data.frame(X, stringsAsFactors=FALSE)
}





#' @importFrom methods as
type_by_datauri <- function(x){
  types <- get_types(x)
  r_types <- vapply(types, r_class, character(length(1)))
  values <- get_values(x)
  
  ## Output must be a list since types can differ
  out <- vector("list", length = length(values))
  for(i in seq_along(values)){
    out[[i]] <- as(utf8me(values[i]), r_types[[i]])
    
  }
  out
}

## Utilities to coerce return type, if recognized
r_class <- function(x){
  switch(gsub("<http://www.w3.org/2001/XMLSchema#(.*)>", "xs:\\1", x),
         "xs:decimal" = "numeric",
         "xs:string" = "character",
         "xs:boolean" = "logical",
         "xs:integer" = "integer",
         "xs:date" = "Date",
         "xs:dateTime" = "POSIXct",
         "character"
  )
}

## Helper utilitiesfor parsing data URIs
get_values <- function(x) gsub('\"(([^\\^])+)\"\\^*(.*)',  "\\1", x)
get_types <- function(x) gsub('\"(([^\\^])+)\"\\^*(.*)',  "\\3", x)
#' @importFrom stringi stri_unescape_unicode
# https://stackoverflow.com/questions/48602294
utf8me <- function(x){ 
  removed_quotes <- gsub('\"', '', x)
  stringi::stri_unescape_unicode(removed_quotes)
}


## so that we can do as("2018-02-05", "Date") in type_by_datauri
setAs("character", "Date", function(from) as.Date(as.character(from)))
setAs("character", "POSIXct", function(from) as.POSIXct(as.character(from)))

#' Concatenate rdf Objects.  
#' 
#' All subsequent rdf objects will be appended to the first rdf object
#' Note: this does not free memory from any of the individual rdf objects
#' Note: It is generally better to avoid the use of this function by passing
#' an existing rdf object to and rdf_parse or rdf_add objects. Multiple active
#' rdf objects can cause problems when using disk-based storage backends.   
#'   
#' @method c rdf
#' @export
#' @param ... objects to be concatenated
c.rdf <- function(...){
  rdfs <- list(...)
  loc <- tempdir()
  rdf <- rdfs[[1]]
  for(i in seq_along(rdfs)){
    f <- file.path(loc,paste0(i, ".rdf"))
    rdf_serialize(rdfs[[i]],f, format = "turtle") 
    rdf_parse(f, rdf = rdf, format = "turtle")
    file.remove(f)
  }
  unlink(loc)
  rdf
}



#' @importFrom redland librdf_model_size
#' @export
length.rdf <- function(x){
  redland::librdf_model_size(x$model@librdf_model)
}

#' @export
print.rdf <- function(x, ...){
  cat(format.rdf(x, ...), sep = "\n")
}


#' @importFrom stringi stri_unescape_unicode
#' @export
# @param max_print maximum number of lines to print of rdf preview
# @param max_preview if number of triples exceeds this, no preview
# will be displayed, since preview method must currently serialize
# entire triplestore.
format.rdf <- function(x,
                       format = getOption("rdf_print_format", "nquads"),
                       max_print = getOption("rdf_max_print", 10L),
                       max_preview = 1e5,
                       ...){
  n <- redland::librdf_model_size(x$model@librdf_model)
  header <- paste0("Total of ", n, " triples, stored in ", x$storage@type, "\n",
                  "-------------------------------\n")
  if(n < max_preview){
    tmp <- tempfile()
    rdf_serialize(x, 
                  tmp,
                  format = format,
                  ...)
    ## Fix encoding on nquads, ntriples 
    footer <- NULL
    if(n > max_print){
      footer <- paste0("\n\n... with ", n-max_print, " more triples")
    }
    txt <- paste0(header, stringi::stri_unescape_unicode(
      paste(readLines(tmp, n = max_print), collapse = "\n")),
      footer)
    unlink(tmp)
  } else {
    txt <- paste(header, "\n (preview supressed for performance)")
  }
  txt
}

#' Serialize an RDF Document
#'
#' @inheritParams rdf_parse
#' @inheritParams rdf_query
#' @param doc file path to write out to. If null, will write to character.
#' @param namespace a named character containing the prefix to namespace bindings. \code{names(namespace)} are the prefixes, whereas \code{namespace} are the namespaces
#' @param prefix (optional) for backward compatibility. See \code{namespace}. It contains the matching prefixes to the namespaces in \code{namespace} and is set automatically if you provide \code{namespace} as a named character vector.
#' @param ... additional arguments to \code{redland::serializeToFile}
#' @return rdf_serialize returns the output file path `doc` invisibly.
#'   This makes it easier to use rdf_serialize in pipe chains with
#'   \code{\link{rdf_parse}}.
#' @importFrom methods new
#' @importClassesFrom redland Serializer
#' @importMethodsFrom redland setNameSpace serializeToFile freeSerializer
#' @importFrom jsonld jsonld_compact
#'
#' @export
#' @examples
#' infile <- system.file("extdata", "dc.rdf", package="redland")
#' out <- tempfile("file", fileext = ".rdf")
#'
#' some_rdf <- rdf_parse(infile)
#' rdf_add(some_rdf,
#'     subject = "http://www.dajobe.org/dave-beckett",
#'     predicate = "http://www.w3.org/1999/02/22-rdf-syntax-ns#type",
#'     object = "http://xmlns.com/foaf/0.1/Person")
#' rdf_serialize(some_rdf, out)
#'
#' ## With a namespace
#' rdf_serialize(some_rdf,
#'           out,
#'           format = "turtle",
#'           namespace = c(dc = "http://purl.org/dc/elements/1.1/",
#'           foaf = "http://xmlns.com/foaf/0.1/")
#'           )
#'
#' readLines(out)
rdf_serialize <- function(rdf,
                          doc = NULL,
                          format = c("guess",
                                     "rdfxml",
                                     "nquads",
                                     "ntriples",
                                     "turtle",
                                     "jsonld"),
                          namespace = NULL,
                          prefix = names(namespace),
                          base = getOption("rdf_base_uri", "localhost://"),
                          ...){
  
  format <- match.arg(format)
  if(format == "guess" & !is.null(doc)){
    format <- guess_format(doc)
  }
  
  
  ## redlands doesn't support jsonld. So write as nquads and then transform
  jsonld_output <- format == "jsonld"
  if(jsonld_output){
    format <- "nquads"
  }
  
  mimetype <- rdf_mimetypes[format]
  
  serializer <-
    new("Serializer", rdf$world,
        name = format, mimeType = mimetype)
  
  if(!is.null(namespace) && is.character(namespace) && length(namespace) >= 1 && length(namespace) == length(prefix)){
    ix = 1:length(namespace)
    for (i in ix) {
      redland::setNameSpace(serializer,
                            rdf$world,
                            namespace = namespace[i],
                            prefix = prefix[i]) 
    }
  }
  
  if(is.null(doc)){
    doc <- redland::serializeToCharacter(serializer, rdf$world, 
                                         rdf$model, baseUri = base, ...)
  } else {
    status <-
      redland::serializeToFile(serializer, rdf$world, 
                               rdf$model, doc, baseUri = base, ...)
  }
  
  if(jsonld_output){
    txt <- paste(readLines(doc), collapse = "\n")
    if(length(txt) > 0){ ## don't attempt to write empty file into json
      json <- jsonld::jsonld_from_rdf(txt,
                                      options = list(
                                        base = base,
                                        format = "application/nquads"))
      compact_json <- jsonld_compact(json, "{}")
      writeLines(compact_json, doc)
    }
  }
  
  redland::freeSerializer(serializer)
  invisible(doc)
}

testthat::context("Parsers and Serializers")

doc <- system.file("extdata/example.rdf", package="redland")
out <- "testing.rdf"



testthat::test_that("we can serialize to character", {
                      rdf <- rdf_parse(doc) 
                      txt <- rdf_serialize(rdf, format = "nquads")
                      testthat::expect_is(txt, "character")
                      testthat::expect_match(txt, "John Smith")
                      rdf_free(rdf)
                      })


testthat::test_that("we can parse (in rdfxml)
                    and serialize (in nquads) a simple rdf graph", {
                      rdf <- rdf_parse(doc) 
                      rdf_serialize(rdf, out, "nquads")
                      roundtrip <- rdf_parse(out, "nquads")
                      testthat::expect_is(roundtrip, "rdf")
                      
                      rdf_free(roundtrip)
                      rdf_free(rdf)
                      })

testthat::test_that("we can add a namespace on serializing", {
  rdf <- rdf_parse(doc)
  rdf_serialize(rdf,
                out,
                namespace = "http://purl.org/dc/elements/1.1/",
                prefix = "dc")
  roundtrip <- rdf_parse(doc)
  testthat::expect_is(roundtrip, "rdf")
  
  rdf_free(rdf)
  rdf_free(roundtrip)
  
})

################################################################


testthat::test_that("we can parse and serialize json-ld", {
  rdf <- rdf_parse(doc)
  rdf_serialize(rdf, "out.json")
  roundtrip <- rdf_parse("out.json")
  testthat::expect_is(roundtrip, "rdf")
  rdf_serialize(rdf, "out.jsonld")
  unlink("out.json")
  unlink("out.jsonld")
  rdf_free(roundtrip)
  rdf_free(rdf)
  
})

testthat::test_that("we can parse and serialize nquads", {
  rdf <- rdf_parse(doc)
  rdf_serialize(rdf, "out.nquads")
  roundtrip <- rdf_parse("out.nquads")
  testthat::expect_is(roundtrip, "rdf")
  unlink("nquads")
  rdf_free(roundtrip)
  rdf_free(rdf)
  
})
testthat::test_that("we can parse and serialize ntriples", {
  rdf <- rdf_parse(doc)
  rdf_serialize(rdf, "out.nt")
  roundtrip <- rdf_parse("out.nt")
  testthat::expect_is(roundtrip, "rdf")
  unlink("out.nt")
  rdf_serialize(rdf, "out.ntriples")
  unlink("out.ntriples")

  rdf_free(roundtrip)
  rdf_free(rdf)
})
testthat::test_that("we can parse and serialize tutle", {
  rdf <- rdf_parse(doc)
  rdf_serialize(rdf, "out.ttl")
  roundtrip <- rdf_parse("out.ttl")
  testthat::expect_is(roundtrip, "rdf")
  unlink("out.ttl")
  rdf_serialize(rdf, "out.turtle")
  unlink("out.turtle")
  
  rdf_free(roundtrip)
  rdf_free(rdf)
})
testthat::test_that("we can parse and serialize rdfxml", {
  rdf <- rdf_parse(doc)
  rdf_serialize(rdf, "out.rdf")
  roundtrip <- rdf_parse("out.rdf")
  testthat::expect_is(roundtrip, "rdf")

  unlink("out.rdf")
  rdf_serialize(rdf, "out.xml")
  unlink("out.xml")
  
  rdf_free(roundtrip)
  rdf_free(rdf)
})

################################################################


testthat::test_that("we can parse by guessing on the file extension", {
  ex <- system.file("extdata/person.nq", package="rdflib")
  rdf <- rdf_parse(ex)
  rdf_serialize(rdf, "tmp.nq", base = "http://schema.org/")
  roundtrip <- rdf_parse("tmp.nq", "turtle")
  testthat::expect_is(roundtrip, "rdf")
  unlink("tmp.nq")
  rdf_free(rdf)
})


testthat::test_that("we can serialize turtle with a baseUri", {
  ex <- system.file("extdata/person.nq", package="rdflib")
  rdf <- rdf_parse(ex, "nquads")
  rdf_serialize(rdf, out, "turtle", base = "http://schema.org/")
  roundtrip <- rdf_parse(out, "turtle")
  testthat::expect_is(roundtrip, "rdf")
  rdf_free(rdf)
})


## JSON-LD tests with default base uri

  testthat::test_that("@id is not a URI, we should get localhost", {
    ex <- '{
      "@context": "http://schema.org/",
      "@id": "person_id",
      "name": "Jane Doe"
    }'
    
    rdf <- rdf_parse(ex, "jsonld")
    testthat::expect_output(cat(format(rdf, "nquads")), "localhost")
    rdf_free(rdf)
  })

  testthat::test_that("@id is a URI, we should not get localhost", {
    ex <- '{
      "@context": "http://schema.org/",
      "@id": "uri:person_id",
      "name": "Jane Doe"
    }'
    rdf <- rdf_parse(ex, "jsonld")
    testthat::expect_false(grepl("localhost", format(rdf, "nquads")))
    rdf_free(rdf)
  })  
  
  testthat::test_that("we can alter the base URI", {
    ex <- '{
      "@id": "person_id",
      "schema:name": "Jane Doe"
    }'
    options(rdf_base_uri = "http://example.com/")
    rdf <- rdf_parse(ex, "jsonld")
    testthat::expect_output(cat(format(rdf, "nquads")), "http://example.com")
    rdf_free(rdf)
    
    
    options(rdf_base_uri = "")
    rdf <- rdf_parse(ex, "jsonld")
    testthat::expect_length(rdf, 0)
    rdf_free(rdf)
    
    options(rdf_base_uri = NULL)
  })
  
  
  
  

  
  
testthat::test_that("we can parse into an existing rdf model", {
    rdf1 <- rdf_parse(system.file("extdata/ex.xml", package = "rdflib"))
    rdf2 <- rdf_parse(system.file("extdata/ex2.xml", package = "rdflib"),
                      rdf = rdf1)
    
    testthat::expect_is(rdf1, "rdf")
    testthat::expect_is(rdf2, "rdf")
    testthat::expect_identical(rdf1, rdf2)
    rdf_free(rdf1)
    
    ## NOTE: rdf is same pointer as rdf1, not a new pointer.  cannot free twice
})
  
  

testthat::test_that("we can parse from a url", {
  # CRAN seems okay with tests requiring an internet connection
  #testthat::skip_on_cran()
  rdf <- rdf_parse("https://tinyurl.com/ycf95c9h")
  testthat::expect_is(rdf, "rdf")
  
  rdf_free(rdf)
})

testthat::test_that("we can parse from a text string", {
  
  
  rdf <- rdf_parse(doc)
  txt <- rdf_serialize(rdf, format = "rdfxml")
  testthat::expect_is(txt, "character")
  roundtrip <- rdf_parse(txt, format="rdfxml")
  testthat::expect_is(roundtrip, "rdf")
  rdf_free(rdf)
  
  string <-
    '
  _:b0 <http://schema.org/jobTitle> "Professor" .
  _:b0 <http://schema.org/name> "Jane Doe" .
  _:b0 <http://schema.org/age> "35" .
  _:b0 <http://schema.org/url> <http://www.janedoe.com> .
  '
  rdf <- rdf_parse(string, "nquads")
  testthat::expect_is(rdf, "rdf")
  
  rdf_free(rdf)
})

unlink(out)

testthat::context("test-rdf.R")

doc <- system.file("extdata/example.rdf", package="redland")
out <- "testing.rdf"


testthat::test_that("We can initialize and free rdf objects", {
  rdf <- rdf()
  
  testthat::expect_is(rdf, "rdf")
  testthat::expect_is(rdf$world, "World")
  testthat::expect_is(rdf$model, "Model")
  testthat::expect_is(rdf$storage, "Storage")
  
  rdf_free(rdf)
  
  ## Confirm `rdf` has been removed from the workspace
  testthat::expect_false("rdf" %in% ls())
  
  RDF_graph <- rdf()
  testthat::expect_true("RDF_graph" %in% ls())
  rdf_free(RDF_graph)
  testthat::expect_false("RDF_graph" %in% ls())
})

testthat::test_that("we can concatenate rdfs", {
  rdf1 <- rdf_parse(system.file("extdata/ex.xml", package = "rdflib"))
  rdf2 <- rdf_parse(system.file("extdata/ex2.xml", package = "rdflib"))
  rdf <- c(rdf1, rdf2)
  testthat::expect_is(rdf, "rdf")
  
  rdf_free(rdf1)
  rdf_free(rdf2)
  ## NOTE: rdf is same pointer as rdf1, not a new pointer.  cannot free twice
})



testthat::test_that("we can add, parse and serialize json-ld", {
  x <- rdf()
  x <- rdf_add(x, 
               subject="http://www.dajobe.org/",
               predicate="http://purl.org/dc/elements/1.1/language",
               object="en")
  rdf_serialize(x, out, "jsonld")
  rdf <- rdf_parse(out, format = "jsonld")
  testthat::expect_is(rdf, "rdf")
  rdf_free(x)
  rdf_free(rdf)
  
})




testthat::test_that("print and format work", {
  rdf <- rdf_parse(doc)
  txt <- format(rdf, format = "rdfxml")
  testthat::expect_output(print(rdf), ".*johnsmith.*")
  
  testthat::expect_is(txt, "character")
  rdf_free(rdf)
})






unlink(out)

#' Check for BDB support
#' 
#' Detect whether Berkeley Database for disk-based storage of RDF graphs
#' is available.  Disk-based storage requires redland package
#' to be installed from source with support for the Berkeley DB 
#' (libdb-dev on Ubuntu, berkeley-db on homebrew), otherwise `rdf()` will
#' fall back to in-memory storage with a warning.
#' 
#' @return TRUE if BDB support is detected, false otherwise
#' @export
#' @examples 
#' rdf_has_bdb()
rdf_has_bdb <- function(){
 rdf_storage("BDB", new_db = TRUE, check_only = TRUE)
}

rdf_has_virtuoso <- function(user="dba", 
                             password="dba", 
                             dsn="Local Virtuoso"){
#  has_driver <- rdf_storage("virtuoso", 
#                            user = user, 
#                            password = password, 
#                            dsn = dsn, 
#                            check_only=TRUE)
  r <- tryCatch(rdf("virtuoso", user = user, 
           password = password, dsn = dsn, fallback = FALSE),
           error = function(e) FALSE)
  if(is.logical(r)){
    has_connection <- r
  } else {
    rdf_add(r, "", "dc:name", "bob")
    if(length(r) >= 1){
      has_connection <- TRUE
    } else { 
      has_connection <- FALSE
    }
    rdf_free(r)
  }

  has_connection
  
}

testthat::context("RDF Storage")

testthat::test_that("SQLite Backend", {
  testthat::skip_on_cran()
  
  testthat::skip_if_not(rdf_storage("sqlite", new_db = TRUE, 
                                    check_only = TRUE, name = "rdflib.sqlite"))
  testthat::expect_silent(r <- rdf(storage="sqlite", 
                                   new_db = TRUE,
                                   name="rdflib.sqlite"))
  rdf_add(r, "", "dc:name", "bob")
  testthat::expect_match(format(r, "nquads"), "bob")
  testthat::expect_is(r, "rdf")
  rdf_free(r)
  unlink("rdflib.sqlite")
  
  
})

testthat::test_that("Postgres Backend", {
  testthat::skip_on_travis()
  testthat::skip_on_cran()
  
  testthat::skip_if_not(rdf_storage("postgres", 
                                    host="postgres", user="postgres",
                                    password="rdflib", new_db = TRUE,
                                    check_only = TRUE))
  testthat::expect_silent(
    rdf <- rdf(storage="postgres", host = "postgres", 
               user="postgres", password="rdflib", new_db = TRUE)
  )
  
  rdf_add(rdf, "", "dc:name", "bob")
  testthat::expect_match(format(rdf, "nquads"), "bob")
  testthat::expect_is(rdf, "rdf")
  rdf_free(rdf)
  
})


## Note: `mysql` is the name default database created by mariadb
testthat::test_that("MySQL Backend", {
  testthat::skip_on_cran()
  testthat::skip_on_travis()
  testthat::skip_if_not(rdf_storage("mysql", host = "mariadb", 
                                    user="root", password="rdflib",
                                    database = "mysql",
                                    new_db=TRUE, check_only = TRUE ))
  testthat::expect_silent(
    rdf <- rdf(storage="mysql", host = "mariadb", 
               user="root", password="rdflib", 
               database = "mysql",
               new_db = TRUE)
  )
  rdf_add(rdf, "", "dc:name", "bob")
  expect_match(format(rdf, "nquads"), "bob")
  testthat::expect_is(rdf, "rdf")
  rdf_free(rdf)
  
})


testthat::test_that("Virtuoso Backend", {
  testthat::skip_on_travis()
  testthat::skip_on_cran()
  
  testthat::skip_if_not(rdflib:::rdf_has_virtuoso())
  
  testthat::expect_silent(
    r <- rdf(storage = "virtuoso", 
             user = "dba", 
             password = "dba", 
             dsn = "Local Virtuoso", 
             new_db = TRUE)
    )
  rdf_add(r, "", "dc:name", "bob")
  
  testthat::expect_match(format(r, "nquads"), "bob")
  testthat::expect_is(r, "rdf")
  rdf_free(r)
  
})



testthat::test_that("We warn if we cannot use disk-based storage", {
  testthat::skip_if(rdf_has_bdb())
  testthat::expect_warning(rdf <- rdf(storage = "BDB"), "BDB driver not found")
  ## Falls back on memory-based storage, still creates rdf
  testthat::expect_is(rdf, "rdf")
  rdf_free(rdf)
  
})

testthat::test_that("We can use BDB storage", {
  testthat::skip_if_not(rdf_has_bdb())
  
  # not sure why this is now failing on appveyor
  testthat::skip_on_os("windows")
  
  testthat::expect_silent(rdf <- rdf(storage="BDB", new_db = TRUE))
  
  rdf_add(rdf, "", "dc:name", "bob")
  expect_match(format(rdf, "nquads"), "bob")
  testthat::expect_is(rdf, "rdf")
  rdf_free(rdf)
  
  ## We can reconnect to disk based storage after freeing
  rdf2 <- rdf(storage = "BDB", new_db = FALSE)
  expect_match(format(rdf2, "nquads"), "bob")
  rdf_free(rdf2)
  
  unlink("rdflib-po2s.db")
  unlink("rdflib-so2p.db")
  unlink("rdflib-sp2o.db")
})

#' write object out as nquads
#'
#' @param x an object that can be represented as nquads
#' @param file output filename
#' @param ... additional parameters, see examples
#'
#' @export
#'
#' @examples
#' tmp <- tempfile(fileext = ".nq")
#' library(datasets)
#' 
#' ## convert data.frame to nquads
#' write_nquads(iris, tmp)
#' rdf <- read_nquads(tmp)
#' 
#' ## or starting a native rdf object
#' write_nquads(rdf, tempfile(fileext = ".nq"))
write_nquads <- function(x, file, ...){
  UseMethod("write_nquads")
}


#' read an nquads file
#' @param file path to nquads file
#' @param ... additional arguments to [rdf_parse()]
#' @return an rdf object.  See [rdf_parse()]
#' 
#' @examples
#' tmp <- tempfile(fileext = ".nq")
#' library(datasets)
#' write_nquads(iris, tmp)
#' read_nquads(tmp)
#' 
#' @export
read_nquads <- function(file, ...){
  rdf_parse(file, "nquads", ...)
}

#' @export
write_nquads.rdf <- function(x, file, ...){
  rdf_serialize(x, file, "nquads", ...)
}
  
#' @export
write_nquads.data.frame <- function(x, 
                         file,
                         ...){
  
  
  
  df <- normalize_table(x, ...)
  poor_mans_nquads(df, file, ...)
}


#' @importFrom tidyr gather
#' @importFrom dplyr left_join
normalize_table <- function(df, key_column = NULL, ...){
  ## gather looses col-classes, so pre-compute them (with base R)
  col_classes <- data.frame(datatype = 
                              vapply(df, 
                                     xs_class, 
                                     character(1)))
  col_classes$predicate <- rownames(col_classes)
  rownames(col_classes) <- NULL
  
  ## Use row names as key (subject), unless a key column is specified
  ## Should we verify that requested key column is indeed a unique key first?
  out <- df
  if (is.null(key_column)) {
    out$subject <- as.character(1:dim(out)[[1]])
  } else {
    names(out)[names(out) == key_column] <- "subject"
  }
  
  ## FIXME consider taking an already-gathered table to avoid dependency?
  
  suppressWarnings(# Possible warnings about mixed types
    out <- tidyr::gather(out, 
                       key = "predicate", 
                       value = "object", 
                       -"subject"))
  
  ## merge is Slow! ~ 5 seconds for 800K triples 
  ## (almost as much time as rdf_parse)
  # merge(out, col_classes, by = "predicate")
  
  dplyr::left_join(out, col_classes, by = "predicate")
  
}



## x is a data.frame with columns: subject, predicate, object, & datatype
#' @importFrom utils write.table
poor_mans_nquads <- function(x, file, prefix = NULL, ...){
  
  if (is.null(prefix)) {
    prefix <- paste0(deparse(substitute(x)), ":")
    warning(paste("prefix not declared, using", prefix))
  }
  
  prefix <- uri_prefix(prefix)
  ## Currently written to be base-R compatible, 
  ## but a tidyverse implementation may speed serialization.  
  ## However, this seems to be fast enough that it is rarely the bottleneck
  
  ## NOTE: paste0 is a little slow ~ 1 s on 800K triples
  ## No datatype on blank (missing) nodes
  
  blank_object <-is.na(x$object)
  blank_subject <- is.na(x$subject)
  
  x$datatype[blank_object] <- as.character(NA)
  ## NA needs to become a unique blank node number, could do uuid or _:r<rownum>
  x$object[blank_object] <- paste0("_:r", which(blank_object))
  x$subject[blank_subject] <- paste0("_:r", which(blank_subject))
  
  ## strings and URIs do not get a datatype
  needs_type <- !is.na(x$datatype)
  
  ## URIs that are not blank nodes need <>
  x$subject[!blank_subject] <- paste0("<", prefix, x$subject[!blank_subject], ">")
  ## Predicate is always a URI
  x$predicate <- paste0("<", prefix, x$predicate, ">")
  
  ## Strings should be quoted
  is_string <- !grepl("\\w+:\\w.*", x$object) &
    !needs_type & !blank_object
  x$object[is_string] <- paste0('\"', x$object[is_string] , '\"')
  
  ## URIs should be <> instead, but not blanks!
  x$object[!blank_object] <- gsub("(^\\w+:\\w.*$)", "<\\1>", 
                                  x$object[!blank_object])
  
  ## assumes datatype is not empty (e.g. string)
  x$object[needs_type] <- paste0('\"', x$object[needs_type], 
                                 '\"^^<', x$datatype[needs_type], ">")
  
  ## quads needs a graph column
  x$graph <- "."
  
  ## write table is a little slow, ~ 1s on 800K triples, 
  ## but readr cannot write in nquads style
  
  ## drop datatype
  x <- x[c("subject", "predicate", "object", "graph")]       
  utils::write.table(x, file, col.names = FALSE, quote = FALSE, row.names = FALSE)
}

#' Parse RDF Files
#'
#' @param doc path, URL, or literal string of the rdf document to parse
#' @param format rdf serialization format of the doc,
#' one of "rdfxml", "nquads", "ntriples", "turtle"
#' or "jsonld". If not provided, will try to guess based
#' on file extension and fall back on rdfxml.
#' @param rdf an existing rdf triplestore to extend with triples from
#' the parsed file.  Default will create a new rdf object.
#' @param base the base URI to assume for any relative URIs (blank nodes)
#' @param ... additional parameters (not implemented)
#'
#' @return an rdf object, containing the redland world
#'  and model objects
#' @importClassesFrom redland World Storage Model Parser
#' @importMethodsFrom redland parseFileIntoModel freeParser
#' @importFrom jsonld jsonld_to_rdf
#' @export
#'
#' @examples
#' doc <- system.file("extdata", "dc.rdf", package="redland")
#' rdf <- rdf_parse(doc)
#'
rdf_parse <- function(doc,
                      format = c("guess",
                                 "rdfxml",
                                 "nquads",
                                 "ntriples",
                                 "turtle",
                                 "jsonld"),
                      rdf = NULL,
                      base = getOption("rdf_base_uri", "localhost://"),
                      ...){
  
  format <- match.arg(format)
  if(format == "guess"){
    format <- guess_format(doc)
  }
  
  ## if we get a string as input, we'll store it in tmp file here
  ## which we can later be sure to clean up.
  tmp_string <- tempfile()
  ## if we get json-ld, we'll need a temp location to serialize that too:
  tmp_json <- tempfile()
  
  # convert string input or url to local file
  doc <- text_or_url_to_doc(doc, tmp_string)
  
  ## redlands doesn't support jsonld. So rewrite as nquads using jsonld package
  ## We use tmp to avoid altering input doc, since parsing a local file should
  ## be a read-only task!
  if(format == "jsonld"){
    x <- jsonld::jsonld_to_rdf(doc, 
                               options = 
           list(base = getOption("rdf_base_uri", "localhost://"),
                format = "application/nquads"))
    writeLines(x, tmp_json)
    format <- "nquads"
    doc <- tmp_json
  }
  
  if(is.null(rdf)){
    rdf <- rdf()
  }
    
  mimetype <- unname(rdf_mimetypes[format])
  parser <- new("Parser", rdf$world, name = format, mimeType = mimetype)
  redland::parseFileIntoModel(parser, rdf$world, doc, rdf$model, baseUri = base)
  
  redland::freeParser(parser)
  unlink(tmp_string)
  unlink(tmp_json)  
  
  ## return rdf object (pointer)
  rdf
}



# rdf functions like working with local files
# this helper function allows us to also use URLs or strings
#' @importFrom utils download.file
text_or_url_to_doc <- function(x, tmp = tempfile()){
  if(file.exists(x)){
    return(x) 
  } else if(grepl("^https?://", x)) {
    utils::download.file(x, tmp, quiet = TRUE)
    return(tmp)
  } else {
    writeLines(x, tmp)
    return(tmp)
  }
}

testthat::context("SPARQL Queries")

doc <- system.file("extdata/example.rdf", package="redland")
out <- "testing.rdf"

testthat::test_that("we can make CONSTRUCT queries", {

  testthat::skip("No test for SPARQL CONSTRUCT")
  
  ## No errors but no return either, not sure 
  ## what the correct construction is.  
  ## currently registers as a skipped test.  
  rdf <- rdf()
  rdf_add(rdf, 
          "http://carlboettiger.info", 
          "http://schema.org/name", 
          "Carl Boettiger")
  sparql <- 'CONSTRUCT { ?x ?p ?o . }
             WHERE { ?x ?p ?o . }'
  
  rdf_query(rdf, sparql, data.frame = FALSE)

  
  rdf_free(rdf)
  
})

testthat::test_that("we can make sparql queries", {
  sparql <-
    'PREFIX dc: <http://purl.org/dc/elements/1.1/>
  SELECT ?a ?c
  WHERE { ?a dc:creator ?c . }'
  
  rdf <- rdf_parse(doc)
  match <- rdf_query(rdf, sparql)
  testthat::expect_length(match, 2)
  
  rdf_free(rdf)
  
})

testthat::test_that("SPARQL handles data types", {
  
  rdf <- rdf()
  rdf_add(rdf, "", "ex:integer", 33L)
  rdf_add(rdf, "", "ex:decimal", 3.141)
  rdf_add(rdf, "", "ex:decimal", 2.718)
  rdf_add(rdf, "", "ex:logical", TRUE)
  rdf_add(rdf, "", "ex:Date", Sys.Date())
  rdf_add(rdf, "", "ex:POSIXct", Sys.time())
  rdf_add(rdf, "", "ex:factor", as.factor("text"))
  rdf_add(rdf, "", "ex:string", "text")
  
  ## Select ?s as well to avoid silly warning
  testthat::expect_is(rdf, "rdf")
  match <- rdf_query(rdf, 'SELECT ?o ?s WHERE { ?s <ex:Date> ?o }')
  testthat::expect_type(match$o[[1]], "double")
  testthat::expect_is(match$o[[1]], "Date")
  
  match <- rdf_query(rdf, 'SELECT ?o ?s WHERE { ?s <ex:POSIXct> ?o }')
  testthat::expect_is(match$o[[1]], "POSIXct")
  testthat::expect_type(match$o[[1]], "double")
  
  match <- rdf_query(rdf, 'SELECT ?o ?s WHERE { ?s <ex:decimal> ?o }')
  testthat::expect_is(match$o[[1]], "numeric")
  testthat::expect_type(match$o[[1]], "double")
  
  ## readr fails to detect "true" as a logical
  #match <- rdf_query(rdf, 'SELECT ?o ?s WHERE { ?s <ex:logical> ?o }')
  #testthat::expect_is(match$o[[1]], "logical")
  #testthat::expect_type(match$o[[1]], "logical")
  
  match <- rdf_query(rdf, 'SELECT ?o ?s WHERE { ?s <ex:integer> ?o }')
  testthat::expect_is(match$o[[1]], "numeric")
  testthat::expect_type(match$o[[1]], "double")
  
  match <- rdf_query(rdf, 'SELECT ?o ?s WHERE { ?s <ex:string> ?o }')
  testthat::expect_is(match$o[[1]], "character")
  testthat::expect_type(match$o[[1]], "character")
  
  
  ## Matching mixed type results in all types treated as character
  # vector, since o is a single column....
  match <- rdf_query(rdf, 'SELECT ?s ?p ?o WHERE { ?s ?p ?o }')
  testthat::expect_is(match$o, "character")
  testthat::expect_is(match, "data.frame")
  

  rdf_free(rdf)
  
})

unlink(out)

#' Free Memory Associated with RDF object
#' 
#' @param rdf an rdf object
#' @param rm logical, default TRUE. Remove pointer from parent.frame()?
#' Usually a good idea since referring to a pointer after it has been
#' removed can crash R.
#' @details Free all pointers associated with an rdf object. 
#' Frees memory associated with the storage, world, and model
#' objects. 
#' @export
#' @examples 
#' rdf <- rdf()
#' rdf_free(rdf)
#' rm(rdf)
rdf_free <- function(rdf, rm = TRUE){
  redland::freeModel(rdf$model)
  redland::freeStorage(rdf$storage)
  redland::freeWorld(rdf$world)

  ## Remove pointer if possible
  ## Since referring to this pointer after it has been
  ## freed would otherwise just crash R.
  if(rm){
    rm(list = deparse(substitute(rdf)), envir = parent.frame())
  }
}

testthat::context("RDF Add")

testthat::test_that("we can print.rdf UTF-8", {
  r <- rdf()
  rdf_add(r,
          subject="", 
          predicate="http://schema.org/name", 
          object="Maëlle Salmon")
  
  testthat::expect_output(print(r), "Maëlle")
  rdf_free(r)
})



testthat::test_that("we can initialize add triples to rdf graph", {
  rdf <- rdf()
  rdf <- rdf_add(rdf, 
                 subject="http://www.dajobe.org/",
                 predicate="http://schema.org/dateCreated",
                 object=as.Date("2015-01-01"))
  testthat::expect_is(rdf, "rdf")
  rdf_free(rdf)
  
})


testthat::test_that("we can initialize add triples to rdf graph", {
  rdf <- rdf()
  rdf <- rdf_add(rdf, 
                 subject=NA,
                 predicate="http://schema.org/dateCreated",
                 object=NA)
  testthat::expect_is(rdf, "rdf")
  rdf_free(rdf)
  
})

testthat::test_that("other rdf_add examples work",{
  
   rdf <- rdf()
   rdf_add(rdf, 
       subject="http://www.dajobe.org/",
       predicate="http://purl.org/dc/elements/1.1/language",
       object="en")
       
   ## non-URI string in subject indicates a blank subject
   ## (prefixes to "_:b0")
   rdf_add(rdf, "b0", "http://schema.org/jobTitle", "Professor") 
   
   ## identically a blank subject.  
   ## Note rdf is unchanged when we add the same triple twice.
   rdf_add(rdf, "b0", "http://schema.org/jobTitle", "Professor", 
           subjectType = "blank") 
           
   ## blank node with empty string creates a default blank node id
   rdf_add(rdf, "", "http://schema.org/jobTitle", "Professor")   
                       
   
   ## Subject and Object both recognized as URI resources:
   rdf_add(rdf, 
           "https://orcid.org/0000-0002-1642-628X",
           "http://schema.org/homepage", 
           "http://carlboettiger.info")  
  
    ## Force object to be literal, not URI resource        
   rdf_add(rdf, 
           "https://orcid.org/0000-0002-1642-628X",
           "http://schema.org/homepage", 
           "http://carlboettiger.info",
           objectType = "literal")  
           
  
   testthat::expect_is(rdf, "rdf")
   rdf_free(rdf)
})

#' Coerce an object into RDF
#' 
#' Coerce an object into RDF
#' @param x an object to coerce into RDF (list, list-like, or data.frame)
#' @param rdf An existing rdf object, (by default a new object will be initialized)
#' @param prefix A default vocabulary (URI prefix) to assume for all predicates
#' @param base A base URI to assume for blank subject nodes
#' @param context a named list mapping any string to a URI
#' @param key_column name of a column which should be 
#' treated as the primary key in a table. must be unique
#' @export
#' 
#' @examples 
#' as_rdf(mtcars)
#' as_rdf(list(repo = "rdflib", owner = list("id", "ropensci")))
as_rdf <- function(x, 
                   rdf = NULL,
                   prefix = NULL, 
                   base = getOption("rdf_base_uri", "localhost://"), 
                   context = NULL,
                   key_column = NULL) UseMethod("as_rdf")


#' @export
as_rdf.list <- function(x, 
                        rdf = NULL,
                        prefix = NULL, 
                        base = getOption("rdf_base_uri", "localhost://"), 
                        context = NULL,
                        key_column = NULL){
  
  if(is.null(rdf)){
    rdf <- rdf()  
  }
  ## unbox length-1 lists so we can apply a context successfully
  if(is.list(x) && length(x) == 1) x <- x[[1]]
  
  json <- jsonlite::toJSON(x, pretty = TRUE, auto_unbox = TRUE, force = TRUE)
  jsonld_context <- json_context(prefix, base, context)
  json2 <- paste0('{\n"@context":', jsonld_context, 
                  ',\n',  '"@graph": ', json,  '}')
  rdflib::rdf_parse(json2, "jsonld", rdf = rdf)
  invisible(rdf)
}


# helper function (identical to plyr::compact)
compact <- function (l) Filter(Negate(is.null), l)
json_context <- function(prefix = NULL, 
                         base = getOption("rdf_base_uri", "localhost://"), 
                         context = NULL){
  jsonlite::toJSON(
    compact(c(list("@vocab" = prefix,
                   "@base" = base),
              context)), auto_unbox = TRUE)
}



# test2 <- tibble(
#   age = 5L,
#   name = "bob",
#   height = 1.9,
#   address = "x:address",
#   knows = NA
# )
# as_rdf.data.frame(test2, prefix = "x:", key = "nquads")


## tidy data to rdf
#' @export
as_rdf.data.frame <- function(x,  
                              rdf = NULL, 
                              prefix = NULL, 
                              base = getOption("rdf_base_uri",
                                               "localhost://"), 
                              context = NULL, 
                              key_column = NULL){
  
  if(is.null(rdf)){
    rdf <- rdf()  
  }
  file <- tempfile()
  write_nquads(x, file = file, prefix = prefix, key_column = key_column)
  rdf <- rdf_parse(file, rdf = rdf, format = "nquads")
  unlink(file)
  invisible(rdf)
}


## Not used
## tidy data to rdf: use rownames as key column
## Note: this method is too slow to be practical on very large data frames
iterative_rdf_add <- function(df, prefix = "x:", base = prefix){
  rownames <- rownames(df)
  colnames <- colnames(df)
  for(i in 1:dim(df)[[1]]){
    for(j in 1:dim(df)[[2]]){
      rdf_add(rdf, 
              subject = paste0(base, as.character(rownames[[i]])),
              predicate = paste0(prefix, colnames[[j]]),
              object = df$object[[i]])
    }
  }
  invisible(rdf)
}

# Must match parser name & q 1.0 mimetype listed at:
# http://librdf.org/raptor/api/raptor-formats-types-by-parser.html
# 3 turtle options listed but only text/turtle works. 
# support may depend on libs avialable depending on when and how redland is compiled
rdf_mimetypes <- c("nquads" = "text/x-nquads",
                   "ntriples" = "application/n-triples",
                   "rdfxml" = "application/rdf+xml",
                   "trig" = "application/x-trig",
                   "turtle" = "text/turtle")

## My redland version does not find support for these, probably optional 
## additional dependency needed when compiling redland
# - trig  (application/x-trig)
# - n3 (text/n3) 
# - rdfa (application/xhtml+xml, or text/html)
# - rss (application/rss+xml or text/rss)



guess_format <- function(doc){
  switch(gsub(".*\\.(\\w+)$", "\\1", basename(doc)),
         "xml" = "rdfxml",
         "rdf" = "rdfxml",
         "json" = "jsonld",
         "nq" = "nquads",
         "nt" = "ntriples",
         "ttl" = "turtle",
         "jsonld" = "jsonld",
         "quads" = "nquads",
         "turtle" = "turtle",
         "rdfxml")
}


## Don't explicitly type characters as strings, since this is default
xs_class <- function(x){
  
    type <- switch(class(x)[[1]],
                   "numeric" = "xs:decimal",
                   "factor" = "xs:string",
                   "logical" = "xs:boolean",
                   "integer" = "xs:integer",
                   "Date" = "xs:date",
                   "POSIXct" = "xs:dateTime",
                   NULL
    )
  
  
  string <- gsub("^xs:", 
                 "http://www.w3.org/2001/XMLSchema#",
                 type)
  ## consistent return length, character(1)
  if (length(string) == 0) {
    string <- as.character(NA)
  }
  string
}


uri_prefix <- function(x){
  abs_uri <- grepl("^\\w+://", x)
  if (abs_uri) {
    if (!grepl("[#/]$", x)) return(paste0(x, "#"))
    return(x)
  }
  if (!grepl(":$", x)) return(paste0(x, ":"))
  x
}

context("utilities")

test_that("we can correctly set prefix formats", {

  uri_prefix(  "http://namespace.com/" )
  uri_prefix(  "http://namespace.com/terms#" )
  
  expect_match( uri_prefix( "schema" ), "schema:")
  expect_match( uri_prefix( "schema:" ), "schema:") 
})

#' Perform a SPARQL Query
#'
#' @param rdf an rdf object (e.g. from \code{\link{rdf_parse}})
#' @param query a SPARQL query, as text string
#' @param data.frame logical, should the results be returned as a data.frame? 
#' @param ... additional arguments to a redland initialize-Query
#'
#' @return a data.frame of all query results (default.)  Columns will 
#' be named according to variable names in the SPARQL query. Returned
#' object values will be coerced to match the corresponding R type 
#' to any associated datatype URI, if provided. If a column would 
#' result in mixed classes (e.g. strings and numerics), all types
#' in the column will be coerced to character strings. If `data.frame`
#' is false, results will be returned as a list with each element 
#' typed by its data URI.   
#' 
#'     
#' @importClassesFrom redland Query
#' @importMethodsFrom redland executeQuery
#' @importMethodsFrom redland freeQuery freeQueryResults
#' @export
#' @examples
#' doc <- system.file("extdata", "dc.rdf", package="redland")
#'
#' sparql <-
#' 'PREFIX dc: <http://purl.org/dc/elements/1.1/>
#'  SELECT ?a ?c
#'  WHERE { ?a dc:creator ?c . }'
#'
#' rdf <- rdf_parse(doc)
#' rdf_query(rdf, sparql)
#'
rdf_query <- function(rdf, query, data.frame = TRUE, ...){
  queryObj <- new("Query", rdf$world, query, ...)
  
  # ... defaults are: base_uri=NULL, query_language="sparql", query_uri=NULL
  
  queryResult <- redland::executeQuery(queryObj, rdf$model)
  out <- getResults(queryResult)
  redland::freeQueryResults(queryResult)
  redland::freeQuery(queryObj)
  
  out
}

## Notes
## readr does a pretty good job guessing types returned from sparql
## character, numeric, integer, Dates, POSIXct work fine
## logicals are denoted `true` and `false`, which readr mistakes for characters

## Redland only exports the getNextResult parser, which is extremely slow on large returns

#' @importFrom readr read_csv
#' @importFrom redland librdf_query_results_to_string2
getResults <- function(queryResult, format = "csv", ...){
  mimetype <- switch(format,
                     "csv" = "text/csv; charset=utf-8",
                     NULL)
  readr::read_csv(I(redland::librdf_query_results_to_string2(
                            queryResult@librdf_query_results, 
                            format, mimetype, NULL, NULL)),
                  progress = FALSE, show_col_types = FALSE,
                  ...)
}

#' Add RDF Triples
#'
#' add a triple (subject, predicate, object) to the RDF graph
#'
#' @param rdf an rdf object
#' @param subject character string containing the subject
#' @param predicate character string containing the predicate
#' @param object character string containing the object
#' @param subjectType the Node type of the subject, i.e. "uri", "blank"
#' @param objectType the Node type of the object, i.e. "literal", "uri", "blank"
#' @param datatype_uri the datatype URI to associate with a object literal value
#'
#' @return Silently returns the updated RDF graph (rdf object).
#' Since the rdf object simply contains external pointers
#' to the model object in C code, note that the input object is modified
#' directly, so you need not assign the output of rdf_add() to anything.
#'
#' @details 
#' 
#'  `rdf_add()` will automatically 'duck type' nodes (if looks like a duck...).
#'  That is, strings that look like URIs will be declared as URIs. (See
#'   [URI](https://en.wikipedia.org/wiki/Uniform_Resource_Identifier)).
#'  Predicate should always be a URI (e.g. URL or  a `prefix:string`),
#'  cannot be blank or literal.  Subjects that look like strings will be 
#'  treated as [Blank Nodes](https://en.wikipedia.org/wiki/Blank_node) (i.e.
#'  will be prefixed with `_:`).  An empty subject, `""`, will create a
#'  blank node with random name.  Objects that look like URIs will be
#'  typed as resource nodes, otherwise as literals.  An empty object `""`
#'  will be treated as blank node.  Set `subjectType` or `objectType` 
#'  explicitly to override this behavior, e.g. to treat an object URI
#'  as a literal string.  NAs are also treated as blank nodes in subject
#'  or object  See examples for details.
#' 
#' @references <https://en.wikipedia.org/wiki/Uniform_Resource_Identifier>
#' @importClassesFrom redland Statement
#' @importMethodsFrom redland addStatement freeStatement
#' @export
#'
#' @examples
#' rdf <- rdf()
#' rdf_add(rdf, 
#'     subject="http://www.dajobe.org/",
#'     predicate="http://purl.org/dc/elements/1.1/language",
#'     object="en")
#'     
#' ## non-URI string in subject indicates a blank subject
#' ## (prefixes to "_:b0")
#' rdf_add(rdf, "b0", "http://schema.org/jobTitle", "Professor") 
#' 
#' ## identically a blank subject.  
#' ## Note rdf is unchanged when we add the same triple twice.
#' rdf_add(rdf, "b0", "http://schema.org/jobTitle", "Professor", 
#'         subjectType = "blank") 
#'         
#' ## blank node with empty string creates a default blank node id
#' rdf_add(rdf, "", "http://schema.org/jobTitle", "Professor")   
#'                     
#' 
#' ## Subject and Object both recognized as URI resources:
#' rdf_add(rdf, 
#'         "https://orcid.org/0000-0002-1642-628X",
#'         "http://schema.org/homepage", 
#'         "http://carlboettiger.info")  
#'
#'  ## Force object to be literal, not URI resource        
#' rdf_add(rdf, 
#'         "https://orcid.org/0000-0002-1642-628X",
#'         "http://schema.org/homepage", 
#'         "http://carlboettiger.info",
#'         objectType = "literal")  
#'         
#'
rdf_add <- function(rdf, subject, predicate, object, 
                    subjectType = as.character(NA), 
                    objectType = as.character(NA), 
                    datatype_uri = as.character(NA)){
  

  ## determine appropriate datatype URI in standard XSD Namespace 
  ## based on the R object class, if no datatype has been provided.
  if(is.na(datatype_uri)){
    datatype_uri <- xs_class(object)
  }
  
  ## predicate as blank node
  if(is.na(subject)){
    subject <- ""
  }
  if(is.na(object)){
    object <- ""
    datatype_uri <- as.character(NA)
  }

  
  stmt <- new("Statement", world = rdf$world, 
              subject, predicate, as.character(object),
              subjectType, objectType, datatype_uri)
  redland::addStatement(rdf$model, stmt)
  
  redland::freeStatement(stmt)
  ## rdf object is a list of pointers, modified in pass-by-reference
  invisible(rdf)
}

#' Initialize an `rdf` Object
#'
#' @param storage Storage backend to use; see details
#' @param host host address for mysql, postgres, or virtuoso storage
#' @param port port for mysql (mysql storage defaults to mysql standard port, 3306)
#' or postgres (postgres storage defaults to postgres standard port, 4321)
#' @param user user name for postgres, mysql, or virtuoso
#' @param password password for postgres, mysql, or virtuoso
#' @param database name of the database to be created/used
#' @param charset charset for virtuoso database, if desired
#' @param dir directory of where to write sqlite or berkeley database.
#' @param dsn Virtuoso dsn, either "Local Virtuoso" or "Remote Virtuoso"
#' @param name name for the storage object created. Default is usually fine.  
#' @param new_db logical, default FALSE. Create new database or connect to existing?
#' @param fallback logical, default TRUE. If requested storage system cannot initialize,
#' should `rdf()` fall back on memory (default) or throw an error (fallback=FALSE)?
#' @return an rdf object
#' @details an rdf Object is a list of class 'rdf', consisting of
#' three pointers to external C objects managed by the redland library.
#' These are the `world` object: basically a top-level pointer for
#' all RDF models, and a `model` object: a collection of RDF statements,
#' and a `storage` object, indicating how these statements are stored.
#' 
#' `rdflib` defaults to an in-memory hash-based storage structure. 
#' which should be best for most use cases. For very large triplestores,
#' disk-based storage will be necessary. Enabling external storage devices
#' will require additional libraries and custom compiling. See the storage
#' vignette for details.  
#'
#' @importClassesFrom redland World Model Storage
#' @importMethodsFrom redland freeWorld freeModel freeStorage
#' @importFrom utils capture.output
#' @export
#'
#' @examples
#' x <- rdf()
#' 
rdf <- function(storage = c("memory", "BDB", "sqlite", 
                            "postgres", "mysql", "virtuoso"), 
                host = NULL,
                port = NULL,
                user = NULL,
                password = NULL,
                database = NULL,
                charset = NULL,
                dir = NULL,
                dsn = "Local Virtuoso",
                name = "rdflib",
                new_db = FALSE,
                fallback = TRUE){
  
  world <- new("World")
  store <- rdf_storage(storage, world, host, port, user, password, 
                       database, charset, dir, dsn, name, new_db, fallback)
  model <- new("Model", world = world, storage = store, options = "")
  structure(list(world = world, model = model, storage = store),
            class = "rdf")
}






rdf_storage <- function(storage = c("memory", 
                                    "BDB", 
                                    "sqlite", 
                                    "postgres", 
                                    "mysql", 
                                    "virtuoso"),
                        world = NULL,
                        host = NULL,
                        port = NULL,
                        user = NULL,
                        password = NULL,
                        database = NULL,
                        charset = NULL,
                        dir = NULL,
                        dsn = "Local Virtuoso",
                        name = "rdflib",
                        new_db = FALSE,
                        fallback = TRUE,
                        check_only = FALSE){
  if(is.null(world)){
    world <- new("World")
  }
  storage <- match.arg(storage)
  
  new <- NULL
  if(new_db){
    new <- "yes"
  }
  if(is.null(dir)){
    dir <- "."
  }
  
  options <- options_to_str(
    switch(storage,
      memory = list("hash-type" = "memory"),
         BDB = list(new = new, "hash-type" = "bdb", dir = dir),  
      sqlite = list(new = new, dir = dir),
    postgres = list(new = new, host = host, port = port,
                    database = database, user = user, password = password),
       mysql = list(new = new, host = host, port = port,
                    database = database, user = user, password = password),
    virtuoso = list(dsn = dsn, user = user, password = password, 
                    database = database, host = host, charset = charset),
    list()
    ))
  
  store <- switch(storage,
    memory = new("Storage", world, "hashes", name = name, options = options),
    BDB = new("Storage", world, "hashes", name = name, options = options),
    sqlite = new("Storage", world, "sqlite", name = name, options = options),
    postgres = new("Storage", world, "postgresql", name = name, options = options),
    mysql =  new("Storage", world, "mysql", name = name, options = options),
    virtuoso =  new("Storage", world, "virtuoso", name = name, options = options)
  ) 
  
  continue <- !is_null_pointer(store)
  
  if(check_only){
    freeStorage(store)
    freeWorld(world)
    return(continue)
  }
  
  if(!continue){
    if(fallback){
       warning(paste(storage, "driver not found. Falling back on in-memory storage"))
       redland::freeStorage(store)
       store <- new("Storage", world)
     } else {
       stop(paste(storage, "not found"))
     }
  }
 
  store
}


compact <- function(l){ Filter(Negate(is.null), l)}
options_to_str <- function(x){
  x <- compact(x)
  n <- names(x)
  out <- character(0)
  for(i in seq_along(x)){
    out <- paste0(c(out, paste0(n[[i]], "=", "'", x[[i]], "'")), collapse=",")
  }
  out
}

#' @importFrom redland is.null.externalptr
is_null_pointer <- function(x){
  redland::is.null.externalptr(x@librdf_storage@ref)
#  utils::capture.output(base::print.default(x@librdf_storage@ref)) == 
#  "<pointer: 0x0>"
}


#' rdflib: Tools to Manipulate and Query Semantic Data
#'
#' The Resource Description Framework, or RDF is a widely used
#' data representation model that forms the cornerstone of the 
#' Semantic Web. 'RDF' represents data as a graph rather than 
#' the familiar data table or rectangle of relational databases.
#'
#' 
#' It has three main goals:
#'
#' \itemize{
#' \item Easily read, write, and convert between all major RDF serialization formats
#' \item Support SPARQL queries to extract data from an RDF graph into a data.frame
#' \item Support JSON-LD format as a first-class citizen in RDF manipulations
#' }
#'
#' For more information, see the Wikipedia pages for RDF, SPARQL, and JSON-LD:
#' 
#' \itemize{
#' \item \url{https://en.wikipedia.org/wiki/Resource_Description_Framework}
#' \item \url{https://en.wikipedia.org/wiki/SPARQL}
#' \item \url{https://en.wikipedia.org/wiki/JSON-LD}
#' }
#'
#' To learn more about rdflib, start with the vignettes:
#' `browseVignettes(package = "rdflib")`
#'
#'  Configurations via `options()`
#' 
#' `rdf_print_format`:
#'  
#' - NULL or "nquads" (default)
#' - any valid serializer name: e.g. "rdfxml", "jsonld", "turtle",  "ntriples"
#'   
#' `rdf_base_uri`:
#' 
#' - Default base URI to use (when serializing JSON-LD only at this time)
#'     default is "localhost://"
#'     
#' `rdf_max_print`:
#' 
#' - maximum number of lines to print from rdf, default 10
#'
#'
"_PACKAGE"

#!/usr/bin/env bash

# script_name="${0}"
script_dirname="$( dirname -- "${0}" )"
# script_basename="$( basename -- "${0}" )"

set -eo pipefail

wait_http_okay() {
    local timeout=30
    for var in "${@}"
    do
        eval "local ${var}"
    done
    1>&2 echo "INFO: checking if ${url} (${label}) is okay"
    for ((i=1;i<=${timeout};i++))
    do
        curl --silent "${url}" --fail -o /dev/null && {
            1>&2 echo "INFO: url ${url} (${label}) is okay"
            return 0
        }
        kill -0 "${pid}" 2>/dev/null 1>&2 || {
            1>&2 echo "ERROR: pid ${pid} (${label}) not running anymore, see log dump below for the possible cause."
            return 1
        }
        1>&2 echo "INFO: url ${url} (${label}) not okay, waiting ..."
        sleep 2
    done
    1>&2 echo "ERROR: timed out trying to load ${url} (${label}), see log dump below for the possible cause."
    return 1
}

assert_pid_running() {
    for var in "${@}"
    do
        eval "local ${var}"
    done

    kill -0 "${pid}" 2>/dev/null 1>&2 || {
        1>&2 echo "ERROR: pid ${pid} (${label}) not running anymore, see log dump below for the possible cause."
        return 1
    }
    1>&2 echo "INFO: pid ${pid} (${label}) running"
    return 0
}

kill_and_wait_pid() {
    local timeout=10
    for var in "${@}"
    do
        eval "local ${var}"
    done
    1>&2 echo "INFO: killing pid ${pid} (${label})"
    kill "${pid}"
    for ((i=1;i<=${timeout};i++))
    do
        kill -0 "${pid}" 2>/dev/null 1>&2 || {
            1>&2 echo "INFO: pid ${pid} (${label}) is dead "
            return 0
        }
        1>&2 echo "INFO: pid ${pid} (${label}) is not dead, waiting ..."
        sleep 2
    done
    1>&2 echo "ERROR: wait timeout for pid ${pid} (${label})"
    return 1
}

fuseki_pid_normal=""
fuseki_pid_tdb=""

fuseki_log_normal=""
fuseki_log_tdb=""

exit_handler() {
    1>&2 declare -p fuseki_pid_normal fuseki_pid_tdb
    if [ -z "${FUSEKI_SKIP_SHUTDOWN}" ]
    then
        if [ -n "${fuseki_pid_normal}" ]
        then
            kill_and_wait_pid "pid=${fuseki_pid_normal}" "label=fuseki-normal" || :
        fi
        if [ -n "${fuseki_pid_tdb}" ]
        then
            kill_and_wait_pid "pid=${fuseki_pid_tdb}" "label=fuseki-tdb" || :
        fi
    else
        1>&2 echo "WARNING: not killing fuseki fuseki_pid_normal=${fuseki_pid_normal} fuseki_pid_tdb=${fuseki_pid_tdb}"
    fi

    local -a dump_cmd
    if [ -n "${FUSEKI_DUMP_FULL_LOGS}" ]
    then
        dump_cmd=(cat)
    else
        dump_cmd=(tail -15)
    fi

    if [ -n "${FUSEKI_DUMP_LOGS}" ] || [ "${xrc}" != "0" ]
    then
        1>&2 echo "dumping fuseki_log_normal=${fuseki_log_normal}"
        "${dump_cmd[@]}" "${fuseki_log_normal}" || :
        1>&2 echo "dumping fuseki_log_tdb=${fuseki_log_tdb}"
        "${dump_cmd[@]}"  "${fuseki_log_tdb}" || :
    fi
}


main() {
    : "${LOCALSTATEDIR:=${script_dirname}/var}"
    # : ${FUSEKI_PORT:=3030}
    : "${XDG_CACHE_HOME:=${HOME}/.cache}"

    local jena_uri="https://archive.apache.org/dist/jena/binaries/apache-jena-fuseki-3.17.0.tar.gz"
    local jena_sha512="2b92f3304743da335f648c1be7b5d7c3a94725ed0a9b5123362c89a50986690114dcef0813e328126b14240f321f740b608cc353417e485c9235476f059bd380"
    local jena_archive_basename
    jena_archive_basename="$(basename "${jena_uri}")"
    local jena_archive="${XDG_CACHE_HOME}/${jena_archive_basename}"
    local jena_stem="${jena_archive_basename%%.tar.gz}"

    1>&2 declare -p jena_uri jena_archive XDG_CACHE_HOME LOCALSTATEDIR
    if ! [ -e "${jena_archive}" ]
    then
        mkdir -vp "${XDG_CACHE_HOME}"
        curl "${jena_uri}" -o "${jena_archive}"
    fi
    printf "%s  %s\n" "${jena_sha512}" "${jena_archive_basename}" > "${jena_archive}.sha512"
    (cd "${XDG_CACHE_HOME}" && shasum -c -a512 "${jena_archive_basename}.sha512") || {
        echo 1>&2 "ERROR: digest verification failed"
        rm -rv "${jena_archive}" "${jena_archive}.sha512"
    }
    mkdir -vp "${LOCALSTATEDIR}"
    tar -zxf "${jena_archive}" -C "${LOCALSTATEDIR}"

    local FUSEKI_HOME=${LOCALSTATEDIR}/${jena_stem}
    1>&2 declare -p FUSEKI_HOME
    export FUSEKI_HOME

    fuseki_base_normal="${FUSEKI_HOME}/run-normal"
    fuseki_base_tdb="${FUSEKI_HOME}/run-tdb"
    local fuseki_pidfile_normal="${fuseki_base_normal}/server.pid"
    local fuseki_pidfile_tdb="${fuseki_base_tdb}/server.pid"

    mkdir -vp "${fuseki_base_normal}" "${fuseki_base_tdb}"
    local fuseki_port_normal=3030
    local fuseki_port_tdb=3031

    1>&2 declare -p fuseki_base_normal fuseki_base_tdb fuseki_pidfile_normal fuseki_pidfile_tdb fuseki_port_normal fuseki_port_tdb

    trap exit_handler EXIT

    fuseki_log_normal="${fuseki_base_normal}/out.log"
    fuseki_log_tdb="${fuseki_base_tdb}/out.log"

    1>&2 declare -p fuseki_log_normal fuseki_log_tdb

    1>&2 echo "INFO: starting fuseki: normal"
    FUSEKI_BASE="${fuseki_base_normal}" bash "${FUSEKI_HOME}/fuseki-server" \
        --port "${fuseki_port_normal}" --debug \
        --update --mem /db &>"${fuseki_log_normal}" &
    fuseki_pid_normal="${!}"
    echo "${fuseki_pid_normal}" > "${fuseki_pidfile_normal}"

    1>&2 echo "INFO: starting fuseki: tdb"
    FUSEKI_BASE="${fuseki_base_tdb}" bash "${FUSEKI_HOME}/fuseki-server" \
        --port "${fuseki_port_tdb}" --debug \
        --update --memTDB --set tdb:unionDefaultGraph=true /db &>"${fuseki_log_tdb}" &
    fuseki_pid_tdb="${!}"
    echo "${fuseki_pid_tdb}" > "${fuseki_pidfile_tdb}"

    1>&2 declare -p fuseki_pid_tdb fuseki_pid_normal

    wait_http_okay "pid=${fuseki_pid_normal}" "url=http://localhost:${fuseki_port_normal}/" "label=fuseki-normal"
    wait_http_okay "pid=${fuseki_pid_tdb}" "url=http://localhost:${fuseki_port_tdb}/" "label=fuseki-tdb"

    assert_pid_running "pid=${fuseki_pid_tdb}" "label=fuseki-tdb"
    assert_pid_running "pid=${fuseki_pid_normal}" "label=fuseki-normal"

    local -a args=("${@}")

    1>&2 echo "running: ${args[*]}"

    local xrc=1
    "${args[@]}"; xrc="${?}"
    exit "${xrc}"
}

main "${@}"

# RDFLib Contributing Guide

Thank you for considering contributing to RDFLib. This project has no formal
funding or full-time maintainers, and relies entirely on independent
contributors to keep it alive and relevant.

## Ways to contribute

Some ways in which you can contribute to RDFLib are:

- Address open issues:
  [![GitHub issues](https://img.shields.io/github/issues/RDFLib/rdflib)](https://github.com/RDFLib/rdflib/issues)
- Fix
  [expected failure](https://docs.pytest.org/en/latest/how-to/skipping.html#xfail-mark-test-functions-as-expected-to-fail)
  tests: [![GitHub search query](https://img.shields.io/badge/GitHub-search-green)](https://github.com/search?q=xfail+repo%3ARDFLib%2Frdflib+path%3Atest%2F**.py&amp%3Btype=code&type=code)
- Add additional
  [expected failure](https://docs.pytest.org/en/latest/how-to/skipping.html#xfail-mark-test-functions-as-expected-to-fail)
  tests for open issues:
  [![GitHub issues](https://img.shields.io/github/issues/RDFLib/rdflib)](https://github.com/RDFLib/rdflib/issues)
- Add tests for untested code:
  [![Coveralls branch](https://img.shields.io/coveralls/RDFLib/rdflib/main.svg)](https://coveralls.io/r/RDFLib/rdflib?branch=main)
- Review pull requests marked with the
  [![review wanted](https://img.shields.io/badge/-review%20wanted-28ead2)](https://github.com/RDFLib/rdflib/labels/review%20wanted)
  label.
- Answer questions on Stack Overflow:
  [![Stack Exchange questions](https://img.shields.io/stackexchange/stackoverflow/t/rdflib)](https://stackoverflow.com/questions/tagged/rdflib)
- Convert
  [`unittest`](https://docs.python.org/3/library/unittest.html)
  based tests to
  [`pytest`](https://docs.pytest.org/en/latest/)
  based tests:
  [![GitHub search query](https://img.shields.io/badge/GitHub-search-green)](https://github.com/search?q=unittest+repo%3ARDFLib%2Frdflib+path%3Atest%2F**.py&type=code)
- Add, correct or improve docstrings:
  [![rtd latest](https://img.shields.io/badge/docs-latest-informational)](https://rdflib.readthedocs.io/en/latest/)
- Update the RDFLib Wikipedia entry:
  [![Wikipedia: RDFLib](https://img.shields.io/badge/Wikipedia-RDFLib-informational)](https://en.wikipedia.org/wiki/RDFLib)
- Update the RDFLib Wikidata entry:
  [![Wikidata: Q7276224](https://img.shields.io/badge/Wikidata-Q7276224-informational)](https://www.wikidata.org/wiki/Q7276224)
- Participate on Gitter/Matrix chat:
  [![Gitter](https://badges.gitter.im/RDFLib/rdflib.svg)](https://gitter.im/RDFLib/rdflib?utm_source=badge&utm_medium=badge&utm_campaign=pr-badge) [![Matrix](https://img.shields.io/matrix/rdflib:matrix.org?label=matrix.org%20chat)](https://matrix.to/#/#RDFLib_rdflib:gitter.im)
- Participate in GitHub discussions:
  [![GitHub Discussions](https://img.shields.io/github/discussions/RDFLib/rdflib)](https://github.com/RDFLib/rdflib/discussions)
- Fix linting failures (see ruff settings in `pyproject.toml` and `#
  noqa:` directives in the codebase).

## Pull Requests

Contributions that involve changes to the RDFLib repository have to be made with
pull requests and should follow the [RDFLib developers guide](./developers.md).

For changes that add features or affect the public API of RDFLib, it is
recommended to first open an issue to discuss the change before starting to work
on it. That way you can get feedback on the design of the feature before
spending time on it.

## Code of Conduct

All contributions to the project should be consistent with the
[code of conduct](./CODE_OF_CONDUCT.md) adopted by RDFLib.

# Default Branch Name

!!! success "Status"
    Accepted

## Context

In recent years usage of the word `master` has become somewhat controversial [as noted by SFC][SFC-BNAMING] and consequently default branch name of Git repos has become `main`, both in Git itself [according to SFC][SFC-BNAMING] and in Git hosting solutions such as GitHub [documentation][GH-BRANCHES].

## Decision

RDFLib's default branch will be renamed from `master` to `main`. This is primarily to stay in line with modern conventions and to adhere to the principle of least surprise.

## Consequences

Anticipated negative consequences:

* Some links to old code will be broken.
* Some people's workflow may break unexpectedly and need adjusting.
* Any code and systems reliant on the old default branch name will fail.

Anticipated positive consequences:

* It will become a bit easier to work with RDFLib for developers that are used
  to `main` as the default branch.

## References

[GH-BRANCHES]: https://docs.github.com/en/pull-requests/collaborating-with-pull-requests/proposing-changes-to-your-work-with-pull-requests/about-branches#about-the-default-branch "GitHub: About the default branch"
[SFC-BNAMING]: https://sfconservancy.org/news/2020/jun/23/gitbranchname/ "Regarding Git and Branch Naming"

# Upgrading from version 6 to 7

## Python version

RDFLib 7 requires Python 3.8.1 or later.

## New behaviour for `publicID` in `parse` methods

Before version 7, the `publicID` argument to the [`parse()`][rdflib.graph.ConjunctiveGraph.parse] and [`parse()`][rdflib.graph.Dataset.parse] methods was used as the name for the default graph, and triples from the default graph in a source were loaded into the graph
named `publicID`.

In version 7, the `publicID` argument is only used as the base URI for relative URI resolution as defined in [IETF RFC 3986](https://datatracker.ietf.org/doc/html/rfc3986#section-5.1.4).

To accommodate this change, ensure that use of the `publicID` argument is consistent with the new behaviour.

If you want to load triples from a format that does not support named graphs into a named graph, use the following code:

```python
from rdflib import ConjunctiveGraph

cg = ConjunctiveGraph()
cg.get_context("example:graph_name").parse("http://example.com/source.ttl", format="turtle")
```

If you want to move triples from the default graph into a named graph, use the following code:

```python
from rdflib import ConjunctiveGraph

cg = ConjunctiveGraph()
cg.parse("http://example.com/source.trig", format="trig")
destination_graph = cg.get_context("example:graph_name")
for triple in cg.default_context.triples((None, None, None)):
    destination_graph.add(triple)
    cg.default_context.remove(triple)
```

[Eclipse RDF4J](https://rdf4j.org/) is an open-source Java framework for working with RDF. Out of the box, it supports running RDF database servers and allows clients to interact with these servers through an HTTP protocol known as the RDF4J REST API. This API fully supports all SPARQL 1.1 W3C Recommendations and also provides additional operations for managing RDF4J concepts such as repositories and transactions. For more details about the RDF4J REST API and its capabilities, see the [official documentation](https://rdf4j.org/documentation/reference/rest-api/).

RDFLib provides two options to interface with the RDF4J REST API:

1. **Store integration** - A full RDFLib Store implementation that lets users interact with RDF4J repositories seamlessly through RDFLib’s [`Graph`][rdflib.Graph] and [`Dataset`][rdflib.Dataset] classes.
2. **RDF4J Client** - A lower-level client that gives users more control over interactions with the RDF4J REST API. It covers all operations supported by the RDF4J REST API, including managing repositories and transactions.

To use RDFLib with RDF4J, first install RDFLib with the optional `rdf4j` extra.

```shell
pip install rdflib[rdf4j]
```

!!! note

    The minimum RDF4J protocol version supported by RDFLib is `12`. Versions less than 12 will raise an [`RDF4JUnsupportedProtocolError`][rdflib.contrib.rdf4j.exceptions.RDF4JUnsupportedProtocolError].

## RDF4J Store

An RDF4J Store connects to a single RDF4J repository. If you need to work with multiple repositories, you can create multiple Store instances. The RDF4J Store exposes only the subset of repository operations that map directly with RDFLib’s Graph and Dataset interfaces. If you need to perform additional operations such as managing repositories, you should use the RDF4J Client instead.

### Connecting to an existing repository

To get started, import the RDF4J Store class, create an instance of it, and pass it as the store parameter when creating a [`Graph`][rdflib.Graph] or [`Dataset`][rdflib.Dataset].

The following example connects to a local RDF4J server running on port 7200 and accesses an existing repository with the identifier `my-repository`.

If the server requires basic authentication, you can optionally pass a (username, password) tuple to the `auth` parameter.

```python
from rdflib.plugins.stores.rdf4j import RDF4JStore

store = RDF4JStore(
    base_url="http://localhost:7200/",
    repository_id="my-repository",
    auth=("username", "password"),
)
```

### Creating a new repository

If the repository does not exist (and the `create` parameter is set to `False`), an exception will be raised. To create the repository, set the `create` parameter to `True`. This will create the repository using RDF4J’s default configuration settings and then connect to it.

```python
store = RDF4JStore(
    base_url="http://localhost:7200/",
    repository_id="my-repository",
    auth=("username", "password"),
    create=True,
)
```

To create a repository with a custom RDF4J configuration, pass the `configuration` as a string to the configuration parameter.

See the RDF4J documentation for more information on [Repository and SAIL Configuration](https://rdf4j.org/documentation/reference/configuration/).

```python
store = RDF4JStore(
    base_url="http://localhost:7200/",
    repository_id="my-repository",
    auth=("username", "password"),
    configuration=configuration,
    create=True,
)
```

### Using the store with RDFLib

Once the store is created, you can create a [`Graph`][rdflib.Graph] or [`Dataset`][rdflib.Dataset] object and use it as you would with any other RDFLib store.

```python
from rdflib import Dataset

ds = Dataset(store=store)
```

For more information on how to use RDFLib Graphs and Datasets, see the subsections under [Usage](index.md).

### Namespaces

When connecting to an RDF4J repository, RDFLib automatically registers a set of built-in namespace prefixes. To disable this behavior, assign a new RDFLib [`NamespaceManager`][rdflib.namespace.NamespaceManager] instance with `bind_namespaces` set to the string `"none"`.

```python
ds = Dataset(store=store)
ds.namespace_manager = NamespaceManager(ds, "none")
```

See [`NamespaceManager`][rdflib.namespace.NamespaceManager] for more namespace binding options.

## RDF4J Client

This section covers examples of how to use the RDF4J Client. For the full reference documentation of the RDF4J Client, see [rdflib.contrib.rdf4j.client](apidocs/rdflib.contrib.rdf4j.client.md).

### Creating a client instance

The `RDF4JClient` class is the main entry point for interacting with the RDF4J REST API. To create an instance, pass the base URL of the RDF4J server to the constructor and optionally a username and password tuple for basic authentication.

The preferred way to create a client instance is to use Python's context manager syntax (`with` statement). When using this syntax, the client will automatically close when the block is exited.

```python
from rdflib.contrib.rdf4j.client import RDF4JClient

with RDF4JClient("http://localhost:7200/", auth=("admin", "admin")) as client:
    # Perform your operations here.
    ...
```

Alternatively, you can create a client instance and manually close it when you are done using it.

```python
from rdflib.contrib.rdf4j.client import RDF4JClient

client = RDF4JClient("http://localhost:7200/", auth=("admin", "admin"))
try:
    # Perform your operations here.
    ...
finally:
    client.close()
```

!!! Note

    You can pass a token string directly to the `auth` parameter instead of a (username, password) tuple. This will set the Authorization header to use Bearer token authentication for all requests.

### HTTP client configuration

The RDF4J Client uses the [httpx](https://www.python-httpx.org/) library for making HTTP requests. When creating an RDF4J Client instance, any additional keyword arguments to [`RDF4JClient`][rdflib.contrib.rdf4j.client.RDF4JClient] will be passed on to the underlying [`httpx.Client`](https://www.python-httpx.org/api/#client) instance.

For example, setting additional headers (such as an Authorization header) for all requests can be done as follows:

```python
token = "secret"
headers = {
    "Authorization": f"Bearer {token}"
}
with RDF4JClient("http://localhost:7200/", headers=headers) as client:
    # Perform your operations here.
    ...
```

The [`httpx.Client`](https://www.python-httpx.org/api/#client) instance is available on the RDF4J client's [`http_client`][rdflib.contrib.rdf4j.client.RDF4JClient.http_client] property.

```python
client.http_client
```

All requests reuse the same underlying HTTP client session, providing connection pooling and performance benefits.

### The repository manager

The RDF4J Client provides a [`RepositoryManager`][rdflib.contrib.rdf4j.client.RepositoryManager] class that allows you to manage RDF4J repositories. It does not represent a repository itself; instead, it is responsible for creating, deleting, listing, and retrieving [`Repository`][rdflib.contrib.rdf4j.client.Repository] instances.

You can access the repository manager on the RDF4J client instance using the [`repositories`][rdflib.contrib.rdf4j.client.RDF4JClient.repositories] property.

```python
client.repositories
```

#### Create a repository

To create a new repository, call the [`create`][rdflib.contrib.rdf4j.client.RepositoryManager.create] method on the repository manager. You must provide the repository identifier as well as the RDF4J repository configuration as an RDF string. The default configuration format expected is `text/turtle`.

See the RDF4J documentation for more information on [Repository and SAIL Configuration](https://rdf4j.org/documentation/reference/configuration/).

```python
configuration = """
    PREFIX config: <tag:rdf4j.org,2023:config/>

    []    a config:Repository ;
        config:rep.id "my-repository" ;
        config:rep.impl
            [
                config:rep.type "openrdf:SailRepository" ;
                config:sail.impl
                    [
                        config:native.tripleIndexers "spoc,posc" ;
                        config:sail.defaultQueryEvaluationMode "STANDARD" ;
                        config:sail.iterationCacheSyncThreshold "10000" ;
                        config:sail.type "openrdf:NativeStore" ;
                    ] ;
            ] ;
    .
"""
repo = client.repositories.create("my-repository", configuration)
```

For the full reference documentation of the RepositoryManager class, see [`RepositoryManager`][rdflib.contrib.rdf4j.client.RepositoryManager].

### Working with a repository

The [`Repository`][rdflib.contrib.rdf4j.client.Repository] class provides attributes and methods for interacting with a single RDF4J repository. This includes:

- Checking the health of the repository
- Retrieving the number of statements in the repository
- Querying the repository using both SPARQL and the Graph Store Protocol
- Updating the repository using both SPARQL Update and the Graph Store Protocol
- Perform queries and updates in a transaction context
- Manage the namespace prefixes of the repository

#### Retrieve a repository instance

To get started, retrieve a repository instance by repository identifier.

```python
repo = client.repositories.get("my-repository")
```

!!! note

    When retrieving a repository instance, it will automatically perform a health check to ensure that the repository is healthy. To perform a manual health check, simply call the [`health`][rdflib.contrib.rdf4j.client.Repository.health] method.

#### Repository size

To get the number of statements in a repository, call the [`size`][rdflib.contrib.rdf4j.client.Repository.size] method.

```python
repo.size()
```

#### Repository data

Upload some RDF data to the repository. Keep in mind that the [`upload`][rdflib.contrib.rdf4j.client.Repository.upload] method always appends the data to the repository.

```python
data = """
    PREFIX ex: <https://example.com/>
    ex:Bob a ex:Person .
"""
repo.upload(data=data)
```

!!! note

    Methods such as upload that accepts a data argument can automatically handle different kinds of inputs. For example, the data parameter in this case is an RDF string, but it can also be a file path, a file-like object, bytes data, or an RDFLib [`Graph`][rdflib.Graph] or [`Dataset`][rdflib.Dataset].

To overwrite the repository data, use the [`overwrite`][rdflib.contrib.rdf4j.client.Repository.overwrite] method.

!!! warning

    This will delete all existing data in the repository and replace it with the provided data.

```python
repo.overwrite(data=data)
```

There are also methods ([`get`][rdflib.contrib.rdf4j.client.Repository.get] and [`delete`][rdflib.contrib.rdf4j.client.Repository.delete]) for querying and deleting data from the repository using a triple or quad pattern match. However, keep in mind that the patterns cannot include [`BNode`][rdflib.BNode] terms.

The following example demonstrates the [`delete`][rdflib.contrib.rdf4j.client.Repository.delete] method, which deletes all statements with the subject `https://example.com/Bob`.

```python
repo.delete(subject=URIRef("https://example.com/Bob"))
```

The following example demonstrates the [`get`][rdflib.contrib.rdf4j.client.Repository.get] method, which returns a new in-memory [`Dataset`][rdflib.Dataset] object containing all the statements with `https://example.com/Bob` as the subject.

```python
ds = repo.get(subject=URIRef("https://example.com/Bob"))
```

!!! note

    Methods that accept a `graph_name` parameter can restrict the operation to a specific graph. Use `None` to match all graphs, or [`DATASET_DEFAULT_GRAPH_ID`][rdflib.graph.DATASET_DEFAULT_GRAPH_ID] to match only the default graph.

To retrieve a list of graph names in the repository, use the [`graph_names`][rdflib.contrib.rdf4j.client.Repository.graph_names] method. This returns a list of [`IdentifiedNode`][rdflib.term.IdentifiedNode].

```python
repo.graph_names()
```

#### Querying the repository with SPARQL

SPARQL queries can be executed against a repository using the [`query`][rdflib.contrib.rdf4j.client.Repository.query] method.

The return object is a [`Result`][rdflib.query.Result]. You can use [`Result.type`][rdflib.query.Result.type] to determine the type of the result. For example, an ASK query will provide a boolean result in [`Result.askAnswer`][rdflib.query.Result.askAnswer], a DESCRIBE or CONSTRUCT query will provide a [`Graph`][rdflib.Graph] result in [`Result.graph`][rdflib.query.Result.graph], and a SELECT query will provide results in [`Result.bindings`][rdflib.query.Result.bindings].

```python
query = "SELECT * WHERE { ?s ?p ?o } LIMIT 10"
result = repo.query(query)
if result.type == "ASK":
    print(result.askAnswer)
elif result.type in ("CONSTRUCT", "DESCRIBE"):
    print(result.graph)
else:
    # SELECT
    for row in result.bindings:
        print(row["s"])
```

!!! note

    RDF4J supports optional query parameters that further restrict the context of the SPARQL query. To include these query parameters, pass in keyword arguments to the [`query`][rdflib.contrib.rdf4j.client.Repository.query] method. See [RDF4J REST API - Execute SPARQL query](https://rdf4j.org/documentation/reference/rest-api/#tag/SPARQL/paths/~1repositories~1%7BrepositoryID%7D/get) for the list of supported query parameters.

#### Updating the repository with SPARQL Update

To update the repository using SPARQL Update, use the [`update`][rdflib.contrib.rdf4j.client.Repository.update] method.

```python
query = "INSERT DATA { <https://example.com/Bob> <https://example.com/knows> <https://example.com/Alice> }"
repo.update(query)
```

#### Graph Store Manager

RDF4J repositories support the Graph Store Protocol to allow clients to upload and retrieve RDF data. To access the Graph Store Manager, use the [`graphs`][rdflib.contrib.rdf4j.client.Repository.graphs] property.

```python
repo.graphs
```

To add data to a specific graph, use the [`add`][rdflib.contrib.rdf4j.client.GraphStoreManager.add] method.

```python
repo.graphs.add(graph_name="my-graph", data=data)
```

To clear a graph, use the [`clear`][rdflib.contrib.rdf4j.client.GraphStoreManager.clear] method.

!!! warning

    This will delete all existing data in the graph.

!!! note

    RDF4J does not support empty named graphs. Once a graph is cleared, it will be deleted from the repository and will not appear in the list of [`graph_names`][rdflib.contrib.rdf4j.client.Repository.graph_names].

```python
repo.graphs.clear(graph_name="my-graph")
```

To retrieve data from a specific graph, use the [`get`][rdflib.contrib.rdf4j.client.GraphStoreManager.get] method. This returns a new in-memory [`Graph`][rdflib.Graph] object containing the data in the graph.

```python
graph = repo.graphs.get(graph_name="my-graph")
```

To overwrite the data in a specific graph, use the [`overwrite`][rdflib.contrib.rdf4j.client.GraphStoreManager.overwrite] method.

!!! warning

    This will delete all existing data in the graph and replace it with the provided data.

```python
repo.graphs.overwrite(graph_name="my-graph", data=data)
```

### Repository transactions

RDF4J supports transactions, which allow you to group multiple operations into a single atomic unit. To perform operations in a transaction, use the [`transaction`][rdflib.contrib.rdf4j.client.Repository.transaction] method as a context manager.

!!! warning

    Transaction instances are not thread-safe. Do not share a single Transaction instance across multiple threads. Each thread should create its own transaction, or use appropriate synchronization if sharing is required.

The following example demonstrates how to perform operations in a transaction. The transaction will be automatically committed when the block is exited. If an error occurs during the transaction, the transaction will be rolled back and the error will be raised.

```python
with repo.transaction() as txn:
    # Perform your operations here.
    ...

# Exiting the with-statement block will automatically commit or roll back if an error occurs.
```

!!! note

    The methods available on the [`Transaction`][rdflib.contrib.rdf4j.client.Transaction] object vary slightly from the methods available on the [`Repository`][rdflib.contrib.rdf4j.client.Repository] object. This is due to slight variations between the two endpoints (repositories and transactions) in the RDF4J REST API.

For the full reference documentation of the Transaction class, see [`Transaction`][rdflib.contrib.rdf4j.client.Transaction].

### Repository namespace prefixes

RDF4J supports managing namespace prefixes for a repository. To access the namespace manager, use the [`namespaces`][rdflib.contrib.rdf4j.client.Repository.namespaces] property on the [`Repository`][rdflib.contrib.rdf4j.client.Repository] object.

```python
repo.namespaces
```

To set a namespace prefix, use the [`set`][rdflib.contrib.rdf4j.client.RDF4JNamespaceManager.set] method.

```python
repo.namespaces.set("schema", "https://schema.org/")
```

To retrieve a namespace by prefix, use the [`get`][rdflib.contrib.rdf4j.client.RDF4JNamespaceManager.get] method.

```python
namespace = repo.namespaces.get("schema")
```

To remove a namespace prefix, use the [`remove`][rdflib.contrib.rdf4j.client.RDF4JNamespaceManager.remove] method.

```python
repo.namespaces.remove("schema")
```

To list all namespace prefixes, use the [`list`][rdflib.contrib.rdf4j.client.RDF4JNamespaceManager.list] method. This returns a list of [`NamespaceListingResult`][rdflib.contrib.rdf4j.client.NamespaceListingResult]

```python
import dataclasses

namespace_prefixes = repo.namespaces.list()
for prefix, namespace in [dataclasses.astuple(np) for np in namespace_prefixes]:
    print(prefix, namespace)
```

To clear all namespace prefixes, use the [`clear`][rdflib.contrib.rdf4j.client.RDF4JNamespaceManager.clear] method.

!!! warning

    This will remove all namespace prefixes from the repository.

```python
repo.namespaces.clear()
```

# Namespaces and Bindings

RDFLib provides several short-cuts to working with many URIs in the same namespace.

The [`rdflib.namespace`][rdflib.namespace] module defines the [`Namespace`][rdflib.namespace.Namespace] class which lets you easily create URIs in a namespace:

```python
from rdflib import Namespace

EX = Namespace("http://example.org/")
EX.Person  # a Python attribute for EX. This example is equivalent to rdflib.term.URIRef("http://example.org/Person")

# use dict notation for things that are not valid Python identifiers, e.g.:
n['first%20name']  # as rdflib.term.URIRef("http://example.org/first%20name")
```

These two styles of namespace creation - object attribute and dict - are equivalent and are made available just to allow for valid RDF namespaces and URIs that are not valid Python identifiers. This isn't just for syntactic things like spaces, as per the example of `first%20name` above, but also for Python reserved words like `class` or `while`, so for the URI `http://example.org/class`, create it with `EX['class']`, not `EX.class`.

## Common Namespaces

The `namespace` module defines many common namespaces such as RDF, RDFS, OWL, FOAF, SKOS, PROF, etc. The list of the namespaces provided grows with user contributions to RDFLib.

These Namespaces, and any others that users define, can also be associated with prefixes using the [`NamespaceManager`][rdflib.namespace.NamespaceManager], e.g. using `foaf` for `http://xmlns.com/foaf/0.1/`.

Each RDFLib graph has a [`namespace_manager`][rdflib.graph.Graph.namespace_manager] that keeps a list of namespace to prefix mappings. The namespace manager is populated when reading in RDF, and these prefixes are used when serialising RDF, or when parsing SPARQL queries. Prefixes can be bound with the [`bind()`][rdflib.graph.Graph.bind] method:

```python
from rdflib import Graph, Namespace
from rdflib.namespace import FOAF

EX = Namespace("http://example.org/")

g = Graph()
g.bind("foaf", FOAF)  # bind an RDFLib-provided namespace to a prefix
g.bind("ex", EX)      # bind a user-declared namespace to a prefix
```


The [`bind()`][rdflib.graph.Graph.bind] method is actually supplied by the [`NamespaceManager`][rdflib.namespace.NamespaceManager] class - see next.

## NamespaceManager

Each RDFLib graph comes with a [`NamespaceManager`][rdflib.namespace.NamespaceManager] instance in the [`namespace_manager`][rdflib.graph.Graph.namespace_manager] field; you can use the [`bind()`][rdflib.namespace.NamespaceManager.bind] method of this instance to bind a prefix to a namespace URI, as above, however note that the [`NamespaceManager`][rdflib.namespace.NamespaceManager] automatically performs some bindings according to a selected strategy.

Namespace binding strategies are indicated with the `bind_namespaces` input parameter to [`NamespaceManager`][rdflib.namespace.NamespaceManager] instances and may be set via `Graph` also:

```python
from rdflib import Graph
from rdflib.namespace import NamespaceManager

g = Graph(bind_namespaces="rdflib")  # bind via Graph

g2 = Graph()
nm = NamespaceManager(g2, bind_namespaces="rdflib")  # bind via NamespaceManager
```


Valid strategies are:

- core:
  - binds several core RDF prefixes only
  - owl, rdf, rdfs, xsd, xml from the NAMESPACE_PREFIXES_CORE object
  - this is default
- rdflib:
  - binds all the namespaces shipped with RDFLib as DefinedNamespace instances
  - all the core namespaces and all the following: brick, csvw, dc, dcat
  - dcmitype, dcterms, dcam, doap, foaf, geo, odrl, org, prof, prov, qb, sdo
  - sh, skos, sosa, ssn, time, vann, void
  - see the NAMESPACE_PREFIXES_RDFLIB object in [`rdflib.namespace`][rdflib.namespace] for up-to-date list
- none:
  - binds no namespaces to prefixes
  - note this is NOT default behaviour
- cc:
  - using prefix bindings from prefix.cc which is a online prefixes database
  - not implemented yet - this is aspirational

### Re-binding

Note that regardless of the strategy employed, prefixes for namespaces can be overwritten with users preferred prefixes, for example:

```python
from rdflib import Graph
from rdflib.namespace import GEO  # imports GeoSPARQL's namespace

g = Graph(bind_namespaces="rdflib")  # binds GeoSPARQL's namespace to prefix 'geo'

g.bind('geosp', GEO, override=True)
```

[`NamespaceManager`][rdflib.namespace.NamespaceManager] also has a method to normalize a given url:

```python
from rdflib.namespace import NamespaceManager

nm = NamespaceManager(Graph())
nm.normalizeUri(t)
```

For simple output, or simple serialisation, you often want a nice readable representation of a term. All RDFLib terms have a `.n3()` method, which will return a suitable N3 format and into which you can supply a NamespaceManager instance to provide prefixes, i.e. `.n3(namespace_manager=some_nm)`:

```python
>>> from rdflib import Graph, URIRef, Literal, BNode
>>> from rdflib.namespace import FOAF, NamespaceManager

>>> person = URIRef("http://xmlns.com/foaf/0.1/Person")
>>> person.n3()
'<http://xmlns.com/foaf/0.1/Person>'

>>> g = Graph()
>>> g.bind("foaf", FOAF)

>>> person.n3(g.namespace_manager)
'foaf:Person'

>>> l = Literal(2)
>>> l.n3()
'"2"^^<http://www.w3.org/2001/XMLSchema#integer>'

>>> l.n3(NamespaceManager(Graph(), bind_namespaces="core"))
'"2"^^xsd:integer'
```

The namespace manager also has a useful method `compute_qname`. `g.namespace_manager.compute_qname(x)` (or just `g.compute_qname(x)`) which takes a URI and decomposes it into the parts:

```python
self.assertEqual(g.compute_qname(URIRef("http://foo/bar#baz")),
                ("ns2", URIRef("http://foo/bar#"), "baz"))
```

## Namespaces in SPARQL Queries

The `initNs` argument supplied to [`query()`][rdflib.graph.Graph.query] is a dictionary of namespaces to be expanded in the query string. If you pass no `initNs` argument, the namespaces registered with the graphs namespace_manager are used:

```python
from rdflib.namespace import FOAF
graph.query('SELECT * WHERE { ?p a foaf:Person }', initNs={'foaf': FOAF})
```

In order to use an empty prefix (e.g. `?a :knows ?b`), use a `PREFIX` directive with no prefix in the SPARQL query to set a default namespace:

```sparql
PREFIX : <http://xmlns.com/foaf/0.1/>
```

---
title: "A tidyverse lover's intro to RDF"
author: "Carl Boettiger"
date: "`r Sys.Date()`"
output: 
  rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{rdflib Introduction}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
  
---

```{r include=FALSE}
options(rdf_print_format = "nquads")
is_linux <- Sys.info()["sysname"] == "Linux"
knitr::opts_chunk$set(eval=FALSE)
```

In the world of data science, RDF is a bit of an ugly duckling.  Like XML and Java, only without the massive-adoption-that-refuses-to-die part.  In fact RDF is most frequently expressed in XML, and RDF tools are written in Java, which help give RDF has the aesthetics of *steampunk*, of some technology for some futuristic Semantic Web[^1] in a toolset that feels about as lightweight and modern as iron dreadnought.

[^1]: "The semantic web is the future of the internet and always will be." -Peter Norvig, Director of Research at Google

But don't let these appearances deceive you. RDF really is cool. If you've ever gotten carried away using `tidyr::gather` to make everything into one long table, you may have noticed you can just about always get things down to about three columns, as we see with an obligatory `mtcars` data example for `tidyr::gather`:

```{r message = FALSE, warning=FALSE}
library(rdflib)
library(dplyr)
library(tidyr)
library(tibble)
library(jsonld)
```


```{r}
car_triples <- 
mtcars %>% 
  rownames_to_column("Model") %>% 
  gather(attribute,measurement, -Model)
```

```{r echo=FALSE}
DT::datatable(car_triples)
```

If you like long tables like this, RDF is for you. This layout isn't "Tidy Data," where rows are observations and columns are variables, but it is damn useful sometimes.  This format is very liquid, easy to reshape into other structures -- so much so that `tidyr::gather` was originally known as `melt` in the `reshape2` package. It's also a good way to get started thinking about RDF.

## It's all about the triples

Looking at this table closely, we see that *each row is reduced to the most elementary statement you can make from the data*. A row no longer tells you the measurements (observations) *all* attributes (variables) of a given species (key), instead, you get just one fact per row, `Mazda RX4` gets a `mpg` measurement of `21.0`.  In RDF-world, we think of these three-part statements as something very special, which we call **triples**.  RDF is all about these triples. 

The first column came from the row names in this case, the `Model` of car.  This acts serves as a `key` to index the data.frame, i.e. the **subject** being described. The next column is the variable (also called attribute or property) being measured, (that is, column names, other than the key column(s), from the tidy data), called the property or **predicate** in RDF-speak (slash grammar-school jargon). The third column is the actual value measured, more **object** of the predicate. Call it key-property-value or subject-predicate-object, these are our triples.  We can represent just about any data in fully elementary manner.  


&nbsp;          | &nbsp;  | &nbsp;      | &nbsp;
----------------|---------|-------------|-------------
**RDF**         | subject | predicate   | object         
**JSON**        | object  | property    | value
**spreadsheet** | row id  | column name | cell
**data.frame**  | key     | variable    | measurement
**data.frame**  | key     | attribute   | value  

Table: **Table 1**: The many names for triples.  

Table 1 summarizes the many different names associated with triples.  The first naming convention is the terminology typically associated with RDF.  The second set are terms typically associated with JSON data, while the remaining are all examples in tabular or relational data structures.  


## Subject URIs

Using row names as our subject was intuitive but actually a bit sloppy.  `tidyverse` lovers know that `tidyverse` doesn't like rownames, they aren't tidy and have a way of causing trouble.  Of course, we made rownames into a proper column to use `gather`, but we could have taken this one step further.  In true `tidyverse` fashion, this rownames-column is really just one more variable we can observe, one more attribute of the thing we were describing: say, thing A (Car A) is a `car_model_name` as  `Mazda RX4` and thing A also has `mpg` of `21`. We can accomplish such a greater level of abstraction by keeping the Model as just another variable the row ids themselves as the key (i.e. the *subject*) of our triple:

```{r}
car_triples <- 
mtcars %>% 
  rownames_to_column("Model") %>% 
  rowid_to_column("subject") %>% 
  gather(predicate, object, -subject)
```


```{r echo=FALSE}
DT::datatable(car_triples)
```


This is identical to a `gather` of *all* columns, where we have just made the original row ids an explicit column for reference (diligent reader will recognize we would need this information to reverse the operation and `spread` the data back into it's wide form; without it, our transformation is lossy and irreversible).   Our `subject` column now consists only of simple numeric `id`'s, while we have gained an additional triple for every row in the original data which states `Model` of each `id` number (e.g. `1` is `Model` `Mazda RX4`).  Okay, now you're probably thinking: "wait a minute, `1` is not a very unique or specific key, surely that will cause trouble," and you'd be right. For instance, if we performed the same transformation on the iris data, we get triples in the exact same format, ready to `bind_rows`:



```{r}
iris_triples <- iris %>%
  rowid_to_column("subject") %>%
  gather(key = predicate, value = object, -subject)
```

```{r echo=FALSE}
DT::datatable(iris_triples)
```


but in the `iris` data, `1` corresponds to the first individual Iris flower in the measurement data, and not a Mazda RX4.  If we don't want to get confused, we're going to need to make sure our identifiers are unique: not just kind of unique, but unique in the **World** wide.  And what else is unique world-wide? Yup, you guessed it, we are going to use URLs for our subject identifiers, just like the world wide web.  Think of this as a clever out-sourcing to the whole internet domain registry service.  Here, we'll imagine registering each of these example datasets with a separate **base URL**, so instead of a vague `1` to identify the first observation in the `iris` example data, we'll use the URL `http://example.com/iris#1`, which we can now distinguish from `http://example.com/mtcars#1` (and if you're way ahead of me, yes, we'll have more to say about URI vs URL and the use of blank nodes in just a minute).  For example:

```{r}
iris_triples <- iris %>%
  rowid_to_column("subject") %>%
  mutate(subject = paste0("http://example.com/", "iris#", subject)) %>%
  gather(key = predicate, value = object, -subject)

```


```{r echo=FALSE}
DT::datatable(iris_triples)
```

## Predicate URIs

A slightly more subtle version of the same problem can arise with our predicates. Different tables may use the same attribute (i.e. originally, a column name of a variable) for different things -- the attribute labeled `cyl` means "number of cylinders" in `mtcars` data.frame, but could mean something very different in different data.  Luckily we've already seen how to make names unique in RDF turn them into URLs.

```{r mesage = FALSE}
iris_triples <- iris %>%
  rowid_to_column("subject") %>%
  mutate(subject = paste0("http://example.com/", "iris#", subject)) %>%
  gather(key = predicate, value = object, -subject) %>%
  mutate(predicate = paste0("http://example.com/", "iris#", predicate))

```



```{r echo=FALSE}
DT::datatable(iris_triples)
```


At this point the motivation for the name "Linked Data" is probably becoming painfully obvious.  


## Datatype URIs

One more column to go!  But wait a minute, the `object` column is different, isn't it? These measurements don't suffer from the same ambiguity -- after all, there is no confusion if a car has `4` cylinders and an iris has `4` mm long sepals.  However, a new issue has arisen in the data type (e.g. `string`, `boolean`, `double`, `integer`, `dateTime`, etc).  A close look reveals our `object` column is encoded as a `character` and not `numeric` -- how'd that happen?  `tidyr::gather` has coerced the whole column into character strings because some of the values, that is, the `Species` names in `iris` and the Model names in `mtcars`, are text strings (and it couldn't exactly coerce them into integers).  Perhaps this isn't a big deal -- we can often guess the type of an object just by how it looks (so-called [Duck typing](https://en.wikipedia.org/wiki/Duck_typing), because if it quacks like duck...).  Still, being explicit about data types is a Good Thing, so fortunately there's an explicit way to address this too ... oh no ... not ... yes ... more URLs!  

Luckily we don't have to make up `example.com` URLs this time because there's already a well-established list of data types widely used across the internet that were originally developed for use in XML (I warned you) Schemas, listed in see the [W3C RDF DataTypes](https://www.w3.org/TR/rdf11-concepts/#section-Datatypes).  As the standard shows, familiar types `string`, `double`, `boolean`, `integer`, etc are made explicit using the XML Schema URL: `http://www.w3.org/2001/XMLSchema#`, followed by the type; so an integer would be ``http://www.w3.org/2001/XMLSchema#integer`, a character string `http://www.w3.org/2001/XMLSchema#string` etc.  

Because this case is a little different, the URL is attached directly after the object value, which is set off by quotes, using the symbol `^^` (I dunno, but I think two duck feet), such that `5.1` becomes `"5.1"^^http://www.w3.org/2001/XMLSchema#double`.  Wow[^2].  Most of the time we won't have to worry about the type, because, if it quacks... 

[^2]: Couldn't we just have used another column?  Perhaps, but then it wouldn't be a triple.  More to the point, the datatype modifies `object` alone, not the predicate or subject.  


# Triples in `rdflib`

So far, we have explored the concept of triples using familiar `data.frame` structures, but haven't yet introduced any `rdflib` functions.  Though we've been thinking of RDF data in this explicitly tabular three-column structure, that is really just one potentially convenient representation. Just as the same tabular data can be represented in a `data.frame`, written to disk as a `.csv` file, or stored in a database (like MySQL or PostgreSQL), so it is for RDF to even greater degree.  We have separate abstractions for the information itself compared to how it is represented.  

To take advantage of this abstraction, `rdflib` introduces an `rdf` class object. Depending on how this is initialized, this could utilize storage in memory (the default), on disk, or potentially in an array of different databases, (including relational databases like PostgreSQL and rdf-specific ones like Virtuoso, depending on how the underlying `redland` library is compiled -- a topic beyond our scope here).  Here, we simply initialize an `rdf` object using the default in-memory storage:

```{r}
rdf <- rdf()
```

To add triples to this `rdf` object (often called an RDF Model or RDF Graph), we use the function `rdf_add`, which takes a subject, predicate, and object as arguments, as we have just discussed.  A datatype URI can be inferred from the R type used for the object (e.g. `numeric`, `integer`, `logical`, `character`, etc.)


```{r}
base <- paste0("http://example.com/", "iris#")

rdf %>% 
  rdf_add(subject = paste0(base, "obs1"), 
          predicate = paste0(base, "Sepal.Length"), 
          object = 5.1)

rdf
```

The result is displayed as a triple discussed above. This is technically an example of the `nquad` notation we will see later.  Note the inferred datatype URI.


## Dialing back the ugly

This `gather` thing started well, but now are data is looking pretty ugly, not to mention cumbersome.  You have some idea why RDF hasn't taken data science by storm, and we haven't even looked at how ugly this gets when you write it in the RDF/XML serialization yet!  On the upside, we've introduced most of the essential concepts that will let us start to work with data as triples.  Before we proceed further, we'll take a quick look at some of the options for expressing triples in different ways, and also introduce some of the different serializations (ways of representing in text) frequently used to express these triples.  

#### Prefixes for URIs

Long URL strings are one of the most obvious ways that what started off looking like a concise, minimal statement got ugly and cumbersome.  Borrowing from the notion of [Namespaces in XML](https://en.wikipedia.org/wiki/XML_namespace), most RDF tools permit custom prefixes to be declared and swapped in for longer URLs.  A prefix is typically a short string[^3] followed by a `:` that is used in place of the shared root URL. For instance, we might use the prefix `iris:Sepal.Length` and `iris:Sepal.Width` where `iris:` is defined to mean `http://example.com/iris#` in our example above.   

 [^3]: Technically I believe it should be a [NCName](https://www.w3.org/TR/xmlschema-2/#NCName), defined by the regexp `[\i-[:]][\c-[:]]*`.  [Essentially](https://stackoverflow.com/questions/1631396), this says it cannot include symbol characters like `:`, `@`, `$`, `%`, `&`, `/`, `+`, `,`, `;`, whitespace characters or different parenthesis. Furthermore an NCName cannot begin with a number, dot or minus character although they can appear later in an NCName.  
 
#### URI vs URL

While I've referred to these things as [URL](https://en.wikipedia.org/wiki/URL)s, (uniform resource locator, aka web address) technically they can be a broader class of things known as [URI](https://en.wikipedia.org/wiki/Uniform_Resource_Identifier)s (uniform resource identifier).  In addition to including anything that is a URL, URIs include things which are not URLs, like `urn:isbn:0-486-27557-4` or `urn:uuid:aac06f69-7ec8-403d-ad84-baa549133dce`, which are URNs: unique resource numbers in some numbering scheme (e.g. book ISBN numbers, or UUIDs), neither of which are URLs but nonetheless enjoy the same globally unique property.

#### Blank nodes

Sometimes we do not need a globally unique identifier, we just want a way to refer to a node (e.g. subject, and sometimes an object) uniquely in our document. This is the role of a [blank node](https://en.wikipedia.org/wiki/Blank_node) (do follow the link for a better overview).  These are frequently denoted with the prefix `_:`, e.g. we could have replaced the sample IDs as `_:1`, `_:2` instead of the URLs such as `http://example.com/iris#1` etc.  Note that RDF operations need not preserve the actual string pattern in a blank ID name, it means the exact same thing if we replace all the `_:1`s with `_:b1` and `_:2` with `_:b2`,  etc.

In `librdf` we can get a blank node by passing an empty string or character string that is not a URI as the subject.  Here we also use a URI that isn't a URL as predicate: 

```{r}
rdf <- rdf()
rdf %>% rdf_add("",   
                "iris:Sepal.Length", 
                object = 5.1)
rdf
```

Note that we get a blank node, `_:`  with a randomly generated string.  

## Triple notation: `nquads` `rdfxml`, `turtle`, and `nquads`

So far we have relied primarily on a three-column tabular format to represent our triples.  We have also seen the default `print` format used for the `rdf` method, known as [N-Quads](https://www.w3.org/TR/n-quads/) above, which displays a bare, space-separated triple, possibly with a datatype URI attached to the object.  The line ends with a dot, which indicates this is part of the same local triplestore (aka RDF graph or RDF Model).  Technically this could be another URI indicating a unique global address for the triplestore in question.   

We can serialize any `rdf` object out to a file in this format with the `rdf_serialize()` function, e.g.

```{r}
rdf_serialize(rdf, "rdf.nq", format = "nquads")
```

 Just as each of these formats can be serialized with `rdf_serialize()`, each can be read by `rdflib` using the function `rdf_parse()`:


```{r}
doc <- system.file("extdata/example.rdf", package="redland")
rdf <- rdf_parse(doc, format = "rdfxml") 
rdf
```


N-Quads are convenient in that each triple is displayed on a unique line, and the format supports the blank node and Datatype URIs in the manner we have just discussed. Other formats are not so concise.  Rather than print to file, we can simply change the default print format used by `rdflib` to explore the textual layout of the other serializations.  Here is one of the most common classical serializations, `RDF/XML` which expresses triples in an XML-based schema:

```{r}
options(rdf_print_format = "rdfxml")
rdf
```

Just looking at this is probably enough to explain why so many alternative serializations were created.  Another popular format, [`turtle`](https://www.w3.org/TR/turtle/), looks more like `nquads`: 


```{r}
options(rdf_print_format = "turtle")
rdf
```

Here, blank nodes are denoted by `[]`.  `turtle` uses indentation to indicate that all three predicates (`creator`, `description`, `title`) are properties of the same subject.  



### JSON-LD

While formats such as `nquads` and `turtle` provide a much cleaner syntax than RDF/XML, they also introduce a custom format rather than building on a familiar standard (like XML) for which users already have a well-developed set of tools and intuition.  After more than a decade of such challenges (RDF specification started 1997, including an the HTML-embedded serialization of [RDFa](https://en.wikipedia.org/wiki/RDFa) in 2004), a more user friendly specification has emerged in the form of JSON-LD (1.0 W3C specification was released in 2014, the 1.1 specification released in February 2018).  JSON-LD uses the familiar *object notation* of JSON, (which is rapidly replacing XML as the ubiquitous data exchange format, and will be more familiar to many readers than the specialized RDF formats or even XML.  Here is our `rdf` data in the JSON-LD serialization:


```{r}
options(rdf_print_format = "jsonld")
rdf
```

In this serialization, our subject corresponds to "the thing in the curly braces," (i.e. the JSON "object") which is identified by the special `@id` property (omitting `@id` corresponds to a blank node).  The predicate-object pairs in the triple are then just JSON key-value pairs within the curly braces of the given object.  We can make this format look even more natural by stripping out the URLs.  While it is possible to use prefixes in place of URLs, it is more natural to pull them out entirely, e.g. by declaring a default vocabulary in the JSON-LD "Context", like so:

```{r}
rdf_serialize(rdf, "example.json", "jsonld") %>% 
  jsonld_compact(context = '{"@vocab": "http://purl.org/dc/elements/1.1/"}')
```

The context of a JSON-LD file can also define datatypes, use multiple namespaces, and permit different names in the JSON keys from that found in the URLs.  While a complete introduction to JSON-LD is beyond our scope, this representation essentially provides a way to map intuitive JSON structures into precise RDF triples. 



## From tables to Graphs

So far we have considered examples where the data could be represented in tabular form. 
We frequently encounter data that cannot be easily represented in such a format.  For instance, consider
the JSON data in this example:

```{r}
ex <- system.file("extdata/person.json", package="rdflib")
cat(readLines(ex), sep = "\n")
#jsonld_compact(ex, "{}")

```


This JSON object for a `Person` has another JSON object nested inside (a `PostalAddress`).  Yet if we look at this data as `nquads`, we see the familiar flat triple structure: 



```{r}
options(rdf_print_format = "nquads")
rdf <- rdf_parse(ex, "jsonld")
rdf
```

So what has happened?  Note that our `address` has been given the blank node URI `_:b0`, which serves both as the object in the `address` line of the `Person` and as the subject of all the properties belonging to the `PostalAddress`.  In JSON-LD, this structure is referred to as being 'flattened': 

```{r}
jsonld_flatten(ex, context = "https://schema.org/")
```

Note that our JSON-LD structure now starts with an object called `@graph`.  Unlike our opening examples, this data is not tabular in nature, but rather, is formatted as a nested _graph_.  Such nesting is very natural in JSON, where objects can be arranged in a tree-like structure with a single outer-most set of `{}` indicating a root object.  A graph is just a more generic form of a tree structure, where we are agnostic to the root.  (We could in fact use the `@reverse` property on address to create a root `PostalAddress` that contains the `Person`).  In this way, the notion of data as a `graph` offers a powerful generalization to the notion of tabular data.  The `@graph` above consists of two separate objects: a `PostalAddress` (with `id` of `_:b0`) and a `Person` (with an ORCID id).  This layout acts much like a foreign key in a relational database, or as a list-column in `tidyverse` (e.g. see `tidyr::nest()`).  `rdflib` uses this flattened representation when serializing JSON-LD objects.  Note that JSON-LD provides a rich set of utilities to go back and forth between flattened and nested layouts using `jsonld_frame`.  For instance, we can recover the original structure just by specifying a frame that indicates which type we want as the root:

```{r}
jsonld_flatten(ex) %>%
  jsonld_frame('{"@type": "https://schema.org//Person"}') %>%
  jsonld_compact(context = "https://schema.org/")
```

(Recall that compacting just replaces URIs and any type declarations with short names given by the context).  This is somewhat analogous to `join` operations in relational data, or nesting and un-nesting functions in `tidyr`.  However, when working with RDF, the beautiful thing is that the differences between these two representations (nested or flattened) are purely aesthetic.  Both representations have precisely the same semantic meaning, and are thus precisely the same thing in RDF world.  We will never have to orchestrate a join on a foreign key before we can perform desired operations like select and filter on the data.  We don't have to think about how our data is organized, because it is always in the same molten triple format, whatever it is, and however nested it might be.  

Just as we saw `gather` could provide a relatively generic way of transforming a data.frame into RDF triples, JSON-LD defines a relatively simple convention for getting nested data (e.g. lists) into RDF triples.  This convention simply treats JSON `{}` objects as `subjects` (often assigning blank node ids, as we saw with row ids), and key-value pairs (or in R-speak, list names and values) as predicates and objects, respectively.  Any raw JSON file can be treated as JSON-LD, ideally by specifying an appropriate `context`, which serves to map terms into URIs as we saw with data.frames. `JSON-LD` is then already a valid RDF format that we can parse with `rdflib`.  

For instance, here is a simple function for coercing list objects into RDF with a specified context:

```{r}
as_rdf.list <- function(x, context = "https://schema.org/"){
  if(length(x) == 1) x <- x[[1]]
  x[["@context"]] <- context
  json <- jsonlite::toJSON(x, pretty = TRUE, auto_unbox = TRUE, force = TRUE)
  rdflib::rdf_parse(json, "jsonld")
}
```

Here we set a default context (https://schema.org/), and map a few R terms to corresponding schema terms

```{r}
context <- list("https://schema.org/", 
                list(schema = "https://schema.org//",
                     given = "givenName",
                     family = "familyName",
                     title = "name",
                     year = "datePublished",
                     note = "softwareVersion",
                     comment = "identifier",
                     role = "https://www.loc.gov/marc/relators/relaterm.html"))

```


We can now apply our function on arbitrary R `list` objects, such as the `bibentry` class object returned by the `citation()` function:

```{r}
options(rdf_print_format = "nquads") # go back to the default


R <- citation("rdflib")
rdf <- as_rdf.list(R, context)
rdf  
```


## SPARQL: A Graph Query Language

So far, we have spent a lot of words describing how to transform data into RDF, and not much actually _doing anything_ cool with said data. 


_Still working on writing this section_

```{r}
#source(system.file("examples/as_rdf.R", package="rdflib"))
source(system.file("examples/tidy_schema.R", package="rdflib"))

## Testing: Digest some data.frames into RDF and extract back
 cars <- mtcars %>% rownames_to_column("Model")
 x1 <- as_rdf(iris, NULL, "iris:")
 x2 <- as_rdf(cars, NULL, "mtcars:")
 rdf <- c(x1,x2)
```


## SPARQL: Getting back to Tidy Tables!


```{r}
sparql <-
  'SELECT  ?Species ?Sepal_Length ?Sepal_Width ?Petal_Length  ?Petal_Width
WHERE {
 ?s <iris:Species>  ?Species .
 ?s <iris:Sepal.Width>  ?Sepal_Width .
 ?s <iris:Sepal.Length>  ?Sepal_Length . 
 ?s <iris:Petal.Length>  ?Petal_Length .
 ?s <iris:Petal.Width>  ?Petal_Width 
}'

iris2 <- rdf_query(rdf, sparql)
```

```{r echo=FALSE}
DT::datatable(iris2)
```


We can automatically create the a SPARQL query that returns "tidy data".  Tidy data has predicates as columns, objects as values, subjects as rows.  

```{r}
sparql <- tidy_schema("Species",  "Sepal.Length", "Sepal.Width", prefix = "iris")

rdf_query(rdf, sparql)
```


```{r include=FALSE}
unlink("rdf.nq")
unlink("example.json")
```

# RDFLib

RDFLib is a pure Python package for working with [RDF](http://www.w3.org/RDF/). It contains:

* **Parsers & Serializers**
    * for RDF/XML, N3, NTriples, N-Quads, Turtle, TriG, TriX, JSON-LD, HexTuples, RDFa and Microdata

* **Store implementations**
    * memory stores
    * persistent, on-disk stores, using databases such as BerkeleyDB
    * remote SPARQL endpoints

* **Graph interface**
    * to a single graph
    * or to multiple Named Graphs within a dataset

* **SPARQL 1.1 implementation**
    * both Queries and Updates are supported

!!! warning "Security considerations"
    RDFLib is designed to access arbitrary network and file resources, in some
    cases these are directly requested resources, in other cases they are
    indirectly referenced resources.

    If you are using RDFLib to process untrusted documents or queries you should
    take measures to restrict file and network access.

    For information on available security measures, see the RDFLib
    [Security Considerations](security_considerations.md)
    documentation.

## Getting started

If you have never used RDFLib, the following will help get you started:

* [Getting Started](gettingstarted.md)
* [Introduction to Parsing](intro_to_parsing.md)
* [Introduction to Creating RDF](intro_to_creating_rdf.md)
* [Introduction to Graphs](intro_to_graphs.md)
* [Introduction to SPARQL](intro_to_sparql.md)
* [Utilities](utilities.md)
* [Examples](apidocs/examples.md)

## In depth

If you are familiar with RDF and are looking for details on how RDFLib handles it, these are for you:

* [RDF Terms](rdf_terms.md)
* [Namespaces and Bindings](namespaces_and_bindings.md)
* [Persistence](persistence.md)
* [Merging](merging.md)
* [Changelog](changelog.md)
* [Upgrade 6 to 7](upgrade6to7.md)
* [Upgrade 5 to 6](upgrade5to6.md)
* [Upgrade 4 to 5](upgrade4to5.md)
* [Security Considerations](security_considerations.md)

## Versioning

RDFLib follows [Semantic Versioning 2.0.0](https://semver.org/spec/v2.0.0.html), which can be summarized as follows:

Given a version number `MAJOR.MINOR.PATCH`, increment the:

1. `MAJOR` version when you make incompatible API changes
2. `MINOR` version when you add functionality in a backwards-compatible manner
3. `PATCH` version when you make backwards-compatible bug fixes

## For developers

* [Developers guide](developers.md)
* [Documentation guide](docs.md)
* [Contributing guide](CONTRIBUTING.md)
* [Code of Conduct](CODE_OF_CONDUCT.md)
* [Persisting N3 Terms](persisting_n3_terms.md)
* [Type Hints](type_hints.md)
* [Decisions](decisions.md)

## Source Code

The rdflib source code is hosted on GitHub at [https://github.com/RDFLib/rdflib](https://github.com/RDFLib/rdflib) where you can lodge Issues and create Pull Requests to help improve this community project!

The RDFlib organisation on GitHub at [https://github.com/RDFLib](https://github.com/RDFLib) maintains this package and a number of other RDF and RDFlib-related packaged that you might also find useful.

## Further help & Contact

If you would like help with using RDFlib, rather than developing it, please post a question on StackOverflow using the tag `[rdflib]`. A list of existing `[rdflib]` tagged questions can be found [here](https://stackoverflow.com/questions/tagged/rdflib).

You might also like to join RDFlib's [dev mailing list](https://groups.google.com/group/rdflib-dev) or use RDFLib's [GitHub discussions section](https://github.com/RDFLib/rdflib/discussions).

The chat is available at [gitter](https://gitter.im/RDFLib/rdflib) or via matrix [#RDFLib_rdflib:gitter.im](https://matrix.to/#/#RDFLib_rdflib:gitter.im).

# Security Considerations

RDFLib is designed to access arbitrary network and file resources, in some cases these are directly requested resources, in other cases they are indirectly referenced resources.

An example of where indirect resources are accessed is JSON-LD processing, where network or file resources referenced by `@context` values will be loaded and processed.

RDFLib also supports SPARQL, which has federated query capabilities that allow
queries to query arbitrary remote endpoints.

If you are using RDFLib to process untrusted documents or queries, you should
take measures to restrict file and network access.

Some measures that can be taken to restrict file and network access are:

* [Operating System Security Measures](#operating-system-security-measures)
* [Python Runtime Audit Hooks](#python-runtime-audit-hooks)
* [Custom URL Openers](#custom-url-openers)

Of these, operating system security measures are recommended. The other measures work, but they are not as effective as operating system security measures, and even if they are used, they should be used in conjunction with operating system security measures.

## Operating System Security Measures

Most operating systems provide functionality that can be used to restrict network and file access of a process.

Some examples of these include:

* [Open Container Initiative (OCI) Containers](https://www.opencontainers.org/) (aka Docker containers).

  Most OCI runtimes provide mechanisms to restrict network and file
  access of containers. For example, using Docker, you can limit your
  container to only being able to access files explicitly mapped into
  the container and only access the network through a firewall. For more
  information, refer to the documentation of the tool you use to manage
  your OCI containers:

  * [Kubernetes](https://kubernetes.io/docs/home/)
  * [Docker](https://docs.docker.com/)
  * [Podman](https://podman.io/)

* [firejail](https://firejail.wordpress.com/) can be used to
  sandbox a process on Linux and restrict its network and file access.

* File and network access restrictions.

  Most operating systems provide a way to restrict operating system users to
  only being able to access files and network resources that are explicitly
  allowed. Applications that process untrusted input could be run as a user with
  these restrictions in place.

Many other measures are available, however, listing them is outside the scope of this document.

Of the listed measures, OCI containers are recommended. In most cases, OCI containers are constrained by default and can't access the loopback interface and can only access files that are explicitly mapped into the container.

## Python Runtime Audit Hooks

From Python 3.8 onwards, Python provides a mechanism to install runtime audit hooks that can be used to limit access to files and network resources.

The runtime audit hook system is described in more detail in [PEP 578 – Python Runtime Audit Hooks](https://peps.python.org/pep-0578/).

Runtime audit hooks can be installed using the [sys.addaudithook](https://docs.python.org/3/library/sys.html#sys.addaudithook) function, and will then get called when audit events occur. The audit events raised by the Python runtime and standard library are described in Python's [audit events table](https://docs.python.org/3/library/audit_events.html).

RDFLib uses `urllib.request.urlopen` for HTTP, HTTPS and other network access, and this function raises a `urllib.Request` audit event. For file access, RDFLib uses `open`, which raises an `open` audit event.

Users of RDFLib can install audit hooks that react to these audit events and raise an exception when an attempt is made to access files or network resources that are not explicitly allowed.

RDFLib's test suite includes tests which verify that audit hooks can block access to network and file resources.

RDFLib also includes an example that shows how runtime audit hooks can be used to restrict network and file access in [`secure_with_audit`][examples.secure_with_audit].

## Custom URL Openers

RDFLib uses the `urllib.request.urlopen` for HTTP, HTTPS and other network access. This function will use a `urllib.request.OpenerDirector` installed with `urllib.request.install_opener` to open the URLs.

Users of RDFLib can install a custom URL opener that raises an exception when an attempt is made to access network resources that are not explicitly allowed.

RDFLib's test suite includes tests which verify that custom URL openers can be used to block access to network resources.

RDFLib also includes an example that shows how a custom opener can be used to restrict network access in [`secure_with_urlopen`][examples.secure_with_urlopen].

---
title: "A tidyverse lover's intro to RDF"
author: "Carl Boettiger"
date: "`r Sys.Date()`"
output: 
  rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{rdflib Introduction}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
  
---

```{r include=FALSE}
options(rdf_print_format = "nquads")
```

In the world of data science, RDF is a bit of an ugly duckling.  Like XML and Java, only without the massive-adoption-that-refuses-to-die part.  In fact RDF is most frequently expressed in XML, and RDF tools are written in Java, which help give RDF has the aesthetics of *steampunk*, of some technology for some futuristic Semantic Web[^1] in a toolset that feels about as lightweight and modern as iron dreadnought.

[^1]: "The semantic web is the future of the internet and always will be." -Peter Norvig, Director of Research at Google

But don't let these appearances deceive you. RDF really is cool. If you've ever gotten carried away using `tidyr::gather` to make everything into one long table, you may have noticed you can just about always get things down to about three columns, as we see with an obligatory `mtcars` data example for `tidyr::gather`:

```{r message = FALSE, warning=FALSE}
library(rdflib)
library(dplyr)
library(tidyr)
library(tibble)
library(jsonld)
```


```{r}
car_triples <- 
mtcars %>% 
  rownames_to_column("Model") %>% 
  gather(attribute,measurement, -Model)
```

```{r echo=FALSE}
DT::datatable(car_triples)
```

If you like long tables like this, RDF is for you. This layout isn't "Tidy Data," where rows are observations and columns are variables, but it is damn useful sometimes.  This format is very liquid, easy to reshape into other structures -- so much so that `tidyr::gather` was originally known as `melt` in the `reshape2` package. It's also a good way to get started thinking about RDF.

## It's all about the triples

Looking at this table closely, we see that *each row is reduced to the most elementary statement you can make from the data*. A row no longer tells you the measurements (observations) *all* attributes (variables) of a given species (key), instead, you get just one fact per row, `Mazda RX4` gets a `mpg` measurement of `21.0`.  In RDF-world, we think of these three-part statements as something very special, which we call **triples**.  RDF is all about these triples. 

The first column came from the row names in this case, the `Model` of car.  This acts serves as a `key` to index the data.frame, i.e. the **subject** being described. The next column is the variable (also called attribute or property) being measured, (that is, column names, other than the key column(s), from the tidy data), called the property or **predicate** in RDF-speak (slash grammar-school jargon). The third column is the actual value measured, more **object** of the predicate. Call it key-property-value or subject-predicate-object, these are our triples.  We can represent just about any data in fully elementary manner.  


&nbsp;          | &nbsp;  | &nbsp;      | &nbsp;
----------------|---------|-------------|-------------
**RDF**         | subject | predicate   | object         
**JSON**        | object  | property    | value
**spreadsheet** | row id  | column name | cell
**data.frame**  | key     | variable    | measurement
**data.frame**  | key     | attribute   | value  

Table: **Table 1**: The many names for triples.  

Table 1 summarizes the many different names associated with triples.  The first naming convention is the terminology typically associated with RDF.  The second set are terms typically associated with JSON data, while the remaining are all examples in tabular or relational data structures.  


## Subject URIs

Using row names as our subject was intuitive but actually a bit sloppy.  `tidyverse` lovers know that `tidyverse` doesn't like rownames, they aren't tidy and have a way of causing trouble.  Of course, we made rownames into a proper column to use `gather`, but we could have taken this one step further.  In true `tidyverse` fashion, this rownames-column is really just one more variable we can observe, one more attribute of the thing we were describing: say, thing A (Car A) is a `car_model_name` as  `Mazda RX4` and thing A also has `mpg` of `21`. We can accomplish such a greater level of abstraction by keeping the Model as just another variable the row ids themselves as the key (i.e. the *subject*) of our triple:

```{r}
car_triples <- 
mtcars %>% 
  rownames_to_column("Model") %>% 
  rowid_to_column("subject") %>% 
  gather(predicate, object, -subject)
```


```{r echo=FALSE}
DT::datatable(car_triples)
```


This is identical to a `gather` of *all* columns, where we have just made the original row ids an explicit column for reference (diligent reader will recognize we would need this information to reverse the operation and `spread` the data back into it's wide form; without it, our transformation is lossy and irreversible).   Our `subject` column now consists only of simple numeric `id`'s, while we have gained an additional triple for every row in the original data which states `Model` of each `id` number (e.g. `1` is `Model` `Mazda RX4`).  Okay, now you're probably thinking: "wait a minute, `1` is not a very unique or specific key, surely that will cause trouble," and you'd be right. For instance, if we performed the same transformation on the iris data, we get triples in the exact same format, ready to `bind_rows`:



```{r}
iris_triples <- iris %>%
  rowid_to_column("subject") %>%
  gather(key = predicate, value = object, -subject)
```

```{r echo=FALSE}
DT::datatable(iris_triples)
```


but in the `iris` data, `1` corresponds to the first individual Iris flower in the measurement data, and not a Mazda RX4.  If we don't want to get confused, we're going to need to make sure our identifiers are unique: not just kind of unique, but unique in the **World** wide.  And what else is unique world-wide? Yup, you guessed it, we are going to use URLs for our subject identifiers, just like the world wide web.  Think of this as a clever out-sourcing to the whole internet domain registry service.  Here, we'll imagine registering each of these example datasets with a separate **base URL**, so instead of a vague `1` to identify the first observation in the `iris` example data, we'll use the URL `http://example.com/iris#1`, which we can now distinguish from `http://example.com/mtcars#1` (and if you're way ahead of me, yes, we'll have more to say about URI vs URL and the use of blank nodes in just a minute).  For example:

```{r}
iris_triples <- iris %>%
  rowid_to_column("subject") %>%
  mutate(subject = paste0("http://example.com/iris#", subject)) %>%
  gather(key = predicate, value = object, -subject)

```


```{r echo=FALSE}
DT::datatable(iris_triples)
```

## Predicate URIs

A slightly more subtle version of the same problem can arise with our predicates. Different tables may use the same attribute (i.e. originally, a column name of a variable) for different things -- the attribute labeled `cyl` means "number of cylinders" in `mtcars` data.frame, but could mean something very different in different data.  Luckily we've already seen how to make names unique in RDF turn them into URLs.

```{r mesage = FALSE}
iris_triples <- iris %>%
  rowid_to_column("subject") %>%
  mutate(subject = paste0("http://example.com/iris#", subject)) %>%
  gather(key = predicate, value = object, -subject) %>%
  mutate(predicate = paste0("http://example.com/iris#", predicate))

```



```{r echo=FALSE}
DT::datatable(iris_triples)
```


At this point the motivation for the name "Linked Data" is probably becoming painfully obvious.  


## Datatype URIs

One more column to go!  But wait a minute, the `object` column is different, isn't it? These measurements don't suffer from the same ambiguity -- after all, there is no confusion if a car has `4` cylinders and an iris has `4` mm long sepals.  However, a new issue has arisen in the data type (e.g. `string`, `boolean`, `double`, `integer`, `dateTime`, etc).  A close look reveals our `object` column is encoded as a `character` and not `numeric` -- how'd that happen?  `tidyr::gather` has coerced the whole column into character strings because some of the values, that is, the `Species` names in `iris` and the Model names in `mtcars`, are text strings (and it couldn't exactly coerce them into integers).  Perhaps this isn't a big deal -- we can often guess the type of an object just by how it looks (so-called [Duck typing](https://en.wikipedia.org/wiki/Duck_typing), because if it quacks like duck...).  Still, being explicit about data types is a Good Thing, so fortunately there's an explicit way to address this too ... oh no ... not ... yes ... more URLs!  

Luckily we don't have to make up `example.com` URLs this time because there's already a well-established list of data types widely used across the internet that were originally developed for use in XML (I warned you) Schemas, listed in see the [W3C RDF DataTypes](https://www.w3.org/TR/rdf11-concepts/#section-Datatypes).  As the standard shows, familiar types `string`, `double`, `boolean`, `integer`, etc are made explicit using the XML Schema URL: `http://www.w3.org/2001/XMLSchema#`, followed by the type; so an integer would be ``http://www.w3.org/2001/XMLSchema#integer`, a character string `http://www.w3.org/2001/XMLSchema#string` etc.  

Because this case is a little different, the URL is attached directly after the object value, which is set off by quotes, using the symbol `^^` (I dunno, but I think two duck feet), such that `5.1` becomes `"5.1"^^http://www.w3.org/2001/XMLSchema#double`.  Wow[^2].  Most of the time we won't have to worry about the type, because, if it quacks... 

[^2]: Couldn't we just have used another column?  Perhaps, but then it wouldn't be a triple.  More to the point, the datatype modifies `object` alone, not the predicate or subject.  


# Triples in `rdflib`

So far, we have explored the concept of triples using familiar `data.frame` structures, but haven't yet introduced any `rdflib` functions.  Though we've been thinking of RDF data in this explicitly tabular three-column structure, that is really just one potentially convenient representation. Just as the same tabular data can be represented in a `data.frame`, written to disk as a `.csv` file, or stored in a database (like MySQL or PostgreSQL), so it is for RDF to even greater degree.  We have separate abstractions for the information itself compared to how it is represented.  

To take advantage of this abstraction, `rdflib` introduces an `rdf` class object. Depending on how this is initialized, this could utilize storage in memory (the default), on disk, or potentially in an array of different databases, (including relational databases like PostgreSQL and rdf-specific ones like Virtuoso, depending on how the underlying `redland` library is compiled -- a topic beyond our scope here).  Here, we simply initialize an `rdf` object using the default in-memory storage:

```{r}
rdf <- rdf()
```

To add triples to this `rdf` object (often called an RDF Model or RDF Graph), we use the function `rdf_add`, which takes a subject, predicate, and object as arguments, as we have just discussed.  A datatype URI can be inferred from the R type used for the object (e.g. `numeric`, `integer`, `logical`, `character`, etc.)


```{r}
base <- "http://example.com/iris#"

rdf %>% 
  rdf_add(subject = paste0(base, "obs1"), 
          predicate = paste0(base, "Sepal.Length"), 
          object = 5.1)

rdf
```

The result is displayed as a triple discussed above. This is technically an example of the `nquad` notation we will see later.  Note the inferred datatype URI.


## Dialing back the ugly

This `gather` thing started well, but now are data is looking pretty ugly, not to mention cumbersome.  You have some idea why RDF hasn't taken data science by storm, and we haven't even looked at how ugly this gets when you write it in the RDF/XML serialization yet!  On the upside, we've introduced most of the essential concepts that will let us start to work with data as triples.  Before we proceed further, we'll take a quick look at some of the options for expressing triples in different ways, and also introduce some of the different serializations (ways of representing in text) frequently used to express these triples.  

#### Prefixes for URIs

Long URL strings are one of the most obvious ways that what started off looking like a concise, minimal statement got ugly and cumbersome.  Borrowing from the notion of [Namespaces in XML](https://en.wikipedia.org/wiki/XML_namespace), most RDF tools permit custom prefixes to be declared and swapped in for longer URLs.  A prefix is typically a short string[^3] followed by a `:` that is used in place of the shared root URL. For instance, we might use the prefix `iris:Sepal.Length` and `iris:Sepal.Width` where `iris:` is defined to mean `http://example.com/iris#` in our example above.   

 [^3]: Technically I believe it should be a [NCName](https://www.w3.org/TR/xmlschema-2/#NCName), defined by the regexp `[\i-[:]][\c-[:]]*`.  [Essentially](https://stackoverflow.com/questions/1631396), this says it cannot include symbol characters like `:`, `@`, `$`, `%`, `&`, `/`, `+`, `,`, `;`, whitespace characters or different parenthesis. Furthermore an NCName cannot begin with a number, dot or minus character although they can appear later in an NCName.  
 
#### URI vs URL

While I've referred to these things as [URL](https://en.wikipedia.org/wiki/URL)s, (uniform resource locator, aka web address) technically they can be a broader class of things known as [URI](https://en.wikipedia.org/wiki/Uniform_Resource_Identifier)s (uniform resource identifier).  In addition to including anything that is a URL, URIs include things which are not URLs, like `urn:isbn:0-486-27557-4` or `urn:uuid:aac06f69-7ec8-403d-ad84-baa549133dce`, which are URNs: unique resource numbers in some numbering scheme (e.g. book ISBN numbers, or UUIDs), neither of which are URLs but nonetheless enjoy the same globally unique property.

#### Blank nodes

Sometimes we do not need a globally unique identifier, we just want a way to refer to a node (e.g. subject, and sometimes an object) uniquely in our document. This is the role of a [blank node](https://en.wikipedia.org/wiki/Blank_node) (do follow the link for a better overview).  These are frequently denoted with the prefix `_:`, e.g. we could have replaced the sample IDs as `_:1`, `_:2` instead of the URLs such as `http://example.com/iris#1` etc.  Note that RDF operations need not preserve the actual string pattern in a blank ID name, it means the exact same thing if we replace all the `_:1`s with `_:b1` and `_:2` with `_:b2`,  etc.

In `librdf` we can get a blank node by passing an empty string or character string that is not a URI as the subject.  Here we also use a URI that isn't a URL as predicate: 

```{r}
rdf <- rdf()
rdf %>% rdf_add("",   
                "iris:Sepal.Length", 
                object = 5.1)
rdf
```

Note that we get a blank node, `_:`  with a randomly generated string.  

## Triple notation: `nquads` `rdfxml`, `turtle`, and `nquads`

So far we have relied primarily on a three-column tabular format to represent our triples.  We have also seen the default `print` format used for the `rdf` method, known as [N-Quads](https://www.w3.org/TR/n-quads/) above, which displays a bare, space-separated triple, possibly with a datatype URI attached to the object.  The line ends with a dot, which indicates this is part of the same local triplestore (aka RDF graph or RDF Model).  Technically this could be another URI indicating a unique global address for the triplestore in question.   

We can serialize any `rdf` object out to a file in this format with the `rdf_serialize()` function, e.g.

```{r}
rdf_serialize(rdf, "rdf.nq", format = "nquads")
```

 Just as each of these formats can be serialized with `rdf_serialize()`, each can be read by `rdflib` using the function `rdf_parse()`:


```{r}
doc <- system.file("extdata/example.rdf", package="redland")
rdf <- rdf_parse(doc, format = "rdfxml") 
rdf
```


N-Quads are convenient in that each triple is displayed on a unique line, and the format supports the blank node and Datatype URIs in the manner we have just discussed. Other formats are not so concise.  Rather than print to file, we can simply change the default print format used by `rdflib` to explore the textual layout of the other serializations.  Here is one of the most common classical serializations, `RDF/XML` which expresses triples in an XML-based schema:

```{r}
options(rdf_print_format = "rdfxml")
rdf
```

Just looking at this is probably enough to explain why so many alternative serializations were created.  Another popular format, [`turtle`](https://www.w3.org/TR/turtle/), looks more like `nquads`: 


```{r}
options(rdf_print_format = "turtle")
rdf
```

Here, blank nodes are denoted by `[]`.  `turtle` uses indentation to indicate that all three predicates (`creator`, `description`, `title`) are properties of the same subject.  



### JSON-LD

While formats such as `nquads` and `turtle` provide a much cleaner syntax than RDF/XML, they also introduce a custom format rather than building on a familiar standard (like XML) for which users already have a well-developed set of tools and intuition.  After more than a decade of such challenges (RDF specification started 1997, including an the HTML-embedded serialization of [RDFa](https://en.wikipedia.org/wiki/RDFa) in 2004), a more user friendly specification has emerged in the form of JSON-LD (1.0 W3C specification was released in 2014, the 1.1 specification released in February 2018).  JSON-LD uses the familiar *object notation* of JSON, (which is rapidly replacing XML as the ubiquitous data exchange format, and will be more familiar to many readers than the specialized RDF formats or even XML.  Here is our `rdf` data in the JSON-LD serialization:


```{r}
options(rdf_print_format = "jsonld")
rdf
```

In this serialization, our subject corresponds to "the thing in the curly braces," (i.e. the JSON "object") which is identified by the special `@id` property (omitting `@id` corresponds to a blank node).  The predicate-object pairs in the triple are then just JSON key-value pairs within the curly braces of the given object.  We can make this format look even more natural by stripping out the URLs.  While it is possible to use prefixes in place of URLs, it is more natural to pull them out entirely, e.g. by declaring a default vocabulary in the JSON-LD "Context", like so:

```{r}
rdf_serialize(rdf, "example.json", "jsonld") %>% 
  jsonld_compact(context = '{"@vocab": "http://purl.org/dc/elements/1.1/"}')
```

The context of a JSON-LD file can also define datatypes, use multiple namespaces, and permit different names in the JSON keys from that found in the URLs.  While a complete introduction to JSON-LD is beyond our scope, this representation essentially provides a way to map intuitive JSON structures into precise RDF triples. 



## From tables to Graphs

So far we have considered examples where the data could be represented in tabular form. 
We frequently encounter data that cannot be easily represented in such a format.  For instance, consider
the JSON data in this example:

```{r}
ex <- system.file("extdata/person.json", package="rdflib")
cat(readLines(ex), sep = "\n")
#jsonld_compact(ex, "{}")

```


This JSON object for a `Person` has another JSON object nested inside (a `PostalAddress`).  Yet if we look at this data as `nquads`, we see the familiar flat triple structure: 



```{r}
options(rdf_print_format = "nquads")
rdf <- rdf_parse(ex, "jsonld")
rdf
```

So what has happened?  Note that our `address` has been given the blank node URI `_:b0`, which serves both as the object in the `address` line of the `Person` and as the subject of all the properties belonging to the `PostalAddress`.  In JSON-LD, this structure is referred to as being 'flattened': 

```{r}
jsonld_flatten(ex, context = "http://schema.org")
```

Note that our JSON-LD structure now starts with an object called `@graph`.  Unlike our opening examples, this data is not tabular in nature, but rather, is formatted as a nested _graph_.  Such nesting is very natural in JSON, where objects can be arranged in a tree-like structure with a single outer-most set of `{}` indicating a root object.  A graph is just a more generic form of a tree structure, where we are agnostic to the root.  (We could in fact use the `@reverse` property on address to create a root `PostalAddress` that contains the `Person`).  In this way, the notion of data as a `graph` offers a powerful generalization to the notion of tabular data.  The `@graph` above consists of two separate objects: a `PostalAddress` (with `id` of `_:b0`) and a `Person` (with an ORCID id).  This layout acts much like a foreign key in a relational database, or as a list-column in `tidyverse` (e.g. see `tidyr::nest()`).  `rdflib` uses this flattened representation when serializing JSON-LD objects.  Note that JSON-LD provides a rich set of utilities to go back and forth between flattened and nested layouts using `jsonld_frame`.  For instance, we can recover the original structure just by specifying a frame that indicates which type we want as the root:

```{r}
jsonld_flatten(ex) %>%
  jsonld_frame('{"@type": "http://schema.org/Person"}') %>%
  jsonld_compact(context = "http://schema.org")
```

(Recall that compacting just replaces URIs and any type declarations with short names given by the context).  This is somewhat analogous to `join` operations in relational data, or nesting and un-nesting functions in `tidyr`.  However, when working with RDF, the beautiful thing is that the differences between these two representations (nested or flattened) are purely aesthetic.  Both representations have precisely the same semantic meaning, and are thus precisely the same thing in RDF world.  We will never have to orchestrate a join on a foreign key before we can perform desired operations like select and filter on the data.  We don't have to think about how our data is organized, because it is always in the same molten triple format, whatever it is, and however nested it might be.  

Just as we saw `gather` could provide a relatively generic way of transforming a data.frame into RDF triples, JSON-LD defines a relatively simple convention for getting nested data (e.g. lists) into RDF triples.  This convention simply treats JSON `{}` objects as `subjects` (often assigning blank node ids, as we saw with row ids), and key-value pairs (or in R-speak, list names and values) as predicates and objects, respectively.  Any raw JSON file can be treated as JSON-LD, ideally by specifying an appropriate `context`, which serves to map terms into URIs as we saw with data.frames. `JSON-LD` is then already a valid RDF format that we can parse with `rdflib`.  

For instance, here is a simple function for coercing list objects into RDF with a specified context:

```{r}
as_rdf.list <- function(x, context = "http://schema.org"){
  if(length(x) == 1) x <- x[[1]]
  x[["@context"]] <- context
  json <- jsonlite::toJSON(x, pretty = TRUE, auto_unbox = TRUE, force = TRUE)
  rdflib::rdf_parse(json, "jsonld")
}
```

Here we set a default context (http://schema.org), and map a few R terms to corresponding schema terms

```{r}
context <- list("http://schema.org", 
                list(schema = "http://schema.org/",
                     given = "givenName",
                     family = "familyName",
                     title = "name",
                     year = "datePublished",
                     note = "softwareVersion",
                     comment = "identifier",
                     role = "http://www.loc.gov/marc/relators/relaterm.html"))

```


We can now apply our function on arbitrary R `list` objects, such as the `bibentry` class object returned by the `citation()` function:

```{r}
options(rdf_print_format = "nquads") # go back to the default


R <- citation("rdflib")
rdf <- as_rdf.list(R, context)
rdf  
```


## SPARQL: A Graph Query Language

So far, we have spent a lot of words describing how to transform data into RDF, and not much actually _doing anything_ cool with said data. 


_Still working on writing this section_

```{r}
#source(system.file("examples/as_rdf.R", package="rdflib"))
source(system.file("examples/tidy_schema.R", package="rdflib"))

## Testing: Digest some data.frames into RDF and extract back
 cars <- mtcars %>% rownames_to_column("Model")
 x1 <- as_rdf(iris, NULL, "iris:")
 x2 <- as_rdf(cars, NULL, "mtcars:")
 rdf <- c(x1,x2)
```


## SPARQL: Getting back to Tidy Tables!


```{r}
sparql <-
  'SELECT  ?Species ?Sepal_Length ?Sepal_Width ?Petal_Length  ?Petal_Width
WHERE {
 ?s <iris:Species>  ?Species .
 ?s <iris:Sepal.Width>  ?Sepal_Width .
 ?s <iris:Sepal.Length>  ?Sepal_Length . 
 ?s <iris:Petal.Length>  ?Petal_Length .
 ?s <iris:Petal.Width>  ?Petal_Width 
}'

iris2 <- rdf_query(rdf, sparql)
```

```{r echo=FALSE}
DT::datatable(iris2)
```


We can automatically create the a SPARQL query that returns "tidy data".  Tidy data has predicates as columns, objects as values, subjects as rows.  

```{r}
sparql <- tidy_schema("Species",  "Sepal.Length", "Sepal.Width", prefix = "iris")

rdf_query(rdf, sparql)
```


## Multi-table queries & non-rectangular data

In the following vignette, "The Data Lake and Schema On Read," we explore more complex examples with multiple tables from a relational database and nested JSON data.  These highlight the power of SPARQL as a *Graph* query language. 

<!--
Thank you for your contribution to this project. This project has no formal
funding or full-time maintainers, and relies entirely on independent
contributors to keep it alive and relevant.

This pull request template includes some guidelines intended to help
contributors, not to deter contributions. While we prefer that PRs follow our
guidelines, we will not reject PRs solely on the basis that they do not, though
we may take longer to process them as in most cases the remaining work will
have to be done by someone else.

If you have any questions regarding our guidelines, submit the PR as is
and ask.

More detailed guidelines for pull requests are provided in our [developers
guide](https://github.com/RDFLib/rdflib/blob/main/docs/developers.rst).

PRs that are smaller in size and scope will be reviewed and merged quicker, so
please consider if your PR could be split up into more than one independent part
before submitting it, no PR is too small. The maintainers of this project may
also split up larger PRs into smaller, more manageable PRs, if they deem it
necessary.

PRs should be reviewed and approved by at least two people other than the author
using GitHub's review system before being merged. This is less important for bug
fixes and changes that don't impact runtime behaviour, but more important for
changes that expand the RDFLib public API. Reviews are open to anyone, so please
consider reviewing other open pull requests, as this will also free up the
capacity required for your PR to be reviewed.
-->

# Summary of changes

<!--
Briefly explain what changes the pull request is making and why. Ideally, this
should cover all changes in the pull request, as the changes will be reviewed
against this summary to ensure that the PR does not include unintended changes.

Please also explicitly state if the PR makes any changes that are not backwards
compatible.
-->

# Checklist

<!--
If an item on this list doesn't apply to your pull request, just remove it.

If, for some reason, you can't check some items on the checklist, or you are
unsure about them, submit your PR as is and ask for help.
-->

- [ ] Checked that there aren't other open pull requests for
  the same change.
- [ ] Checked that all tests and type checking passes.
- If the change adds new features or changes the RDFLib public API:
  <!-- This can be removed if no new features are added and the RDFLib public API is
  not changed. -->
  - [ ] Created an issue to discuss the change and get in-principle agreement.
  - [ ] Considered adding an example in `./examples`.
- If the change has a potential impact on users of this project:
  <!-- This can be removed if the change does not affect users of this project. -->
  - [ ] Added or updated tests that fail without the change.
  - [ ] Updated relevant documentation to avoid inaccuracies.
  - [ ] Considered adding additional documentation.
- [ ] Considered granting [push permissions to the PR branch](https://docs.github.com/en/pull-requests/collaborating-with-pull-requests/working-with-forks/allowing-changes-to-a-pull-request-branch-created-from-a-fork),
  so maintainers can fix minor issues and keep your PR up to date.

# Querying with SPARQL

## Run a Query

The RDFLib comes with an implementation of the [SPARQL 1.1 Query](http://www.w3.org/TR/sparql11-query/) and [SPARQL 1.1 Update](http://www.w3.org/TR/sparql11-update/) query languages.

Queries can be evaluated against a graph with the [`query()`][rdflib.graph.Graph.query] method, and updates with [`update()`][rdflib.graph.Graph.update].

The query method returns a [`Result`][rdflib.query.Result] instance. For SELECT queries, iterating over this returns [`ResultRow`][rdflib.query.ResultRow] instances, each containing a set of variable bindings. For `CONSTRUCT`/`DESCRIBE` queries, iterating over the result object gives the triples. For `ASK` queries, iterating will yield the single boolean answer, or evaluating the result object in a boolean-context (i.e. `bool(result)`)

For example...

```python
import rdflib
g = rdflib.Graph()
g.parse("http://danbri.org/foaf.rdf#")

knows_query = """
SELECT DISTINCT ?aname ?bname
WHERE {
    ?a foaf:knows ?b .
    ?a foaf:name ?aname .
    ?b foaf:name ?bname .
}"""

qres = g.query(knows_query)
for row in qres:
    print(f"{row.aname} knows {row.bname}")
```

The results are tuples of values in the same order as your `SELECT` arguments. Alternatively, the values can be accessed by variable name, either as attributes, or as items, e.g. `row.b` and `row["b"]` are equivalent. The above, given the appropriate data, would print something like:

```text
Timothy Berners-Lee knows Edd Dumbill
Timothy Berners-Lee knows Jennifer Golbeck
Timothy Berners-Lee knows Nicholas Gibbins
...
```

As an alternative to using `SPARQL`'s `PREFIX`, namespace bindings can be passed in with the `initNs` kwarg, see [namespaces_and_bindings](namespaces_and_bindings.md).

Variables can also be pre-bound, using the `initBindings` kwarg which can pass in a `dict` of initial bindings. This is particularly useful for prepared queries, as described below.

## Update Queries

Update queries are performed just like reading queries but using the [`update()`][rdflib.graph.Graph.update] method. An example:

```python
from rdflib import Graph

# Create a Graph, add in some test data
g = Graph()
g.parse(
    data="""
        <x:> a <c:> .
        <y:> a <c:> .
    """,
    format="turtle"
)

# Select all the things (s) that are of type (rdf:type) c:
qres = g.query("""SELECT ?s WHERE { ?s a <c:> }""")

for row in qres:
    print(f"{row.s}")
# prints:
# x:
# y:

# Add in a new triple using SPARQL UPDATE
g.update("""INSERT DATA { <z:> a <c:> }""")

# Select all the things (s) that are of type (rdf:type) c:
qres = g.query("""SELECT ?s WHERE { ?s a <c:> }""")

print("After update:")
for row in qres:
    print(f"{row.s}")
# prints:
# x:
# y:
# z:

# Change type of <y:> from <c:> to <d:>
g.update("""
         DELETE { <y:> a <c:> }
         INSERT { <y:> a <d:> }
         WHERE { <y:> a <c:> }
         """)
print("After second update:")
qres = g.query("""SELECT ?s ?o WHERE { ?s a ?o }""")
for row in qres:
    print(f"{row.s} a {row.o}")
# prints:
# x: a c:
# z: a c:
# y: a d:
```

## Querying a Remote Service

The `SERVICE` keyword of SPARQL 1.1 can send a query to a remote SPARQL endpoint.

```python
import rdflib

g = rdflib.Graph()
qres = g.query(
    """
    SELECT ?s
    WHERE {
      SERVICE <https://dbpedia.org/sparql> {
        ?s a ?o .
      }
    }
    LIMIT 3
    """
)

for row in qres:
    print(row.s)
```

This example sends a query to [DBPedia](https://dbpedia.org/)'s SPARQL endpoint service so that it can run the query and then send back the result:

```text
<http://www.w3.org/1999/02/22-rdf-syntax-ns#type> <http://www.openlinksw.com/schemas/virtcxml#FacetCategoryPattern>
<http://www.w3.org/2001/XMLSchema#anyURI> <http://www.w3.org/2000/01/rdf-schema#Datatype>
<http://www.w3.org/2001/XMLSchema#anyURI> <http://www.w3.org/2000/01/rdf-schema#Datatype>
```

## Prepared Queries

RDFLib lets you *prepare* queries before execution, this saves re-parsing and translating the query into SPARQL Algebra each time.

The method [`prepareQuery()`][rdflib.plugins.sparql.prepareQuery] takes a query as a string and will return a [`Query`][rdflib.plugins.sparql.sparql.Query] object. This can then be passed to the [`query()`][rdflib.graph.Graph.query] method.

The `initBindings` kwarg can be used to pass in a `dict` of initial bindings:

```python
q = prepareQuery(
    "SELECT ?s WHERE { ?person foaf:knows ?s .}",
    initNs = { "foaf": FOAF }
)

g = rdflib.Graph()
g.parse("foaf.rdf")

tim = rdflib.URIRef("http://www.w3.org/People/Berners-Lee/card#i")

for row in g.query(q, initBindings={'person': tim}):
    print(row)
```

## Custom Evaluation Functions

For experts, it is possible to override how bits of SPARQL algebra are evaluated. By using the [setuptools entry-point](http://pythonhosted.org/distribute/setuptools.html#dynamic-discovery-of-services-and-plugins) `rdf.plugins.sparqleval`, or simply adding to an entry to [`CUSTOM_EVALS`][rdflib.plugins.sparql.CUSTOM_EVALS], a custom function can be registered. The function will be called for each algebra component and may raise `NotImplementedError` to indicate that this part should be handled by the default implementation.

See [`examples/custom_eval.py`][examples.custom_eval]

---
title: The Data Lake and Schema On Read
author: Carl Boettiger
date: "`r Sys.Date()`"
output: 
  rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{The Data Lake and Schema On Read}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
  
---


```{r include = FALSE}
knitr::opts_chunk$set(message=FALSE, warning = FALSE, cache = TRUE)
```


A provocative recent analogy (e.g. [Archer, 2017](https://www.w3.org/blog/2017/04/dive-into-the-semantic-data-lake/)) for thinking about RDF is that of the Data Lake.  Whereas adding new data to a traditional relational database frequently involves laborious wrangling into the an existing rigid data *schema* of tables and columns, RDF can enable a far more simple approach of just 'tossing everything into the data lake.' While this may sound like a recipe for a terrible mess down the road, RDF gives the promise of *schema-on-read*: rather than dictating the shape of your data when first adding it to the database (i.e. schema-on-write), a SPARQL query auto-magically reaches into the lake and retrieves all the relevant data, returning it in exactly the schema you ask for; that is, as a nice single table containing only the requested columns.  

This can serve as a very effective means of data integration (provided a reasonably consistent and diligent use of URIs in identifying subjects and properties (predicates)), since just about any data can be added to the lake without worrying about whether it comes in a schema that matches the existing architecture of the database.  It is this flexibility not to have to define your database schema at the start that is the primary strength of the RDF approach. 

A key advantage of this approach is that it is equally as easy to extract your desired data.frame from non-rectangular data as it is from tables or relational database systems. While the previous vignette focused on simple examples including data from single `data.frames` or small JSON files, this vignette showcases the Data Lake approach on a more complex relational database organized over several tables (`nyflights13` data, used to teach `joins` and other relational data in [Grolemund & Wickham, 2017](http://r4ds.had.co.nz)), and also in a large non-tabular file returned from the GitHub API (e.g. as used in [Bryan & Wickham, 2017 ](https://dcl-2017-04.github.io/curriculum/rectangling.html) to teach data rectangling.) 

Load the necessary libraries to get started: 

```{r libraries}
## Data 
library(nycflights13)
library(repurrrsive)

## for comparison approaches
library(dplyr)
library(purrr)
library(jsonlite)

## Our focal package:
library(rdflib)
## experimental functions for rdflib package
source(system.file("examples/tidy_schema.R", package="rdflib"))

```

Configure RDF storage to use the BDB backend for on-disk storage.

```{r options, include = TRUE}
rdf <- rdf(storage = "BDB", new_db = TRUE)

```

# Relational data


## The `tidyverse` approach

`tidyverse` operations are incredibly effective for working with relational data.  These `dplyr` on the `nyflights13` dataset are easy to write and fast to execute:

```{r tidyverse}
df <- flights %>% 
  left_join(airlines) %>%
  left_join(planes, by="tailnum") %>% 
  select(carrier, name, manufacturer, model) %>% 
  distinct()
head(df)
```
 
Still, joins are often a challenge in data preparation.  Tabular formats can often be sloppy about what is a key column and what is a literal value, and also whether a column with the same name in different tables means the same thing in both.  Both of these things pose challenges for later use when joining data.  RDF representation encourages greater discipline through the use of URIs (though we'll run a bit roughshod over that with the caviler use of `x:` here.)  

This example uses only data that is already part of the relational database.  Adding additional information to the database is frequently more tricky, and can result in a rapidly expanding number of tables that can become difficult to work across.
 

## RDF approach

Okay, now let's dump the `nyflights13` into the data lake. Foreign keys in any table must be represented as URIs and not literal strings: 

```{r as_uri}
uri_flights <- flights %>% 
  mutate(tailnum = paste0("planes:", tailnum),
         carrier = paste0("airlines:", carrier))
```

We write the `data.frame`s out as nquads.  Recall that each cell of a `data.frame` can be represented as a triple, in which the column is the predicate, the primary key (or row number) the subject, and the cell value the object.  We turn column names and primary keys into URIs using a prefix based on the table name.  (Note that `rdflib` does this conversion by merely munging cells and calling `write.table`, it is not a standard `redland` library transform).

```{r into_quads}
system.time({
  write_nquads(airlines,  "airlines.nq", key = "carrier", prefix = "airlines:")
  write_nquads(planes,  "planes.nq", key = "tailnum", prefix = "planes:")
  write_nquads(uri_flights,  "flights.nq", prefix = "flights:")
})
```

We can now read these into our RDF data lake:

```{r write_rdf}

system.time({
  read_nquads("airlines.nq", rdf = rdf)
  read_nquads("flights.nq", rdf = rdf)
  read_nquads("planes.nq", rdf = rdf)

})
```

Note that flights does not have a natural key (somewhat surprisingly, `flight` number is not a unique key for this table, as the same flight number is reused on the same route at different times.)  So, we will treat each row as a unique anonymous key by setting the key to `NULL`.

## Schema on read

We simply define the columns we want and we immediately get back the desired `data.frame`:


```{r query}
s <- 
  'SELECT  ?carrier ?name ?manufacturer ?model ?dep_delay
WHERE {
?flight <flights:tailnum>  ?tailnum .
?flight <flights:carrier>  ?carrier .
?flight <flights:dep_delay>  ?dep_delay .
?tailnum <planes:manufacturer> ?manufacturer .
?tailnum <planes:model> ?model .
?carrier <airlines:name> ?name
}'

system.time(
df <- rdf_query(rdf, s)
)

head(df)
```



Note that in place of joins, we give more semantically meaningful statements about the data:
e.g. `manufacturer` is a property of a `tailnum` (corresponding to a particular physical aircraft), not of a `flight` number.  Departure delay `dep_delay` is a property of a flight, not of an aircraft (`tailnum`).  

This is reminiscent of  the way in which these data are organized in the relational database tables to begin with: we find `deb_delay` in the `flights` table and `manufacturer` in the `planes` table. Good relational design encourages this, but to work with the data the user is often left having to do the required joins, which also creates tables where these semantics are less clear.  




# Non-tabular data

## The `tidyverse` approach

We start with data from [Bryan & Wickham, 2017 ](https://dcl-2017-04.github.io/curriculum/rectangling.html) lesson on data rectangling:

```{r}
f <- system.file("extdata/gh_repos.json", package="repurrrsive")
gh_data <- jsonlite::read_json(f)
```


The original lesson illustrates the power and reasonably concise syntax of the `tidyverse` package, `purrr`, to iterate over this complex structure to extract the necessary data.  In this approach, nesting of the data is largely a nuisance to overcome rather than an asset to the data analyst:

```{r }
gh_flat <- gh_data %>% purrr::flatten()  # abandon nested structure and hope we didn't need it

gh_tibble <- tibble(
  name =     gh_flat %>% map_chr("name"),
  issues =   gh_flat %>% map_int("open_issues_count"),
  wiki =     gh_flat %>% map_lgl("has_wiki"),
  homepage = gh_flat %>% map_chr("homepage", .default = ""),
  owner =    gh_flat %>% map_chr(c("owner", "login"))
)

gh_tibble %>% arrange(name) %>% head()
```

## RDF on non-tabular data

The RDF approach merely treats JSON as JSON-LD within a given vocabulary.  In this context, nesting implicitly provides important semantic information about relationships between the data, which are captured in the RDF triples. Here, we import the JSON data as RDF (and add it to our existing triplestore just for fun)

```{r gh_add}
gh_rdf <- as_rdf(gh_data, rdf = rdf, prefix = "gh:")
```





And we can query it back out of the lake just by selecting the columns of interest. 

```{r gh_query}
s <- 
  'SELECT ?name ?issues ?wiki ?homepage ?owner
WHERE {
?repo <gh:homepage>  ?homepage .
?repo <gh:has_wiki> ?wiki .
?repo <gh:open_issues_count> ?issues .
?repo <gh:name> ?name .
?repo <gh:owner> ?owner_id .
?owner_id <gh:login>  ?owner 
}'

system.time(
rdf_tibble <- rdf_query(rdf, s)
)

head(rdf_tibble)
```

## Going further: A `dplyr`-style syntax for SPARQL?

`dplyr` provides a reasonably intuitive, powerful, and concise interface for many common SQL commands.  Indeed, `dplyr` function calls are literally serialized to the corresponding SQL query when working with a relational database back-end (via `dbplyr`).  Can a similar API be developed for SPARQL queries?

SPARQL syntax is obviously inspired by SQL syntax, and includes many of the same operations found in SQL and `dplyr` (e.g. SELECT, WHERE, FILTER, DISTINCT) as well as other more RDF specific queries. 

If we are willing to make some assumptions about the most common queries, we can start to make some simplifying functions.  For instance, in the above patterns, the variables being returned are always objects from the triples, with columns being named using the corresponding predicate.  Assuming some additional convention to define the prefixes and indicate graph traversal (i.e. nested values), we could have constructed the above query from a call:

```r
tidy_schema(name, open_issues_count, has_wiki, homepage owner.login)
```

Though to generalize to arbitrary labels and predicates this might need to support something like:

```r
tidy_schema(list = list(name = "gh:name", 
                        issues = "gh:open_issues_count", 
                        wiki = "gh:has_wiki", 
                        homepage = "gh:homepage",
                        owner = list("gh:owner", login = "gh:login")))
```

# Persisting Notation 3 Terms

## Using N3 Syntax for Persistence

Blank Nodes, Literals, URI References, and Variables can be distinguished in persistence by relying on Notation 3 syntax convention.

All URI References can be expanded and persisted as:

```turtle
<..URI..>
```

All Literals can be expanded and persisted as:

```turtle
"..value.."@lang or "..value.."^^dtype_uri
```

!!! abstract "Language tag"
    `@lang` is a language tag and `^^dtype_uri` is the URI of a data type associated with the Literal

Blank Nodes can be expanded and persisted as:

```turtle
_:Id
```

!!! info "About skolemization"
    Where Id is an identifier as determined by skolemization. Skolemization is a syntactic transformation routinely used in automatic inference systems in which existential variables are replaced by 'new' functions - function names not used elsewhere - applied to any enclosing universal variables. In RDF, Skolemization amounts to replacing every blank node in a graph by a 'new' name, i.e. a URI reference which is guaranteed to not occur anywhere else. In effect, it gives 'arbitrary' names to the anonymous entities whose existence was asserted by the use of blank nodes: the arbitrariness of the names ensures that nothing can be inferred that would not follow from the bare assertion of existence represented by the blank node. (Using a literal would not do. Literals are never 'new' in the required sense.)

Variables can be persisted as they appear in their serialization `(?varName)` - since they only need be unique within their scope (the context of their associated statements)

These syntactic conventions can facilitate term round-tripping.

## Variables by Scope

Would an interface be needed in order to facilitate a quick way to aggregate all the variables in a scope (given by a formula identifier)? An interface such as:

```python
def variables(formula_identifier)
```

## The Need to Skolemize Formula Identifiers

It would seem reasonable to assume that a formula-aware store would assign Blank Node identifiers as names of formulae that appear in a N3 serialization. So for instance, the following bit of N3:

```
{?x a :N3Programmer} => {?x :has :Migrane}
```

Could be interpreted as the assertion of the following statement:

```turtle
_:a log:implies _:b
```

However, how are `_:a` and `_:b` distinguished from other Blank Nodes? A formula-aware store would be expected to persist the first set of statements as quoted statements in a formula named `_:a` and the second set as quoted statements in a formula named `_:b`, but it would not be cost-effective for a serializer to have to query the store for all statements in a context named `_:a` in order to determine if `_:a` was associated with a formula (so that it could be serialized properly).

## Relying on `log:Formula` Membership

The store could rely on explicit `log:Formula` membership (via `rdf:type` statements) to model the distinction of Blank Nodes associated with formulae. However, would these statements be expected from an N3 parser or known implicitly by the store? i.e., would all such Blank Nodes match the following pattern:

```turtle
?formula rdf:type log:Formula
```

## Relying on an Explicit Interface

A formula-aware store could also support the persistence of this distinction by implementing a method that returns an iterator over all the formulae in the store:

```python
def formulae(triple=None)
```

This function would return all the Blank Node identifiers assigned to formulae or just those that contain statements matching the given triple pattern and would be the way a serializer determines if a term refers to a formula (in order to properly serializer it).

How much would such an interface reduce the need to model formulae terms as first class objects (perhaps to be returned by the [`triples()`][rdflib.Graph.triples] function)? Would it be more useful for the [`Graph`][rdflib.Graph] (or the store itself) to return a Context object in place of a formula term (using the formulae interface to make this determination)?

Conversely, would these interfaces (variables and formulae) be considered optimizations only since you have the distinction by the kinds of terms triples returns (which would be expanded to include variables and formulae)?

## Persisting Formula Identifiers

This is the most straight forward way to maintain this distinction - without relying on extra interfaces. Formula identifiers could be persisted distinctly from other terms by using the following notation:

```
{_:bnode} or {<.. URI ..>}
```

This would facilitate their persistence round-trip - same as the other terms that rely on N3 syntax to distinguish between each other.

# Upgrading 5.0.0 to 6.0.0

6.0.0 fully adopts Python 3 practices and drops Python 2 support so it is neater, faster and generally more modern than 5.0.0. It also tidies up the [`Graph`][rdflib.graph.Graph] API (removing duplicate functions) so it does include a few breaking changes. Additionally, there is a long list of PRs merged into 6.0.0 adding a number of small fixes and features which are listed below.

RDFLib version 5.0.0 was released in 2020, 3 years after the previous version (4.2.2) and is fundamentally 5.0.0 compatible with. If you need very long-term backwards-compatibility or Python 2 support, you need 5.0.0.

## Major Changes

The most notable changes in RDFLib 6.0.0 are:

### Python 3.7+

* The oldest version of python you can use to run RDFLib is now 3.7.
* This is a big jump from RDFLib 5.0.0 that worked on python 2.7 and 3.5.
* This change is to allow the library maintainers to adopt more modern development tools, newer language features, and avoid the need to support EOL versions of python in he future

### JSON-LD integration and JSON-LD 1.1

* The json-ld serializer/parser plugin was by far the most commonly used RDFLib addon.
* Last year we brought it under the RDFLib org in Github
* Now for 6.0.0 release the JSON-LD serializer and parser are integrated into RDFLib core
* This includes the experimental support for the JSON-LD v1.1 spec
* You no longer need to install the json-ld dependency separately.

## All Changes

This list has been assembled from Pull Request and commit information.

### General Bugs Fixed

* Pr 451 redux
  [PR #978](https://github.com/RDFLib/rdflib/pull/978)

### Enhanced Features

* Register additional serializer plugins for SPARQL mime types.
  [PR #987](https://github.com/RDFLib/rdflib/pull/987)

### SPARQL Fixes

* Total order patch patch
  [PR #862](https://github.com/RDFLib/rdflib/pull/862)

### Code Quality and Cleanups

* a slightly opinionated autopep8 run
  [PR #870](https://github.com/RDFLib/rdflib/pull/870)

### Testing

* 3.7 for travis
  [PR #864](https://github.com/RDFLib/rdflib/pull/864)

### Documentation Fixes

* Fix a doc string in the query module
  [PR #976](https://github.com/RDFLib/rdflib/pull/976)

### Integrate JSON-LD into RDFLib

[PR #1354](https://github.com/RDFLib/rdflib/pull/1354)

# Persistence

RDFLib provides an [`abstracted Store API`][rdflib.store.Store]
for persistence of RDF and Notation 3. The [`Graph`][rdflib.graph.Graph] class works with instances of this API (as the first argument to its constructor) for triple-based management of an RDF store including: garbage collection, transaction management, update, pattern matching, removal, length, and database management ([`Graph.open()`][rdflib.graph.Graph.open] / [`Graph.close()`][rdflib.graph.Graph.close] / [`Graph.destroy()`][rdflib.graph.Graph.destroy]).

Additional persistence mechanisms can be supported by implementing this API for a different store.

## Stores currently shipped with core RDFLib

* [`Memory`][rdflib.plugins.stores.memory.Memory] - not persistent!
* [`BerkeleyDB`][rdflib.plugins.stores.berkeleydb.BerkeleyDB] - on disk persistence via Python's [berkeleydb package](https://pypi.org/project/berkeleydb/)
* [`SPARQLStore`][rdflib.plugins.stores.sparqlstore.SPARQLStore] - a read-only wrapper around a remote SPARQL Query endpoint
* [`SPARQLUpdateStore`][rdflib.plugins.stores.sparqlstore.SPARQLUpdateStore] - a read-write wrapper around a remote SPARQL query/update endpoint pair

## Usage

In most cases, passing the name of the store to the Graph constructor is enough:

```python
from rdflib import Graph

graph = Graph(store='BerkeleyDB')
```

Most stores offering on-disk persistence will need to be opened before reading or writing. When peristing a triplestore, rather than a ConjuntiveGraph quadstore, you need to specify an identifier with which you can open the graph:

```python
graph = Graph('BerkeleyDB', identifier='mygraph')

# first time create the store:
graph.open('/home/user/data/myRDFLibStore', create=True)

# work with the graph:
data = """
PREFIX : <https://example.org/>

:a :b :c .
:d :e :f .
:d :g :h .
"""
graph.parse(data=data, format="ttl")

# when done!
graph.close()
```

When done, [`close()`][rdflib.graph.Graph.close] must be called to free the resources associated with the store.

## Additional store plugins

More store implementations are available in RDFLib extension projects:

* [rdflib-sqlalchemy](https://github.com/RDFLib/rdflib-sqlalchemy) – a store which supports a wide-variety of RDBMS backends
* [rdflib-leveldb](https://github.com/RDFLib/rdflib-leveldb) – a store on top of Google's [LevelDB](https://code.google.com/p/leveldb/) key-value store
* [rdflib-kyotocabinet](https://github.com/RDFLib/rdflib-kyotocabinet) – a store on top of the [Kyoto Cabinet](http://fallabs.com/kyotocabinet/) key-value store

## Example

* [`examples.berkeleydb_example`][examples.berkeleydb_example] contains an example for using a BerkeleyDB store.
* [`examples.sparqlstore_example`][examples.sparqlstore_example] contains an example for using a SPARQLStore.

# Creating RDF triples

## Creating Nodes

RDF data is a graph where the nodes are URI references, Blank Nodes or Literals. In RDFLib, these node types are represented by the classes [`URIRef`][rdflib.term.URIRef], [`BNode`][rdflib.term.BNode], and [`Literal`][rdflib.term.Literal]. `URIRefs` and `BNodes` can both be thought of as resources, such a person, a company, a website, etc.

* A `BNode` is a node where the exact URI is not known - usually a node with identity only in relation to other nodes.
* A `URIRef` is a node where the exact URI is known. In addition to representing some subjects and predicates in RDF graphs, `URIRef`s are always used to represent properties/predicates
* `Literals` represent object values, such as a name, a date, a number, etc. The most common literal values are XML data types, e.g. string, int... but custom types can be declared too

Nodes can be created by the constructors of the node classes:

```python
from rdflib import URIRef, BNode, Literal

bob = URIRef("http://example.org/people/Bob")
linda = BNode()  # a GUID is generated

name = Literal("Bob")  # passing a string
age = Literal(24)  # passing a python int
height = Literal(76.5)  # passing a python float
```

Literals can be created from Python objects, this creates `data-typed literals`. For the details on the mapping see [rdflibliterals](rdf_terms.md).

For creating many `URIRefs` in the same `namespace`, i.e. URIs with the same prefix, RDFLib has the [`Namespace`][rdflib.namespace.Namespace] class

```python
from rdflib import Namespace

n = Namespace("http://example.org/people/")

n.bob  # == rdflib.term.URIRef("http://example.org/people/bob")
n.eve  # == rdflib.term.URIRef("http://example.org/people/eve")
```

This is very useful for schemas where all properties and classes have the same URI prefix. RDFLib defines Namespaces for some common RDF/OWL schemas, including most W3C ones:

```python
from rdflib.namespace import CSVW, DC, DCAT, DCTERMS, DOAP, FOAF, ODRL2, ORG, OWL, \
                           PROF, PROV, RDF, RDFS, SDO, SH, SKOS, SOSA, SSN, TIME, \
                           VOID, XMLNS, XSD

RDF.type
# == rdflib.term.URIRef("http://www.w3.org/1999/02/22-rdf-syntax-ns#type")

FOAF.knows
# == rdflib.term.URIRef("http://xmlns.com/foaf/0.1/knows")

PROF.isProfileOf
# == rdflib.term.URIRef("http://www.w3.org/ns/dx/prof/isProfileOf")

SOSA.Sensor
# == rdflib.term.URIRef("http://www.w3.org/ns/sosa/Sensor")
```


## Adding Triples to a graph

We already saw in [intro_to_parsing](intro_to_parsing.md), how triples can be added from files and online locations with the [`parse()`][rdflib.graph.Graph.parse] function.

Triples can also be added within Python code directly, using the [`add()`][rdflib.graph.Graph.add] function:

[`add()`][rdflib.graph.Graph.add] takes a 3-tuple (a "triple") of RDFLib nodes. Using the nodes and namespaces we defined previously:

```python
from rdflib import Graph, URIRef, Literal, BNode
from rdflib.namespace import FOAF, RDF

g = Graph()
g.bind("foaf", FOAF)

bob = URIRef("http://example.org/people/Bob")
linda = BNode()  # a GUID is generated

name = Literal("Bob")
age = Literal(24)

g.add((bob, RDF.type, FOAF.Person))
g.add((bob, FOAF.name, name))
g.add((bob, FOAF.age, age))
g.add((bob, FOAF.knows, linda))
g.add((linda, RDF.type, FOAF.Person))
g.add((linda, FOAF.name, Literal("Linda")))

print(g.serialize())
```

outputs:

```turtle
@prefix foaf: <http://xmlns.com/foaf/0.1/> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

<http://example.org/people/Bob> a foaf:Person ;
    foaf:age 24 ;
    foaf:knows [ a foaf:Person ;
            foaf:name "Linda" ] ;
    foaf:name "Bob" .
```

For some properties, only one value per resource makes sense (i.e they are *functional properties*, or have a max-cardinality of 1). The [`set()`][rdflib.graph.Graph.set] method is useful for this:

```python
from rdflib import Graph, URIRef, Literal
from rdflib.namespace import FOAF

g = Graph()
bob = URIRef("http://example.org/people/Bob")

g.add((bob, FOAF.age, Literal(42)))
print(f"Bob is {g.value(bob, FOAF.age)}")
# prints: Bob is 42

g.set((bob, FOAF.age, Literal(43)))  # replaces 42 set above
print(f"Bob is now {g.value(bob, FOAF.age)}")
# prints: Bob is now 43
```


[`value()`][rdflib.graph.Graph.value] is the matching query method. It will return a single value for a property, optionally raising an exception if there are more.

You can also add triples by combining entire graphs, see [graph-setops](intro_to_graphs.md).

## Removing Triples

Similarly, triples can be removed by a call to [`remove()`][rdflib.graph.Graph.remove]:

When removing, it is possible to leave parts of the triple unspecified (i.e. passing `None`), this will remove all matching triples:

```python
g.remove((bob, None, None))  # remove all triples about bob
```


## An example

LiveJournal produces FOAF data for their users, but they seem to use `foaf:member_name` for a person's full name but `foaf:member_name` isn't in FOAF's namespace and perhaps they should have used `foaf:name`

To retrieve some LiveJournal data, add a `foaf:name` for every `foaf:member_name` and then remove the `foaf:member_name` values to ensure the data actually aligns with other FOAF data, we could do this:

```python
from rdflib import Graph
from rdflib.namespace import FOAF

g = Graph()
# get the data
g.parse("http://danbri.livejournal.com/data/foaf")

# for every foaf:member_name, add foaf:name and remove foaf:member_name
for s, p, o in g.triples((None, FOAF['member_name'], None)):
    g.add((s, FOAF['name'], o))
    g.remove((s, FOAF['member_name'], o))
```

!!! info "Foaf member name"
    Since rdflib 5.0.0, using `foaf:member_name` is somewhat prevented in RDFlib since FOAF is declared as a [`ClosedNamespace`][rdflib.namespace.ClosedNamespace] class instance that has a closed set of members and `foaf:member_name` isn't one of them! If LiveJournal had used RDFlib 5.0.0, an error would have been raised for `foaf:member_name` when the triple was created.


## Creating Containers & Collections

There are two convenience classes for RDF Containers & Collections which you can use instead of declaring each triple of a Containers or a Collections individually:

* [`Container`][rdflib.container.Container] (also `Bag`, `Seq` & `Alt`) and
* [`Collection`][rdflib.collection.Collection]

See their documentation for how.

# Merging graphs

Graphs share blank nodes only if they are derived from graphs described by documents or other structures (such as an RDF dataset) that explicitly provide for the sharing of blank nodes between different RDF graphs. Simply downloading a web document does not mean that the blank nodes in a resulting RDF graph are the same as the blank nodes coming from other downloads of the same document or from the same RDF source.

RDF applications which manipulate concrete syntaxes for RDF which use blank node identifiers should take care to keep track of the identity of the blank nodes they identify. Blank node identifiers often have a local scope, so when RDF from different sources is combined, identifiers may have to be changed in order to avoid accidental conflation of distinct blank nodes.

For example, two documents may both use the blank node identifier "_:x" to identify a blank node, but unless these documents are in a shared identifier scope or are derived from a common source, the occurrences of "_:x" in one document will identify a different blank node than the one in the graph described by the other document. When graphs are formed by combining RDF from multiple sources, it may be necessary to standardize apart the blank node identifiers by replacing them by others which do not occur in the other document(s).

_(copied directly from <https://www.w3.org/TR/rdf11-mt/#shared-blank-nodes-unions-and-merges>_

In RDFLib, blank nodes are given unique IDs when parsing, so graph merging can be done by simply reading several files into the same graph:

```python
from rdflib import Graph

graph = Graph()

graph.parse(input1)
graph.parse(input2)
```

`graph` now contains the merged graph of `input1` and `input2`.

!!! warning "Blank Node Collision"
    However, the set-theoretic graph operations in RDFLib are assumed to be performed in sub-graphs of some larger data-base (for instance, in the context of a [`Dataset`][rdflib.graph.Dataset]) and assume shared blank node IDs, and therefore do NOT do _correct_ merging, i.e.:

    ```python
    from rdflib import Graph

    g1 = Graph()
    g1.parse(input1)

    g2 = Graph()
    g2.parse(input2)

    graph = g1 + g2
    ```

    May cause unwanted collisions of blank-nodes in `graph`.

# Utilities & convenience functions

For RDF programming, RDFLib and Python may not be the fastest tools, but we try hard to make them the easiest and most convenient to use and thus the *fastest* overall!

This is a collection of hints and pointers for hassle-free RDF coding.

## Functional properties

Use [`value()`][rdflib.graph.Graph.value] and [`set()`][rdflib.graph.Graph.set] to work with *functional property* instances, i.e. properties than can only occur once for a resource.

```python
from rdflib import Graph, URIRef, Literal, BNode
from rdflib.namespace import FOAF, RDF

g = Graph()
g.bind("foaf", FOAF)

# Add demo data
bob = URIRef("http://example.org/people/Bob")
g.add((bob, RDF.type, FOAF.Person))
g.add((bob, FOAF.name, Literal("Bob")))
g.add((bob, FOAF.age, Literal(38)))

# To get a single value, use 'value'
print(g.value(bob, FOAF.age))
# prints: 38

# To change a single of value, use 'set'
g.set((bob, FOAF.age, Literal(39)))
print(g.value(bob, FOAF.age))
# prints: 39
```

## Slicing graphs

Python allows slicing arrays with a `slice` object, a triple of `start`, `stop` and `step-size`:

```python
for i in range(20)[2:9:3]:
    print(i)
# prints:
# 2, 5, 8
```

RDFLib graphs override `__getitem__` and we pervert the slice triple to be a RDF triple instead. This lets slice syntax be a shortcut for [`triples()`][rdflib.graph.Graph.triples], [`subject_predicates()`][rdflib.graph.Graph.subject_predicates], [`__contains__()`][rdflib.graph.Graph.__contains__], and other Graph query-methods:

```python
from rdflib import Graph, URIRef, Literal, BNode
from rdflib.namespace import FOAF, RDF

g = Graph()
g.bind("foaf", FOAF)

# Add demo data
bob = URIRef("http://example.org/people/Bob")
bill = URIRef("http://example.org/people/Bill")
g.add((bob, RDF.type, FOAF.Person))
g.add((bob, FOAF.name, Literal("Bob")))
g.add((bob, FOAF.age, Literal(38)))
g.add((bob, FOAF.knows, bill))

print(g[:])
# same as
print(iter(g))

print(g[bob])
# same as
print(g.predicate_objects(bob))

print(g[bob: FOAF.knows])
# same as
print(g.objects(bob, FOAF.knows))

print(g[bob: FOAF.knows: bill])
# same as
print((bob, FOAF.knows, bill) in g)

print(g[:FOAF.knows])
# same as
print(g.subject_objects(FOAF.knows))
```

See [`examples.slice`][examples.slice] for a complete example.

!!! warning "Slicing Caution"
    Slicing is convenient for run-once scripts for playing around
    in the Python `REPL`, however since slicing returns
    tuples of varying length depending on which parts of the
    slice are bound, you should be careful using it in more
    complicated programs. If you pass in variables, and they are
    `None` or `False`, you may suddenly get a generator of
    different length tuples back than you expect.

## SPARQL Paths

[SPARQL property paths](http://www.w3.org/TR/sparql11-property-paths/) are possible using overridden operators on URIRefs. See [`examples.foafpaths`][examples.foafpaths] and [`rdflib.paths`][rdflib.paths].

## Serializing a single term to N3

For simple output, or simple serialisation, you often want a nice
readable representation of a term. All terms (URIRef, Literal etc.) have a
`n3`, method, which will return a suitable N3 format:

```python
from rdflib import Graph, URIRef, Literal
from rdflib.namespace import FOAF

# A URIRef
person = URIRef("http://xmlns.com/foaf/0.1/Person")
print(person.n3())
# prints: <http://xmlns.com/foaf/0.1/Person>

# Simplifying the output with a namespace prefix:
g = Graph()
g.bind("foaf", FOAF)

print(person.n3(g.namespace_manager))
# prints foaf:Person

# A typed literal
l = Literal(2)
print(l.n3())
# prints "2"^^<http://www.w3.org/2001/XMLSchema#integer>

# Simplifying the output with a namespace prefix
# XSD is built in, so no need to bind() it!
l.n3(g.namespace_manager)
# prints: "2"^^xsd:integer
```

## Parsing data from a string

You can parse data from a string with the `data` param:

```python
from rdflib import Graph

g = Graph().parse(data="<a:> <p:> <p:>.")
for r in g.triples((None, None, None)):
    print(r)
# prints: (rdflib.term.URIRef('a:'), rdflib.term.URIRef('p:'), rdflib.term.URIRef('p:'))
```

## Command Line tools

RDFLib includes a handful of commandline tools, see [`rdflib.tools`][rdflib.tools].

---
title: "Working with Database Storage Backends in `rdflib`"
author: "Carl Boettiger"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Vignette Title}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  cache = TRUE
)
```


By default, `rdflib` stores triples in memory, which can be read in from and written out to files on disk using a variety of serializations (`rdfxml`, `json-ld`, `ntriples`, `nquads` or `turtle`).  This is analgous to reading in a `data.frame` object from a `csv`, `tsv`, or other text-delimited file, which requires the full data object to be read into memory before a user can manipulate it.  This approach runs into limitations with very large datasets, which may exceed the memory available.  Just as R packages such as `dplyr` offer the ability to perform manipulations with very large data through a connection to a database backend such as SQLite, MySQL or PostgreSQL, `rdflib` can rely on these and other databases to provide a disk-based backend. 

### Installation

Unfortunately, at this time, support for these storage devices is not included in the prebuild Mac and Windows binaries of the `redland` R package.  Users wanting to take advantage of disk-based storage must instead build and install the `redland` R package from source; and in some cases, also build the `redland` C libraries from source.  This package provides a [Dockerfile](https://github.com/ropensci/rdflib/blob/master/inst/docker/Dockerfile) containing a portable recipe for building the library with support for the 5 main backend storage devices: SQLite, MySQL, PostgreSQL, Virtuoso, and Berkeley DB.  This vignette documents the installation and use of these devices.  

Also see the [official documentation](http://librdf.org/docs/api/redland-storage-modules.html) for the redland C library discussing all supported storage devices.  


# Benchmarks

Here we show examples of reading in a modest set of 190,000 triples from an `nquads` file and executing a simple SPARQL query using each the five backends supported by `rdflib`.  First we load the libraries and prepare an example file by triple-ifying 10,000 rows of the `flights` data.frame:

```{r}
library(rdflib)
library(nycflights13)
```

```{r}
example <- flights[1e4,]

system.time(
write_nquads(example, "example.nq", prefix = "flights")
)

system.time(
write_nquads(flights, "flights.nq", prefix = "flights")
)
```

## In Memory

Because the dataset is small enough to easily fit in memory, the default in-memory option is an obvious choice with excellent performance.  Note that this option will not be possible with larger triplestores (e.g. with millions or more triples).  Our testing has found that even on machines with 100GB+ of RAM, the redland in-memory backend is not able to take advantage of that additional memory and disk-based backends are required. 

```{r}
triplestore <- rdf()
```

```{r}

system.time(
  read_nquads("example.nq", rdf = triplestore) # smaller set
)

```

```{r}
query <- 
  'SELECT  ?carrier ?flight ?origin ?dep_delay
WHERE {
?flight <flights:carrier>  ?carrier .
?flight <flights:dep_delay>  ?dep_delay .
?flight <flights:origin> ?origin
}'
```

```{r}
system.time(
df <- rdf_query(triplestore, query)
)
```

```{r}
rdf_free(triplestore)
```


## BDB

The [Berkeley DB](https://en.wikipedia.org/wiki/Berkeley_DB) is a simple key-value store 

> It is the most mature and primary persistent store and suitable for large models, tested in the 2-3 million range.

Berkeley DB is a simple disk-based storage option.  Install both the redland libraries and the berkeley-db (e.g. `bd-dev` on Debian/Ubuntu) libraries, and then install `redland` from source.  BDB is relatively fast for data too large for memory.


```{r}
triplestore <- rdf(storage="BDB", new_db = TRUE)
system.time(
  read_nquads("flights.nq", rdf = triplestore)
)
```


```{r}
system.time(
df <- rdf_query(triplestore, query)
)
```

```{r}
rdf_free(triplestore)
## Becuase BDB is just a hash table, redland needs three separate files:
unlink("rdflib-po2s.db")
unlink("rdflib-so2p.db")
unlink("rdflib-sp2o.db")
```

## Virtuoso

Unlike the other backends which use general purpose relational databases or key-value stores, [Virtuoso](https://en.wikipedia.org/wiki/Virtuoso_Universal_Server) dedicated to RDF-based data.  Virtuoso is a popular open source database for RDF with a rich set of built-in interfaces and features, but we can also interact with it directly through the `redland` bindings just like we do any other backend in `rdflib`.  Virtuoso setup may be slightly more involved for individuals unfamiliar with it, but will probably provide the best performance in the case of very large triplestores.  The example below shows a new database setup, but `rdflib` can also connect to any existing Virtuoso database. 

```{r}
triplestore <- rdf(storage = "virtuoso", 
                   user = "dba", 
                   password = "dba", 
                   dsn = "Local Virtuoso"
                   )

system.time(
  read_nquads("flights.nq", rdf = triplestore)
)
```


```{r}
system.time(
df <- rdf_query(triplestore, query)
)
df <- rdf_query(triplestore, "SELECT ?s ?p ?o WHERE{ ?s ?p ?o }")
df
```


```{r}
rdf_free(triplestore)
```

Or on remote virtuoso:


```{r}
triplestore <- rdf(storage = "virtuoso", 
                   user = "dba", 
                   password = "dba", 
                   host = "virtuoso:1111"
                   )

system.time(
  read_nquads("flights.nq", rdf = triplestore)
)
```




## POSTGRES

Postgres and MySQL are ubiquitious relational databases.  This backend requires the `redland` binaries are built from source with this support enabled, which is not the case for pre-built Mac or Linux binaries.   

```{r}
triplestore <- rdf(storage = "postgres", 
                   host = "postgres", 
                   user = "postgres", 
                   password = "rdflib", 
                   new_db = TRUE)
system.time(
  read_nquads("flights.nq", rdf = triplestore)
)
```



```{r}
system.time(
df <- rdf_query(triplestore, query)
)
```

```{r}
rdf_free(triplestore)
```  

## MySQL

```{r}
triplestore <- rdf(storage = "mysql", 
                   host = "mariadb", 
                   user = "root", 
                   password = "rdflib", 
                   database = "mysql",
                   new_db = TRUE
                  )
  read_nquads("flights.nq", rdf = triplestore)
 
```


```{r}
system.time(df <- rdf_query(triplestore, query))
```

```{r}
rdf_free(triplestore)

```




## SQLite


[SQLite](https://en.wikipedia.org/wiki/SQLite) is relatively easy to set up, but appears to have rather poor overall performance. Requires SQLite development libraries installed (should work 'out-of-the-box' with Mac binaries for `redland` package).  


```{r}
triplestore <- rdf(storage="sqlite", new_db = TRUE, name="rdflib.sqlite")

system.time(
  read_nquads("flights.nq", rdf = triplestore)
)
```


```{r}
system.time(
df <- rdf_query(triplestore, query)
)
```

```{r}
rdf_free(triplestore)
```

# Building redland with full backend support

Getting full support for the above backend databases through the `redland` R package is not trivial.  The `redland` R package provides bindings to the redland libraries.  Unfortunately, commonly available binary versions of those libraries, such as `librdf0-dev` on Debian, `redland` on Mac OSX brew, and the statically linked versions for Mac and Windows shipping in the R package, do not build those libraries with the optional support for all backends. (NOTE: it is the C library itself which must be built from source with these options, not just the R package source).  Consequently, users must build `librdf` from the original sources, <https://github.com/dajobe/librdf> with all backend linking libraries available, and then also build the `redland` R package from source, to be able to access all of these backends. On a Debian or Ubuntu system this looks like the following:

```bash
apt-get update && apt-get -yq install \
libxml2-dev \
libcurl4-openssl-dev \
libssl-dev \
git \
automake \
libtool \
gtk-doc-tools \
bison \
flex \
libgmp-dev  \
libmhash-dev \
libgcrypt20-dev \
libpcre3-dev \
libv8-dev \
libjq-dev \
libpq-dev \
libdb-dev \
libsqlite3-dev \
libmariadbclient-dev \
librdf-storage-virtuoso \
virtuoso-server \
unixodbc-dev
```

Now we can build `raptor` (parsers), `rasqal` (sparql queries) and `rdflib` from source:

```
git clone git://github.com/dajobe/raptor.git && \
cd raptor && \
./autogen.sh && \
make && \
make install && \
cd .. && \
git clone git://github.com/dajobe/rasqal.git && \
cd rasqal && \
./autogen.sh && \
make && \
make install && \
cd .. && \
git clone git://github.com/dajobe/librdf.git && \
cd librdf && \
./autogen.sh && \
make && \
make install
```

A See the [Dockerfile](https://github.com/ropensci/rdflib/tree/master/inst/docker/Dockerfile) in `inst/docker` for an example of this, or simply use the [Rocker-based](https://rocker-project.org) image `ropensci/rdflib`.  

# Testing

`rdflib` uses Circle-CI to test database backends using `docker-compose`.  `docker-compose` pulls down dedicated docker containers for `postgres` and `mariadb`, along with the `ropensci/rdflib` container, which includes a version of `redland` compiled with support for all five major backend storage systems.  See the [Dockerfile](https://github.com/ropensci/rdflib/tree/master/inst/docker/Dockerfile) in `inst/docker` and associated [docker-compose.yml]((https://github.com/ropensci/rdflib/tree/master/docker-compose.yml) used in testing for an example of this configuration.  You can also pull the Docker image `ropensci/rdflib` from Docker Hub for testing with all these libraries installed.     





The badge below (also on the package README) indicates that these dedicated tests are passing.

[![CircleCI](https://circleci.com/gh/ropensci/rdflib.svg?style=svg)](https://circleci.com/gh/ropensci/rdflib)

*[HTML]: Hyper Text Markup Language
*[HTTP]: HyperText Transfer Protocol
*[HTTPS]: HyperText Transfer Protocol Secure
*[API]: Application Programming Interface
*[UI]: User Interface
*[CLI]: Command-Line Interface
*[PIP]: Pip Install Packages
*[PyPI]: Python Packaging Index
*[PyPA]: Python Packaging Authority
*[PEP]: Python Enhancement Proposal
*[RDF]: Resource Description Framework
*[N3]: Notation 3, an assertion and logic language which is a superset of RDF
*[TriX]: Triples in XML
*[TriG]: Triples in Graphs
*[RDFa]: Resource Description Framework in Attributes
*[JSON-LD]: JavaScript Object Notation - Linked Data
*[JSON]: JavaScript Object Notation
*[OWL]: Web Ontology Language
*[XML]: Extensible Markup Language
*[SPARQL]: SPARQL Protocol and RDF Query Language
*[URL]: Uniform Resource Locator
*[URI]: Uniform Resource Identifier
*[IRI]: Internationalized Resource Identifier
*[CSV]: Comma-Separated Value
*[TSV]: Tab-Separated Value
*[PSV]: Pipe-Separated Value
*[RegEx]: Regular Expression
*[OBO]: Open Biological and Biomedical Ontology
*[VSCode]: VisualStudio Code
*[PR]: Pull request
*[PRs]: Pull requests

# Getting started with RDFLib

## Installation

RDFLib is open source and is maintained in a [GitHub](https://github.com/RDFLib/rdflib/) repository. RDFLib releases, current and previous, are listed on [PyPI](https://pypi.python.org/pypi/rdflib/)

The best way to install RDFLib is to use `pip` (sudo as required):

```bash
pip install rdflib
```

If you want the latest code to run, clone the `main` branch of the GitHub repo and use that or you can `pip install` directly from GitHub:

```bash
pip install git+https://github.com/RDFLib/rdflib.git@main#egg=rdflib
```

## Support

Usage support is available via questions tagged with `[rdflib]` on [StackOverflow](https://stackoverflow.com/questions/tagged/rdflib) and development support, notifications and detailed discussion through the rdflib-dev group (mailing list): [http://groups.google.com/group/rdflib-dev](http://groups.google.com/group/rdflib-dev)

If you notice a bug or want to request an enhancement, please do so via our Issue Tracker in Github: [http://github.com/RDFLib/rdflib/issues](http://github.com/RDFLib/rdflib/issues)

## How it all works

*The package uses various Python idioms that offer an appropriate way to introduce RDF to a Python programmer who hasn't worked with RDF before.*

The primary interface that RDFLib exposes for working with RDF is a [`Graph`][rdflib.graph.Graph].

RDFLib graphs are un-sorted containers; they have ordinary Python `set` operations (e.g. [`add()`][rdflib.graph.Graph.add] to add a triple) plus methods that search triples and return them in arbitrary order.

RDFLib graphs also redefine certain built-in Python methods in order to behave in a predictable way. They do this by [emulating container types](https://docs.python.org/3.8/reference/datamodel.html#emulating-container-types) and are best thought of as a set of 3-item tuples ("triples", in RDF-speak):

```python
[
    (subject0, predicate0, object0),
    (subject1, predicate1, object1),
    # ...
    (subjectN, predicateN, objectN),
]
```

## A tiny example

```python
from rdflib import Graph

# Create a Graph
g = Graph()

# Parse in an RDF file hosted on the Internet
g.parse("http://www.w3.org/People/Berners-Lee/card")

# Loop through each triple in the graph (subj, pred, obj)
for subj, pred, obj in g:
    # Check if there is at least one triple in the Graph
    if (subj, pred, obj) not in g:
       raise Exception("It better be!")

# Print the number of "triples" in the Graph
print(f"Graph g has {len(g)} statements.")
# Prints: Graph g has 86 statements.

# Print out the entire Graph in the RDF Turtle format
print(g.serialize(format="turtle"))
```

Here a [`Graph`][rdflib.graph.Graph] is created and then an RDF file online, Tim Berners-Lee's social network details, is parsed into that graph. The `print()` statement uses the `len()` function to count the number of triples in the graph.

## A more extensive example

```python
from rdflib import Graph, Literal, RDF, URIRef
# rdflib knows about quite a few popular namespaces, like W3C ontologies, schema.org etc.
from rdflib.namespace import FOAF , XSD

# Create a Graph
g = Graph()

# Create an RDF URI node to use as the subject for multiple triples
donna = URIRef("http://example.org/donna")

# Add triples using store's add() method.
g.add((donna, RDF.type, FOAF.Person))
g.add((donna, FOAF.nick, Literal("donna", lang="en")))
g.add((donna, FOAF.name, Literal("Donna Fales")))
g.add((donna, FOAF.mbox, URIRef("mailto:donna@example.org")))

# Add another person
ed = URIRef("http://example.org/edward")

# Add triples using store's add() method.
g.add((ed, RDF.type, FOAF.Person))
g.add((ed, FOAF.nick, Literal("ed", datatype=XSD.string)))
g.add((ed, FOAF.name, Literal("Edward Scissorhands")))
g.add((ed, FOAF.mbox, Literal("e.scissorhands@example.org", datatype=XSD.anyURI)))

# Iterate over triples in store and print them out.
print("--- printing raw triples ---")
for s, p, o in g:
    print((s, p, o))

# For each foaf:Person in the store, print out their mbox property's value.
print("--- printing mboxes ---")
for person in g.subjects(RDF.type, FOAF.Person):
    for mbox in g.objects(person, FOAF.mbox):
        print(mbox)

# Bind the FOAF namespace to a prefix for more readable output
g.bind("foaf", FOAF)

# print all the data in the Notation3 format
print("--- printing mboxes ---")
print(g.serialize(format='n3'))
```

## A SPARQL query example

```python
from rdflib import Graph

# Create a Graph, parse in Internet data
g = Graph().parse("http://www.w3.org/People/Berners-Lee/card")

# Query the data in g using SPARQL
# This query returns the 'name' of all `foaf:Person` instances
q = """
    PREFIX foaf: <http://xmlns.com/foaf/0.1/>

    SELECT ?name
    WHERE {
        ?p rdf:type foaf:Person .

        ?p foaf:name ?name .
    }
"""

# Apply the query to the graph and iterate through results
for r in g.query(q):
    print(r["name"])

# prints: Timothy Berners-Lee
```

# Navigating Graphs

An RDF Graph is a set of RDF triples, and we try to mirror exactly this in RDFLib. The Python [`Graph`][rdflib.graph.Graph] tries to emulate a container type.

## Graphs as Iterators

RDFLib graphs override [`__iter__()`][rdflib.graph.Graph.__iter__] in order to support iteration over the contained triples:

```python
for s, p, o in someGraph:
    if not (s, p, o) in someGraph:
        raise Exception("Iterator / Container Protocols are Broken!!")
```

This loop iterates through all the subjects(s), predicates (p) & objects (o) in `someGraph`.

## Contains check

Graphs implement [`__contains__()`][rdflib.graph.Graph.__contains__], so you can check if a triple is in a graph with a `triple in graph` syntax:

```python
from rdflib import URIRef
from rdflib.namespace import RDF

bob = URIRef("http://example.org/people/bob")
if (bob, RDF.type, FOAF.Person) in graph:
    print("This graph knows that Bob is a person!")
```

Note that this triple does not have to be completely bound:

```python
if (bob, None, None) in graph:
    print("This graph contains triples about Bob!")
```

## Set Operations on RDFLib Graphs

Graphs override several pythons operators: [`__iadd__()`][rdflib.graph.Graph.__iadd__], [`__isub__()`][rdflib.graph.Graph.__isub__], etc. This supports addition, subtraction and other set-operations on Graphs:

| operation | effect |
|-----------|--------|
| `G1 + G2` | return new graph with union (triples on both) |
| `G1 += G2` | in place union / addition |
| `G1 - G2` | return new graph with difference (triples in G1, not in G2) |
| `G1 -= G2` | in place difference / subtraction |
| `G1 & G2` | intersection (triples in both graphs) |
| `G1 ^ G2` | xor (triples in either G1 or G2, but not in both) |

!!! warning
    Set-operations on graphs assume Blank Nodes are shared between graphs. This may or may not be what you want. See [merging](merging.md) for details.

## Basic Triple Matching

Instead of iterating through all triples, RDFLib graphs support basic triple pattern matching with a [`triples()`][rdflib.graph.Graph.triples] function. This function is a generator of triples that match a pattern given by arguments, i.e. arguments restrict the triples that are returned. Terms that are `None` are treated as a wildcard. For example:

```python
g.parse("some_foaf.ttl")
# find all subjects (s) of type (rdf:type) person (foaf:Person)
for s, p, o in g.triples((None, RDF.type, FOAF.Person)):
    print(f"{s} is a person")

# find all subjects of any type
for s, p, o in g.triples((None,  RDF.type, None)):
    print(f"{s} is a {o}")

# create a graph
bobgraph = Graph()
# add all triples with subject 'bob'
bobgraph += g.triples((bob, None, None))
```

If you are not interested in whole triples, you can get only the bits you want with the methods [`objects()`][rdflib.graph.Graph.objects], [`subjects()`][rdflib.graph.Graph.subjects], [`predicates()`][rdflib.graph.Graph.predicates], [`predicate_objects()`][rdflib.graph.Graph.predicate_objects], etc. Each take parameters for the components of the triple to constraint:

```python
for person in g.subjects(RDF.type, FOAF.Person):
    print("{} is a person".format(person))
```

Finally, for some properties, only one value per resource makes sense (i.e they are *functional properties*, or have a max-cardinality of 1). The [`value()`][rdflib.graph.Graph.value] method is useful for this, as it returns just a single node, not a generator:

```python
# get any name of bob
name = g.value(bob, FOAF.name)
# get the one person that knows bob and raise an exception if more are found
person = g.value(predicate=FOAF.knows, object=bob, any=False)
```


## Graph methods for accessing triples

Here is a list of all convenience methods for querying Graphs:

* [`triples()`][rdflib.graph.Graph.triples]
* [`value()`][rdflib.graph.Graph.value]
* [`subjects()`][rdflib.graph.Graph.subjects]
* [`objects()`][rdflib.graph.Graph.objects]
* [`predicates()`][rdflib.graph.Graph.predicates]
* [`subject_objects()`][rdflib.graph.Graph.subject_objects]
* [`subject_predicates()`][rdflib.graph.Graph.subject_predicates]
* [`predicate_objects()`][rdflib.graph.Graph.predicate_objects]

# RDF terms in rdflib

Terms are the kinds of objects that can appear in a RDFLib's graph's triples. Those that are part of core RDF concepts are: `IRIs`, `Blank Node` and `Literal`, the latter consisting of a literal value and either a [datatype](https://www.w3.org/TR/xmlschema-2/#built-in-datatypes) or an [RFC 3066](https://tools.ietf.org/html/rfc3066) language tag.

!!! info "Origins"
    RDFLib's class for representing IRIs/URIs is called "URIRef" because, at the time it was implemented, that was what the then current RDF specification called URIs/IRIs. We preserve that class name but refer to the RDF object as "IRI".

## Class hierarchy

All terms in RDFLib are sub-classes of the [`Identifier`][rdflib.term.Identifier] class. A class diagram of the various terms is:

![Term Class Hierarchy](_static/term_class_hierarchy.svg)

Nodes are a subset of the Terms that underlying stores actually persist.

The set of such Terms depends on whether or not the store is formula-aware. Stores that aren't formula-aware only persist those terms core to the RDF Model but those that are formula-aware also persist the N3 extensions. However, utility terms that only serve the purpose of matching nodes by term-patterns will probably only be terms and not nodes.

## Python Classes

The three main RDF objects - *IRI*, *Blank Node* and *Literal* are represented in RDFLib by these three main Python classes:

### URIRef

An IRI (Internationalized Resource Identifier) is represented within RDFLib using the [`URIRef class`][rdflib.term.URIRef]. From [the RDF 1.1 specification's IRI section](https://www.w3.org/TR/rdf11-concepts/#section-IRIs):

```python
>>> from rdflib import URIRef
>>> uri = URIRef()  # doctest: +SKIP
Traceback (most recent call last):
  File "<stdin>", line 1, in <module>
TypeError: __new__() missing 1 required positional argument: 'value'
>>> uri = URIRef('')
>>> uri
rdflib.term.URIRef('')
>>> uri = URIRef('http://example.com')
>>> uri
rdflib.term.URIRef('http://example.com')
>>> uri.n3()
'<http://example.com>'
```

### BNodes

In RDF, a blank node (also called BNode) is a node in an RDF graph representing a resource for which an IRI or literal is not given. The resource represented by a blank node is also called an anonymous resource. According to the RDF standard, a blank node can only be used as subject or object in a triple, although in some syntaxes like Notation 3 it is acceptable to use a blank node as a predicate. If a blank node has a node ID (not all blank nodes are labelled in all RDF serializations), it is limited in scope to a particular serialization of the RDF graph, i.e. the node p1 in one graph does not represent the same node as a node named p1 in any other graph  -- [wikipedia](http://en.wikipedia.org/wiki/Blank_node)

See the [`BNode`][rdflib.term.BNode] class' documentation.

```python
>>> from rdflib import BNode
>>> bn = BNode()
>>> bn  # doctest: +SKIP
rdflib.term.BNode('AFwALAKU0')
>>> bn.n3()  # doctest: +SKIP
'_:AFwALAKU0'
```

### Literals

Literals are attribute values in RDF, for instance, a person's name, the date of birth, height, etc. and are stored using simple data types, e.g. *string*, *double*, *dateTime* etc. This usually looks something like this:

```python
name = Literal("Nicholas")  # the name 'Nicholas', as a string

age = Literal(39, datatype=XSD.integer)  # the number 39, as an integer
```

A slightly special case is a *langString* which is a *string* with a language tag, e.g.:

```python
name = Literal("Nicholas", lang="en")  # the name 'Nicholas', as an English string
imie = Literal("Mikołaj", lang="pl")  # the Polish version of the name 'Nicholas'
```

Special literal types indicated by use of a custom IRI for a literal's `datatype` value, for example the [GeoSPARQL RDF standard](https://opengeospatial.github.io/ogc-geosparql/geosparql11/spec.html#_geometry_serializations) invents a custom datatype, `geoJSONLiteral` to indicate [GeoJSON geometry serlializations](https://opengeospatial.github.io/ogc-geosparql/geosparql11/spec.html#_rdfs_datatype_geogeojsonliteral) like this:

```python
GEO = Namespace("http://www.opengis.net/ont/geosparql#")

geojson_geometry = Literal(
    '''{"type": "Point", "coordinates": [-83.38,33.95]}''',
    datatype=GEO.geoJSONLiteral
```

See the [`Literal`][rdflib.term.Literal] class' documentation, followed by notes on Literal from the [RDF 1.1 specification 'Literals' section](https://www.w3.org/TR/rdf11-concepts/#section-Graph-Literal).

A literal in an RDF graph contains one or two named components.

All literals have a lexical form being a Unicode string, which SHOULD be in Normal Form C.

Plain literals have a lexical form and optionally a language tag as defined by [RFC 3066](https://tools.ietf.org/html/rfc3066), normalized to lowercase. An exception will be raised if illegal language-tags are passed to [\_\_new\_\_()][rdflib.term.Literal.__new__].

Typed literals have a lexical form and a datatype URI being an RDF URI reference.

!!! abstract "Language vs. locale"
    When using the language tag, care must be taken not to confuse language with locale. The language tag relates only to human language text. Presentational issues should be addressed in end-user applications.

!!! quote "Case sensitive"
    The case normalization of language tags is part of the description of the abstract syntax, and consequently the abstract behaviour of RDF applications. It does not constrain an RDF implementation to actually normalize the case. Crucially, the result of comparing two language tags should not be sensitive to the case of the original input. -- [RDF Concepts and Abstract Syntax](http://www.w3.org/TR/rdf-concepts/#section-Graph-URIref)

#### Common XSD datatypes

Most simple literals such as *string* or *integer* have XML Schema (XSD) datatypes defined for them, see the figure below. Additionally, these XSD datatypes are listed in the [XSD Namespace class][rdflib.namespace.XSD] that ships with RDFLib, so many Python code editors will prompt you with autocomplete for them when using it.

Remember, you don't *have* to use XSD datatypes and can always make up your own, as GeoSPARQL does, as described above.

![datatype hierarchy](_static/datatype_hierarchy.png)

#### Python conversions

RDFLib Literals essentially behave like unicode characters with an XML Schema datatype or language attribute.

The class provides a mechanism to both convert Python literals (and their built-ins such as time/date/datetime) into equivalent RDF Literals and (conversely) convert Literals to their Python equivalent. This mapping to and from Python literals is done as follows:

| XML Datatype | Python type |
|--------------|-------------|
| None | None [^1] |
| xsd:time | time [^2] |
| xsd:date | date |
| xsd:dateTime | datetime |
| xsd:string | None |
| xsd:normalizedString | None |
| xsd:token | None |
| xsd:language | None |
| xsd:boolean | boolean |
| xsd:decimal | Decimal |
| xsd:integer | long |
| xsd:nonPositiveInteger | int |
| xsd:long | long |
| xsd:nonNegativeInteger | int |
| xsd:negativeInteger | int |
| xsd:int | long |
| xsd:unsignedLong | long |
| xsd:positiveInteger | int |
| xsd:short | int |
| xsd:unsignedInt | long |
| xsd:byte | int |
| xsd:unsignedShort | int |
| xsd:unsignedByte | int |
| xsd:float | float |
| xsd:double | float |
| xsd:base64Binary | base64 |
| xsd:anyURI | None |
| rdf:XMLLiteral | Document (xml.dom.minidom.Document [^3] |
| rdf:HTML | DocumentFragment (xml.dom.minidom.DocumentFragment) |

[^1]: plain literals map directly to value space
[^2]: Date, time and datetime literals are mapped to Python instances using the RDFlib xsd_datetime module, that is based on the [isodate](http://pypi.python.org/pypi/isodate/) package).
[^3]: this is a bit dirty - by accident the `html5lib` parser produces `DocumentFragments`, and the xml parser `Documents`, letting us use this to decide what datatype when round-tripping.

An appropriate data-type and lexical representation can be found using `_castPythonToLiteral`, and the other direction with `_castLexicalToPython`.

All this happens automatically when creating `Literal` objects by passing Python objects to the constructor, and you never have to do this manually.

You can add custom data-types with [`bind()`][rdflib.term.bind], see also [`custom_datatype example`][examples.custom_datatype]

# Loading and saving RDF

## Reading RDF files

RDF data can be represented using various syntaxes (`turtle`, `rdf/xml`, `n3`, `n-triples`, `trix`, `JSON-LD`, etc.). The simplest format is `ntriples`, which is a triple-per-line format.

Create the file `demo.nt` in the current directory with these two lines in it:

```turtle
<http://example.com/drewp> <http://www.w3.org/1999/02/22-rdf-syntax-ns#type> <http://xmlns.com/foaf/0.1/Person> .
<http://example.com/drewp> <http://example.com/says> "Hello World" .
```

On line 1 this file says "drewp is a FOAF Person:. On line 2 it says "drep says "Hello World"".

RDFLib can guess what format the file is by the file ending (".nt" is commonly used for n-triples) so you can just use [`parse()`][rdflib.graph.Graph.parse] to read in the file. If the file had a non-standard RDF file ending, you could set the keyword-parameter `format` to specify either an Internet Media Type or the format name (a [list of available parsers][rdflib.plugins.parsers] is available).

In an interactive python interpreter, try this:

```python
from rdflib import Graph

g = Graph()
g.parse("demo.nt")

print(len(g))
# prints: 2

import pprint
for stmt in g:
    pprint.pprint(stmt)
# prints:
# (rdflib.term.URIRef('http://example.com/drewp'),
#  rdflib.term.URIRef('http://example.com/says'),
#  rdflib.term.Literal('Hello World'))
# (rdflib.term.URIRef('http://example.com/drewp'),
#  rdflib.term.URIRef('http://www.w3.org/1999/02/22-rdf-syntax-ns#type'),
#  rdflib.term.URIRef('http://xmlns.com/foaf/0.1/Person'))
```

The final lines show how RDFLib represents the two statements in the file: the statements themselves are just length-3 tuples ("triples") and the subjects, predicates, and objects of the triples are all rdflib types.

## Reading remote RDF

Reading graphs from the Internet is easy:

```python
from rdflib import Graph

g = Graph()
g.parse("http://www.w3.org/People/Berners-Lee/card")
print(len(g))
# prints: 86
```

[`parse()`][rdflib.Graph.parse] can process local files, remote data via a URL, as in this example, or RDF data in a string (using the `data` parameter).

## Saving RDF

To store a graph in a file, use the [`serialize()`][rdflib.Graph.serialize] function:

```python
from rdflib import Graph

g = Graph()
g.parse("http://www.w3.org/People/Berners-Lee/card")
g.serialize(destination="tbl.ttl")
```

This parses data from http://www.w3.org/People/Berners-Lee/card and stores it in a file `tbl.ttl` in this directory using the turtle format, which is the default RDF serialization (as of rdflib 6.0.0).

To read the same data and to save it as an RDF/XML format string in the variable `v`, do this:

```python
from rdflib import Graph

g = Graph()
g.parse("http://www.w3.org/People/Berners-Lee/card")
v = g.serialize(format="xml")
```

The following table lists the RDF formats you can serialize data to with rdflib, out of the box, and the `format=KEYWORD` keyword used to reference them within `serialize()`:

| RDF Format | Keyword | Notes |
|------------|---------|-------|
| Turtle | turtle, ttl or turtle2 | turtle2 is just turtle with more spacing & linebreaks |
| RDF/XML | xml or pretty-xml | Was the default format, rdflib < 6.0.0 |
| JSON-LD | json-ld | There are further options for compact syntax and other JSON-LD variants |
| N-Triples | ntriples, nt or nt11 | nt11 is exactly like nt, only utf8 encoded |
| Notation-3 | n3 | N3 is a superset of Turtle that also caters for rules and a few other things |
| Trig | trig | Turtle-like format for RDF triples + context (RDF quads) and thus multiple graphs |
| Trix | trix | RDF/XML-like format for RDF quads |
| N-Quads | nquads | N-Triples-like format for RDF quads |

## Working with multi-graphs

To read and query multi-graphs, that is RDF data that is context-aware, you need to use rdflib's [`Dataset`][rdflib.Dataset] class. This an extension to [`Graph`][rdflib.Graph] that know all about quads (triples + graph IDs).

If you had this multi-graph data file (in the `trig` format, using new-style `PREFIX` statement (not the older `@prefix`):

```turtle
PREFIX eg: <http://example.com/person/>
PREFIX foaf: <http://xmlns.com/foaf/0.1/>

eg:graph-1 {
    eg:drewp a foaf:Person .
    eg:drewp eg:says "Hello World" .
}

eg:graph-2 {
    eg:nick a foaf:Person .
    eg:nick eg:says "Hi World" .
}
```

You could parse the file and query it like this:

```python
from rdflib import Dataset
from rdflib.namespace import RDF

g = Dataset()
g.parse("demo.trig")

for s, p, o, g in g.quads((None, RDF.type, None, None)):
    print(s, g)
```

This will print out:

```
http://example.com/person/drewp http://example.com/person/graph-1
http://example.com/person/nick http://example.com/person/graph-2
```

# Plugins

![rdflib plugin "architecture"](_static/plugins-diagram.svg)

Many parts of RDFLib are extensible with plugins, [see setuptools' 'Creating and discovering plugins'](https://packaging.python.org/guides/creating-and-discovering-plugins/). These pages list the plugins included in RDFLib core.

* [`Parser Plugins`][rdflib.plugins.parsers]
* [`Serializer Plugins`][rdflib.plugins.serializers]
* [`Store Plugins`][rdflib.plugins.stores]
* [`Query Results Plugins`][rdflib.plugins.sparql.results]

## Plugin stores

### Built In

The following Stores are contained within the rdflib core package:

| Name | Class |
| --- | --- |
| Auditable | [`AuditableStore`][rdflib.plugins.stores.auditable.AuditableStore] |
| Concurrent | [`ConcurrentStore`][rdflib.plugins.stores.concurrent.ConcurrentStore] |
| SimpleMemory | [`SimpleMemory`][rdflib.plugins.stores.memory.SimpleMemory] |
| Memory | [`Memory`][rdflib.plugins.stores.memory.Memory] |
| SPARQLStore | [`SPARQLStore`][rdflib.plugins.stores.sparqlstore.SPARQLStore] |
| SPARQLUpdateStore | [`SPARQLUpdateStore`][rdflib.plugins.stores.sparqlstore.SPARQLUpdateStore] |
| BerkeleyDB | [`BerkeleyDB`][rdflib.plugins.stores.berkeleydb.BerkeleyDB] |
| default | [`Memory`][rdflib.plugins.stores.memory.Memory] |

### External

The following Stores are defined externally to rdflib's core package, so look to their documentation elsewhere for specific details of use.

| Name | Repository | Notes |
| --- | --- | --- |
| SQLAlchemy | [github.com/RDFLib/rdflib-sqlalchemy](https://github.com/RDFLib/rdflib-sqlalchemy) | An SQLAlchemy-backed, formula-aware RDFLib Store. Tested dialects are: SQLite, MySQL & PostgreSQL |
| leveldb | [github.com/RDFLib/rdflib-leveldb](https://github.com/RDFLib/rdflib-leveldb) | An adaptation of RDFLib BerkeleyDB Store's key-value approach, using LevelDB as a back-end |
| Kyoto Cabinet | [github.com/RDFLib/rdflib-kyotocabinet](https://github.com/RDFLib/rdflib-kyotocabinet) | An adaptation of RDFLib BerkeleyDB Store's key-value approach, using Kyoto Cabinet as a back-end |
| HDT | [github.com/RDFLib/rdflib-hdt](https://github.com/RDFLib/rdflib-hdt) | A Store back-end for rdflib to allow for reading and querying [HDT](https://www.rdfhdt.org/) documents |
| Oxigraph | [github.com/oxigraph/oxrdflib](https://github.com/oxigraph/oxrdflib) | Works with the [Pyoxigraph](https://oxigraph.org/pyoxigraph) Python graph database library |
| pycottas | [github.com/arenas-guerrero-julian/pycottas](https://github.com/arenas-guerrero-julian/pycottas) | A Store backend for querying compressed [COTTAS](https://pycottas.readthedocs.io/#cottas-files) files |

*If you have, or know of a Store implementation and would like it listed here, please submit a Pull Request!*

### Use

You can use these stores like this:

```python
from rdflib import Graph

# use the default memory Store
graph = Graph()

# use the BerkeleyDB Store
graph = Graph(store="BerkeleyDB")
```

In some cases, you must explicitly *open* and *close* a store, for example:

```python
from rdflib import Graph

# use the BerkeleyDB Store
graph = Graph(store="BerkeleyDB")
graph.open("/some/folder/location")
# do things ...
graph.close()
```

## Plugin parsers

These serializers are available in default RDFLib, you can use them by passing the name to graph's [`parse()`][rdflib.graph.Graph.parse] method:

```python
graph.parse(my_url, format='n3')
```

The `html` parser will auto-detect RDFa, HTurtle or Microdata.

It is also possible to pass a mime-type for the `format` parameter:

```python
graph.parse(my_url, format='application/rdf+xml')
```

If you are not sure what format your file will be, you can use [`guess_format()`][rdflib.util.guess_format] which will guess based on the file extension.

| Name    | Class                                                         |
|---------|---------------------------------------------------------------|
| json-ld | [`JsonLDParser`][rdflib.plugins.parsers.jsonld.JsonLDParser]  |
| hext    | [`HextuplesParser`][rdflib.plugins.parsers.hext.HextuplesParser] |
| n3      | [`N3Parser`][rdflib.plugins.parsers.notation3.N3Parser]       |
| nquads  | [`NQuadsParser`][rdflib.plugins.parsers.nquads.NQuadsParser]  |
| patch   | [`RDFPatchParser`][rdflib.plugins.parsers.patch.RDFPatchParser] |
| nt      | [`NTParser`][rdflib.plugins.parsers.ntriples.NTParser]        |
| trix    | [`TriXParser`][rdflib.plugins.parsers.trix.TriXParser]        |
| turtle  | [`TurtleParser`][rdflib.plugins.parsers.notation3.TurtleParser] |
| xml     | [`RDFXMLParser`][rdflib.plugins.parsers.rdfxml.RDFXMLParser]  |

### Multi-graph IDs

Note that for correct parsing of multi-graph data, e.g. TriG, HexTuple, etc., into a `Dataset`, as opposed to a context-unaware `Graph`, you will need to set the `publicID` of the `Dataset` to the identifier of the `default_context` (default graph), for example:

```python
d = Dataset()
d.parse(
    data=""" ... """,
    format="trig",
    publicID=d.default_context.identifier
)
```

(from the file tests/test_serializer_hext.py)

## Plugin serializers

These serializers are available in default RDFLib, you can use them by
passing the name to a graph's [`serialize()`][rdflib.graph.Graph.serialize] method:

```python
print graph.serialize(format='n3')
```

It is also possible to pass a mime-type for the `format` parameter:

```python
graph.serialize(my_url, format='application/rdf+xml')
```

| Name | Class |
|------|-------|
| json-ld | [`JsonLDSerializer`][rdflib.plugins.serializers.jsonld.JsonLDSerializer] |
| n3 | [`N3Serializer`][rdflib.plugins.serializers.n3.N3Serializer] |
| nquads | [`NQuadsSerializer`][rdflib.plugins.serializers.nquads.NQuadsSerializer] |
| nt | [`NTSerializer`][rdflib.plugins.serializers.nt.NTSerializer] |
| hext | [`HextuplesSerializer`][rdflib.plugins.serializers.hext.HextuplesSerializer] |
| patch | [`PatchSerializer`][rdflib.plugins.serializers.patch.PatchSerializer] |
| pretty-xml | [`PrettyXMLSerializer`][rdflib.plugins.serializers.rdfxml.PrettyXMLSerializer] |
| trig | [`TrigSerializer`][rdflib.plugins.serializers.trig.TrigSerializer] |
| trix | [`TriXSerializer`][rdflib.plugins.serializers.trix.TriXSerializer] |
| turtle | [`TurtleSerializer`][rdflib.plugins.serializers.turtle.TurtleSerializer] |
| longturtle | [`LongTurtleSerializer`][rdflib.plugins.serializers.longturtle.LongTurtleSerializer] |
| xml | [`XMLSerializer`][rdflib.plugins.serializers.rdfxml.XMLSerializer] |

### JSON-LD

JSON-LD - 'json-ld' - has been incorporated into RDFLib since v6.0.0.

### RDF Patch

The RDF Patch Serializer - 'patch' - uses the RDF Patch format defined at https://afs.github.io/rdf-patch/. It supports serializing context aware stores as either addition or deletion patches; and also supports serializing the difference between two context aware stores as a Patch of additions and deletions.

### HexTuples

The HexTuples Serializer - 'hext' - uses the HexTuples format defined at https://github.com/ontola/hextuples.

For serialization of non-context-aware data sources, e.g. a single `Graph`, the 'graph' field (6th variable in the Hextuple) will be an empty string.

For context-aware (multi-graph) serialization, the 'graph' field of the default graph will be an empty string and the values for other graphs will be Blank Node IDs or IRIs.

### Longturtle

Longturtle is just the turtle format with newlines preferred over compactness - multiple nodes on the same line to enhance the format's text file version control (think Git) friendliness - and more modern forms of prefix markers - PREFIX instead of @prefix - to make it as similar to SPARQL as possible.

Longturtle is Turtle 1.1 compliant and will work wherever ordinary turtle works, however some very old parsers don't understand PREFIX, only @prefix...

## Plugin query results

Plugins for reading and writing of (SPARQL) [`Result`][rdflib.query.Result] - pass `name` to either [`parse()`][rdflib.query.Result.parse] or [`serialize()`][rdflib.query.Result.serialize]

### Parsers

| Name | Class |
| ---- | ----- |
| csv  | [`CSVResultParser`][rdflib.plugins.sparql.results.csvresults.CSVResultParser] |
| json | [`JSONResultParser`][rdflib.plugins.sparql.results.jsonresults.JSONResultParser] |
| tsv  | [`TSVResultParser`][rdflib.plugins.sparql.results.tsvresults.TSVResultParser] |
| xml  | [`XMLResultParser`][rdflib.plugins.sparql.results.xmlresults.XMLResultParser] |

### Serializers

| Name | Class |
| ---- | ----- |
| csv  | [`CSVResultSerializer`][rdflib.plugins.sparql.results.csvresults.CSVResultSerializer] |
| json | [`JSONResultSerializer`][rdflib.plugins.sparql.results.jsonresults.JSONResultSerializer] |
| txt  | [`TXTResultSerializer`][rdflib.plugins.sparql.results.txtresults.TXTResultSerializer] |
| xml  | [`XMLResultSerializer`][rdflib.plugins.sparql.results.xmlresults.XMLResultSerializer] |

# Decision Records

To ensure that significant changes to RDFLib are made with sufficient consultation, consideration and planning they should be preceded by a decision record that captures the particulars of the decision that lead to the change.

Decision records present the users and maintainers of RDFLib with an opportunity to review decisions before effort is expended to implement the decision in code, and it also makes it possible to review decisions without having to reconstruct them from the code changes that implement them.

Whether a change is significant is hard to measure objectively, but some characteristics that may indicate that a change is significant include:

* It will require changes to code that use RDFLib.
* It cannot be reversed without requiring changes to code that use RDFLib.
* It is onerous to reverse later.
* It increases the maintenance burden of RDFLib.
* It is very large.

Some of these characteristics are not binary but measured in degrees, so some discretion is required when determining if an architectural decision record is appropriate.

Decision records may also be used for changes that do not have any of the listed characteristics if a decision record would be otherwise helpful, for example to capture a decision to change the maintenance process of RDFLib.

Changes not preceded by decision records won't be rejected solely on this basis even if they are deemed significant, and decision records may also be created retrospectively for changes.

Decision records as described here are similar to the concept of [Architectural Decision Records](https://adr.github.io/), though it is slightly broader as it could include decisions which are not classified as architectural.

## Creating a decision record

Decision records should be added to the RDFLib repository in the `./docs/decisions/` directory with a name `{YYYYmmdd}-{title}.md`.

The content of the decision record should succinctly describe the context of the decision, the decision itself, and the status of the decision.

Decision records should preferably follow [Michael Nygard decision record template](https://github.com/joelparkerhenderson/architecture-decision-record/blob/main/templates/decision-record-template-by-michael-nygard/index.md) that he described in a [2011 article](https://cognitect.com/blog/2011/11/15/documenting-architecture-decisions.html) on documenting architecture decisions.

For questions about decision records please reach out to the RDFLib maintainers and community using the options given in [further_help_and_contact].

## Decisions list

- [Default branch](decisions/20220826-default_branch.md)

# Type Hints

This document provides some details about the type hints for RDFLib. More information about type hints can be found [here](https://docs.python.org/3/library/typing.html)

## Rationale for Type Hints

Type hints are code annotations that describe the types of variables, function parameters and function return value types in a way that can be understood by humans, static type checkers like [mypy](http://mypy-lang.org/), code editors like VSCode, documentation generators like mkdocstring, and other tools.

Static type checkers can use type hints to detect certain classes of errors by inspection. Code editors and IDEs can use type hints to provide better auto-completion and documentation generators can use type hints to generate better documentation.

These capabilities make it easier to develop a defect-free RDFLib and they also make it easier for users of RDFLib who can now use static type checkers to detect type errors in code that uses RDFLib.

## Gradual Typing Process

Type hints are being added to RDFLib through a process called [gradual typing](https://en.wikipedia.org/wiki/Gradual_typing). This process involves adding type hints to some parts of RDFLib while leaving the rest without type hints. Gradual typing is being applied to many, long-lived, Python code bases.

This process is beneficial in that we can realize some of the benefits of type hints without requiring that the whole codebase have type hints.

## Intended Type Hints

The intent is to have type hints in place for all of RDFLib and to have these type hints be as accurate as possible.

The accuracy of type hints is determined by both the standards that RDFLib aims to conform to, like RDF 1.1, and the deliberate choices that are made when implementing RDFLib. For example, given that the RDF 1.1 specification stipulates that the subject of an RDF triple cannot be a literal, all functions that accept an *RDF term* to be used as the subject of a triple should have type hints which excludes values that are literals.

There may be cases where some functionality of RDFLib may work perfectly well with values of types that are excluded by the type hints, but if these additional types violate the relevant standards we will consider the correct type hints to be those that exclude values of these types.

## Public Type Aliases

In python, type hints are specified in annotations. Type hints are different from type aliases which are normal python variables that are not intended to provide runtime utility and are instead intended for use in static type checking.

For clarity, the following is an example of a function `foo` with type hints:

```python
def foo(a: int) -> int:
    return a + 1
```

In the function `foo`, the input variable `a` is indicated to be of type `int` and the function is indicated to return an `int`.

The following is an example of a type alias `Bar`:

```python
from typing import Tuple

Bar = Tuple[int, str]
```

RDFLib will provide public type aliases under the `rdflib.typing` package, for example, `rdflib.typing.Triple`, `rdflib.typing.Quad`. Type aliases in the rest of RDFLib should be private (i.e. being with an underscore).

## Versioning, Compatibility and Stability

RDFLib attempts to adhere to [semver 2.0](https://semver.org/spec/v2.0.0.html) which is concerned with the public API of software.

Ignoring type hints, the public API of RDFLib exists implicitly as a consequence of the code of RDFLib and the actual behaviour this entails, the relevant standards that RDFLib is trying to implement, and the documentation of RDFLib, with some interplay between all three of these. RDFLib's public API includes public type aliases, as these are normal python variables and not annotations.

Type hints attempt to formally document RDFLib's implicitly-defined public API in a machine-readable fashion as accurately and correctly as possible within the framework outline earlier in this document.

Type hints do not affect the runtime API or behaviour of RDFLib. In this way then, they are somewhat outside of the scope of semver, however, they still have an impact on the users of RDFLib, even if this impact is not at runtime, but during development. This necessitates some clarity as to what users of RDFLib should expect regarding type hints in RDFLib releases.

Changes to type hints can broadly be classified as follow:

**Type Declaration**
  Adding type hints to existing code that had no explicit type hints, for example, changing

```python
def foo(val):
    return val + 1
```

to

```python
def foo(val: int) -> int:
    return val + 1
```

**Type Refinement**
  Refining existing type hints to be narrower, for example, changing a type hint of `typing.Collection` to `typing.Sequence`.

**Type Corrections**
  Correcting existing type hints which contradict the behaviour of the code or relevant specifications, for example, changing `typing.Sequence` from `typing.Set`

Given semver version components `MAJOR.MINOR.PATCH`, RDFLib will attempt to constrain type hint changes as follow:

| Version Component | Type Declaration | Type Refinement | Type Corrections |
|------------------|-----------------|----------------|-----------------|
| MAJOR | YES | YES | YES |
| MINOR | YES | YES | YES |
| PATCH | NO | NO | YES |

!!! caution "Type Corrections"
    A caveat worth nothing here is that code that passed type validation on one version of RDFLib can fail type validation on a later version of RDFLib that only differs in `PATCH` version component. This is as a consequence of potential *Type Corrections*.

# Upgrading from RDFLib version 4.2.2 to 5.0.0

RDFLib version 5.0.0 appeared over 3 years after the previous release, 4.2.2 and contains a large number of both enhancements and bug fixes. Fundamentally though, 5.0.0 is compatible with 4.2.2.

## Major Changes

### Literal Ordering

Literal total ordering [PR #793](https://github.com/RDFLib/rdflib/pull/793) is implemented. That means all literals can now be compared to be greater than or less than any other literal. This is required for implementing some specific SPARQL features, but it is counter-intuitive to those who are expecting a TypeError when certain normally-incompatible types are compared. For example, comparing a `Literal(int(1), datatype=xsd:integer)` to `Literal(datetime.date(10,01,2020), datatype=xsd:date)` using a `>` or `<` operator in rdflib 4.2.2 and earlier, would normally throw a TypeError, however in rdflib 5.0.0 this operation now returns a True or False according to the Literal Total Ordering according the rules outlined in [PR #793](https://github.com/RDFLib/rdflib/pull/793)

### Removed RDF Parsers

The RDFa and Microdata format RDF parsers were removed from rdflib. There are still other python libraries available to implement these parsers.

## All Changes

This list has been assembled from Pull Request and commit information.

### General Bugs Fixed

* Pr 451 redux
  [PR #978](https://github.com/RDFLib/rdflib/pull/978)
* NTriples fails to parse URIs with only a scheme
  [ISSUE #920](https://github.com/RDFLib/rdflib/issues/920)
  [PR #974](https://github.com/RDFLib/rdflib/pull/974)
* cannot clone it on windows - Remove colons from test result files. Fix #901.
  [ISSUE #901](https://github.com/RDFLib/rdflib/issues/901)
  [PR #971](https://github.com/RDFLib/rdflib/pull/971)
* Add requirement for requests to setup.py
  [PR #969](https://github.com/RDFLib/rdflib/pull/969)
* fixed URIRef including native unicode characters
  [PR #961](https://github.com/RDFLib/rdflib/pull/961)
* DCTERMS.format not working
  [ISSUE #932](https://github.com/RDFLib/rdflib/issues/932)
* infixowl.manchesterSyntax do not encode strings
  [PR #906](https://github.com/RDFLib/rdflib/pull/906)
* Fix blank node label to not contain '_:' during parsing
  [PR #886](https://github.com/RDFLib/rdflib/pull/886)
* rename new SPARQLWrapper to SPARQLConnector
  [PR #872](https://github.com/RDFLib/rdflib/pull/872)
* Fix #859. Unquote and Uriquote Literal Datatype.
  [PR #860](https://github.com/RDFLib/rdflib/pull/860)
* Parsing nquads
  [ISSUE #786](https://github.com/RDFLib/rdflib/issues/786)
* ntriples spec allows for upper-cased lang tag, fixes #782
  [PR #784](https://github.com/RDFLib/rdflib/pull/784)
* Error parsing N-Triple file using RDFlib
  [ISSUE #782](https://github.com/RDFLib/rdflib/issues/782)
* Adds escaped single quote to literal parser
  [PR #736](https://github.com/RDFLib/rdflib/pull/736)
* N3 parse error on single quote within single quotes
  [ISSUE #732](https://github.com/RDFLib/rdflib/issues/732)
* Fixed #725
  [PR #730](https://github.com/RDFLib/rdflib/pull/730)
* test for issue #725: canonicalization collapses BNodes
  [PR #726](https://github.com/RDFLib/rdflib/pull/726)
* RGDA1 graph canonicalization sometimes still collapses distinct BNodes
  [ISSUE #725](https://github.com/RDFLib/rdflib/issues/725)
* Accept header should use a q parameter
  [PR #720](https://github.com/RDFLib/rdflib/pull/720)
* Added test for Issue #682 and fixed.
  [PR #718](https://github.com/RDFLib/rdflib/pull/718)
* Incompatibility with Python3: unichr
  [ISSUE #687](https://github.com/RDFLib/rdflib/issues/687)
* namespace.py include colon in ALLOWED_NAME_CHARS
  [PR #663](https://github.com/RDFLib/rdflib/pull/663)
* namespace.py fix compute_qname missing namespaces
  [PR #649](https://github.com/RDFLib/rdflib/pull/649)
* RDFa parsing Error! `__init__()` got an unexpected keyword argument 'encoding'
  [ISSUE #639](https://github.com/RDFLib/rdflib/issues/639)
* Bugfix: `term.Literal.__add__`
  [PR #451](https://github.com/RDFLib/rdflib/pull/451)
* fixup of #443
  [PR #445](https://github.com/RDFLib/rdflib/pull/445)
* Microdata to rdf second edition bak
  [PR #444](https://github.com/RDFLib/rdflib/pull/444)

### Enhanced Features

* Register additional serializer plugins for SPARQL mime types.
  [PR #987](https://github.com/RDFLib/rdflib/pull/987)
* Pr 388 redux
  [PR #979](https://github.com/RDFLib/rdflib/pull/979)
* Allows RDF terms introduced by JSON-LD 1.1
  [PR #970](https://github.com/RDFLib/rdflib/pull/970)
* make SPARQLConnector work with DBpedia
  [PR #941](https://github.com/RDFLib/rdflib/pull/941)
* ClosedNamespace returns right exception for way of access
  [PR #866](https://github.com/RDFLib/rdflib/pull/866)
* Not adding all namespaces for n3 serializer
  [PR #832](https://github.com/RDFLib/rdflib/pull/832)
* Adds basic support of xsd:duration
  [PR #808](https://github.com/RDFLib/rdflib/pull/808)
* Add possibility to set authority and basepath to skolemize graph
  [PR #807](https://github.com/RDFLib/rdflib/pull/807)
* Change notation3 list realization to non-recursive function.
  [PR #805](https://github.com/RDFLib/rdflib/pull/805)
* Suppress warning for not using custom encoding.
  [PR #800](https://github.com/RDFLib/rdflib/pull/800)
* Add support to parsing large xml inputs
  [ISSUE #749](https://github.com/RDFLib/rdflib/issues/749)
  [PR #750](https://github.com/RDFLib/rdflib/pull/750)
* improve hash efficiency by directly using str/unicode hash
  [PR #746](https://github.com/RDFLib/rdflib/pull/746)
* Added the csvw prefix to the RDFa initial context.
  [PR #594](https://github.com/RDFLib/rdflib/pull/594)
* syncing changes from pyMicrodata
  [PR #587](https://github.com/RDFLib/rdflib/pull/587)
* Microdata parser: updated the parser to the latest version of the microdata->rdf note (published in December 2014)
  [PR #443](https://github.com/RDFLib/rdflib/pull/443)
* Literal.toPython() support for xsd:hexBinary
  [PR #388](https://github.com/RDFLib/rdflib/pull/388)

### SPARQL Fixes

* Total order patch patch
  [PR #862](https://github.com/RDFLib/rdflib/pull/862)
* use <<= instead of deprecated <<
  [PR #861](https://github.com/RDFLib/rdflib/pull/861)
* Fix #847
  [PR #856](https://github.com/RDFLib/rdflib/pull/856)
* RDF Literal "1"^^xsd:boolean should _not_ coerce to True
  [ISSUE #847](https://github.com/RDFLib/rdflib/issues/847)
* Makes NOW() return an UTC date
  [PR #844](https://github.com/RDFLib/rdflib/pull/844)
* NOW() SPARQL should return an xsd:dateTime with a timezone
  [ISSUE #843](https://github.com/RDFLib/rdflib/issues/843)
* fix property paths bug: issue #715
  [PR #822](https://github.com/RDFLib/rdflib/pull/822)
  [ISSUE #715](https://github.com/RDFLib/rdflib/issues/715)
* MulPath: correct behaviour of n3()
  [PR #820](https://github.com/RDFLib/rdflib/pull/820)
* Literal total ordering
  [PR #793](https://github.com/RDFLib/rdflib/pull/793)
* Remove SPARQLWrapper dependency
  [PR #744](https://github.com/RDFLib/rdflib/pull/744)
* made UNION faster by not preventing duplicates
  [PR #741](https://github.com/RDFLib/rdflib/pull/741)
* added a hook to add custom functions to SPARQL
  [PR #723](https://github.com/RDFLib/rdflib/pull/723)
* Issue714
  [PR #717](https://github.com/RDFLib/rdflib/pull/717)
* Use <<= instead of deprecated << in SPARQL parser
  [PR #417](https://github.com/RDFLib/rdflib/pull/417)
* Custom FILTER function for SPARQL engine
  [ISSUE #274](https://github.com/RDFLib/rdflib/issues/274)

### Code Quality and Cleanups

* a slightly opinionated autopep8 run
  [PR #870](https://github.com/RDFLib/rdflib/pull/870)
* remove rdfa and microdata parsers from core RDFLib
  [PR #828](https://github.com/RDFLib/rdflib/pull/828)
* ClosedNamespace KeyError -> AttributeError
  [PR #827](https://github.com/RDFLib/rdflib/pull/827)
* typo in rdflib/plugins/sparql/update.py
  [ISSUE #760](https://github.com/RDFLib/rdflib/issues/760)
* Fix logging in interactive mode
  [PR #731](https://github.com/RDFLib/rdflib/pull/731)
* make namespace module flake8-compliant, change exceptions in that mod…
  [PR #711](https://github.com/RDFLib/rdflib/pull/711)
* delete ez_setup.py?
  [ISSUE #669](https://github.com/RDFLib/rdflib/issues/669)
* code duplication issue between rdflib and pymicrodata
  [ISSUE #582](https://github.com/RDFLib/rdflib/issues/582)
* Transition from 2to3 to use of six.py to be merged in 5.0.0-dev
  [PR #519](https://github.com/RDFLib/rdflib/pull/519)
* sparqlstore drop deprecated methods and args
  [PR #516](https://github.com/RDFLib/rdflib/pull/516)
* python3 code seems shockingly inefficient
  [ISSUE #440](https://github.com/RDFLib/rdflib/issues/440)
* removed md5_term_hash, fixes #240
  [PR #439](https://github.com/RDFLib/rdflib/pull/439)
  [ISSUE #240](https://github.com/RDFLib/rdflib/issues/240)

### Testing

* 3.7 for travis
  [PR #864](https://github.com/RDFLib/rdflib/pull/864)
* Added trig unit tests to highlight some current parsing/serializing issues
  [PR #431](https://github.com/RDFLib/rdflib/pull/431)

### Documentation Fixes

* Fix a doc string in the query module
  [PR #976](https://github.com/RDFLib/rdflib/pull/976)
* setup.py: Make the license field use an SPDX identifier
  [PR #789](https://github.com/RDFLib/rdflib/pull/789)
* Update README.md
  [PR #764](https://github.com/RDFLib/rdflib/pull/764)
* Update namespaces_and_bindings.rst
  [PR #757](https://github.com/RDFLib/rdflib/pull/757)
* DOC: README.md: rdflib-jsonld, https uris
  [PR #712](https://github.com/RDFLib/rdflib/pull/712)
* make doctest support py2/py3
  [ISSUE #707](https://github.com/RDFLib/rdflib/issues/707)
* `pip install rdflib` (as per README.md) gets OSError on Mint 18.1
  [ISSUE #704](https://github.com/RDFLib/rdflib/issues/704)
  [PR #717](https://github.com/RDFLib/rdflib/pull/717)
* Use <<= instead of deprecated << in SPARQL parser
  [PR #417](https://github.com/RDFLib/rdflib/pull/417)
* Custom FILTER function for SPARQL engine
  [ISSUE #274](https://github.com/RDFLib/rdflib/issues/274)

# Writing RDFLib Documentation

These docs are generated with [Material for MkDocs](https://squidfunk.github.io/mkdocs-material).

- When writing doc-strings use markdown and google style.
- API Docs are automatically generated with [`mkdocstring`](https://mkdocstrings.github.io).
- See the [supported admonitions here](https://squidfunk.github.io/mkdocs-material/reference/admonitions/#supported-types)

## Building

To build the documentation you can use `mkdocs` from within the poetry environment. To do this, run the following commands:

```bash
# Install poetry venv
poetry install

# Build the docs
poetry run mkdocs build
```

Built HTML docs will be generated in `site/` and API documentation, generated as markdown from doc-strings, will be placed in `docs/apidocs/`.

API Docs are automatically generated with `mkdocstring`

There is also a [tox](https://tox.wiki/en/latest/) environment for building documentation:

```bash
tox -e docs
```

You can check the built documentation with:

```bash
npx -p live-server live-server site/
```

## Development

Run development server with auto-reload on change to code:

```bash
poetry run mkdocs serve
```

## Tables

The tables in `plugin_*.rst` were generated with `plugintable.py`

# RDFLib developers guide

## Introduction

This document describes the process and conventions to follow when
developing RDFLib code.

* Please be as Pythonic as possible ([PEP 8](https://www.python.org/dev/peps/pep-0008/)).
* Code should be formatted using [black](https://github.com/psf/black) and we use Black v23.1.0, with the black config in `pyproject.toml`.
* Code should also pass [flake8](https://flake8.pycqa.org/en/latest/) linting
  and [mypy](http://mypy-lang.org/) type checking.
* You must supply tests for new code.
* RDFLib uses [Poetry](https://python-poetry.org/docs/master/) for dependency management and packaging.

If you add a new cool feature, consider also adding an example in `./examples`.

## Pull Requests Guidelines

Contributions to RDFLib are made through pull requests (PRs).

For changes that add features or affect the public API of RDFLib, it is recommended to first open an issue to discuss the change before starting to work on it. That way you can get feedback on the design of the feature before spending time on it.

In general, maintainers will only merge PRs if the following conditions are met:

* The PR has been sufficiently reviewed.

  Each PR should be reviewed and approved by at least two people other than the
  author of the PR before it is merged and PRs will be processed faster if
  they are easier to review and approve of.

  Reviews are open to everyone, but the weight assigned to any particular
  review is at the discretion of maintainers.

* Changes that have a runtime impact are covered by unit tests.

  There should either be existing tests that cover the changed code and
  behaviour, or the PR should include tests. For more information about what is
  considered adequate testing see the [Tests section](#tests).

* Documentation that covers something that changed has been updated.

* Type checks and unit tests that are part of our continuous integration workflow pass.

In addition to these conditions, PRs that are easier to review and approve will be processed quicker. The primary factors that determine this are the scope and size of a PR. If there are few changes and the scope is limited, then there is less that a reviewer has to understand and less that they can disagree with. It is thus important to try to split up your changes into multiple independent PRs if possible. No PR is too small.

For PRs that introduce breaking changes, it is even more critical that they are limited in size and scope, as they will likely have to be kept up to date with the `main` branch of this project for some time before they are merged.

It is also critical that your PR is understandable both in what it does and why it does it, and how the change will impact the users of this project, for this reason, it is essential that your PR's description explains the nature of the PR, what the PR intends to do, why this is desirable, and how this will affect the users of this project.

Please note that while we would like all PRs to follow the guidelines given here, we will not reject a PR just because it does not.

## Maintenance Guidelines

This section contains guidelines for maintaining RDFLib. RDFLib maintainers should try to follow these. These guidelines also serve as an indication to RDFLib users what they can expect.

### Breaking changes

Breaking changes to RDFLib's public API should be made incrementally, with small pull requests to the main branch that change as few things as possible.

Breaking changes should be discussed first in an issue before work is started, as it is possible that the change is not necessary or that there is a better way to achieve the same goal, in which case the work on the PR would have been wasted. This will however not be strictly enforced, and no PR will be rejected solely on the basis that it was not discussed upfront.

RDFLib follows [semantic versioning](https://semver.org/spec/v2.0.0.html) and [trunk-based development](https://trunkbaseddevelopment.com/), so if any breaking changes were introduced into the main branch since the last release, then the next release will be a major release with an incremented major version.

Releases of RDFLib will not as a rule be conditioned on specific features, so there may be new major releases that contain very few breaking changes, and there could be no minor or patch releases between two major releases.

#### Rationale

RDFLib has been around for more than a decade, and in this time both Python and RDF have evolved, and RDFLib's API also has to evolve to keep up with these changes and to make it easier for users to use. This will inevitably require breaking changes.

There are more or less two ways to introduce breaking changes to RDFLib's public API:

- Revolutionary: Create a new API from scratch and reimplement it, and when
  ready, release a new version of RDFLib with the new API.
- Evolutionary: Incrementally improve the existing API with small changes and
  release any breaking changes that were made at regular intervals.

While the revolutionary approach seems appealing, it is also risky and time-consuming.

The evolutionary approach puts a lot of strain on the users of RDFLib as they have to adapt to breaking changes more often, but the shortcomings of the RDFLib public API also put a lot of strain on the users of RDFLib. On the other hand, a major advantage of the evolutionary approach is that it is simple and achievable from a maintenance and contributor perspective.

### Deprecating functionality

To whatever extent possible, classes, functions, variables, or parameters that will be removed should be marked for deprecation in documentation, and if possible, should be changed to raise deprecation warnings if used.

There is however no hard requirement that something may only be removed after a deprecation notice has been added, or only after a release was made with a deprecation notice.

Consequently, functionality may be removed without it ever being marked as deprecated.

#### Rationale

Current resource limitations and the backlog of issues make it impractical to first release or incorporate deprecation notices before making quality of life changes.

RDFLib uses semantic versioning and provides type hints, and these are the primary mechanisms for signalling breaking changes to our users.

## Tests

Any new functionality being added to RDFLib *must* have unit tests and should have doc tests supplied.

Typically, you should add your functionality and new tests to a branch of RDFlib and run all tests locally and see them pass. There are currently close to 4,000 tests, with a some expected failures and skipped tests. We won't merge pull requests unless the test suite completes successfully.

Tests that you add should show how your new feature or bug fix is doing what you say it is doing: if you remove your enhancement, your new tests should fail!

Finally, please consider adding simple and more complex tests. It's good to see the basic functionality of your feature tests and then also any tricky bits or edge cases.

### Testing framework

RDFLib uses the [pytest](https://docs.pytest.org/en/latest/) testing framework.

### Running tests

To run RDFLib's test suite with [pytest](https://docs.pytest.org/en/latest/):

```bash
poetry install
poetry run pytest
```

Specific tests can be run by file name. For example:

```bash
poetry run pytest test/test_graph/test_graph.py
```

For more extensive tests, including tests for the [berkleydb](https://www.oracle.com/database/technologies/related/berkeleydb.html) backend, install extra requirements before executing the tests.

```bash
poetry install --all-extras
poetry run pytest
```

### Writing tests

New tests should be written for [pytest](https://docs.pytest.org/en/latest/) instead of for python's built-in `unittest` module as pytest provides advanced features such as parameterization and more flexibility in writing expected failure tests than `unittest`.

A primer on how to write tests for pytest can be found [here](https://docs.pytest.org/en/latest/getting-started.html#create-your-first-test).

The existing tests that use `unittest` work well with pytest, but they should ideally be updated to the pytest test-style when they are touched.

Test should go into the `test/` directory, either into an existing test file with a name that is applicable to the test being written, or into a new test file with a name that is descriptive of the tests placed in it. Test files should be named `test_*.py` so that [pytest can discover them](https://docs.pytest.org/en/latest/explanation/goodpractices.html#conventions-for-python-test-discovery).

## Running static checks

Check formatting with [black](https://github.com/psf/black), making sure you use
our black.toml config file:

```bash
poetry run black .
```

Check style and conventions with [ruff](https://docs.astral.sh/ruff/linter/):

```bash
poetry run ruff check
```

Any issues that are found can potentially be fixed automatically using:

```bash
poetry run ruff check --fix
```

Check types with [mypy](http://mypy-lang.org/):

```bash
poetry run mypy --show-error-context --show-error-codes
```

## pre-commit and pre-commit ci

We have [pre-commit](https://pre-commit.com/) configured with [black](https://github.com/psf/black) for formatting code.

Some useful commands for using pre-commit:

```bash
# Install pre-commit.
pip install --user --upgrade pre-commit

# Install pre-commit hooks, this will run pre-commit
# every time you make a git commit.
pre-commit install

# Run pre-commit on changed files.
pre-commit run

# Run pre-commit on all files.
pre-commit run --all-files
```

There is also two tox environments for pre-commit:

```bash
# run pre-commit on changed files.
tox -e precommit

# run pre-commit on all files.
tox -e precommitall
```

There is no hard requirement for pull requests to be processed with pre-commit (or the underlying processors), however doing this makes for a less noisy codebase with cleaner history.

We have enabled [https://pre-commit.ci/](https://pre-commit.ci/) and this can be used to automatically fix pull requests by commenting `pre-commit.ci autofix` on a pull request.

## Using tox

RDFLib has a [tox](https://tox.wiki/en/latest/index.html) config file that makes it easier to run validation on all supported python versions.

```bash
# Install tox.
pip install tox

# List the tox environments that run by default.
tox -e

# Run the default environments.
tox

# List all tox environments, including ones that don't run by default.
tox -a

# Run a specific environment.
tox -e py39 # default environment with py39
tox -e py311-extra # extra tests with py311

# Override the test command.
# the below command will run `pytest test/test_translate_algebra.py`
# instead of the default pytest command.
tox -e py39,py311 -- pytest test/test_translate_algebra.py
```

## `go-task` and `Taskfile.yml`

A `Taskfile.yml` is provided for [go-task](https://taskfile.dev/#/) with various commands that facilitate development.

Instructions for installing go-task can be seen in the [go-task installation guide](https://taskfile.dev/#/installation).

Some useful commands for working with the task in the taskfile is given below:

```bash
# List available tasks.
task -l

# Configure the environment for development
task configure

# Run basic validation
task validate

# Build docs
task docs

# Run live-preview on the docs
task docs:live-server

# Run the py310 tox environment
task tox -- -e py310
```

The [Taskfile usage documentation](https://taskfile.dev/#/usage) provides more information on how to work with taskfiles.

## Development container

To simplify the process of getting a working development environment to develop rdflib in we provide a [Development Container](https://devcontainers.github.io/containers.dev/) (*devcontainer*) that is configured in [Docker Compose](https://docs.docker.com/compose/). This container can be used directly to run various commands, or it can be used with [editors that support Development Containers](https://devcontainers.github.io/containers.dev/supporting).

!!! bug "Rootless docker"
    The devcontainer is intended to run with a
    [rootless docker](https://docs.docker.com/engine/security/rootless/)
    daemon so it can edit files owned by the invoking user without
    an invovled configuration process.

    Using a rootless docker daemon also has general security benefits.

To use the development container directly:

```bash
# Build the devcontainer docker image.
docker-compose build

# Configure the system for development.
docker-compose run --rm run task configure

# Run the validate task inside the devtools container.
docker-compose run --rm run task validate

# Run extensive tests inside the devtools container.
docker-compose run --rm run task EXTENSIVE=true test

# To get a shell into the devcontainer docker image.
docker-compose run --rm run bash
```

The devcontainer also works with [Podman Compose](https://github.com/containers/podman-compose).

Details on how to use the development container with [VSCode](https://code.visualstudio.com/) can found in the [Developing inside a Container](https://code.visualstudio.com/docs/remote/containers) page. With the VSCode [development container CLI](https://code.visualstudio.com/docs/remote/devcontainer-cli) installed the following command can be used to open the repository inside the development container:

```bash
# Inside the repository base directory
cd ./rdflib/

# Build the development container.
devcontainer build .

# Open the code inside the development container.
devcontainer open .
```

## Writing documentation

We use mkdocs for generating HTML docs, see [docs](docs.md).

## Continuous Integration

We used GitHub Actions for CI, see: [https://github.com/RDFLib/rdflib/actions](https://github.com/RDFLib/rdflib/actions)

If you make a pull-request to RDFLib on GitHub, GitHub Actions will automatically test your code and we will only merge code passing all tests.

Please do *not* commit tests you know will fail, even if you're just pointing out a bug. If you commit such tests, flag them as expecting to fail.

## Compatibility

RDFlib 7.0.0 release and later only support Python 3.8.1 and newer.

RDFlib 6.0.0 release and later only support Python 3.7 and newer.

RDFLib 5.0.0 maintained compatibility with Python versions 2.7, 3.4, 3.5, 3.6, 3.7.

## Releasing

Create a release-preparation pull request with the following changes:

* Updated version and date in [`CITATION.cff`](../CITATION.cff).
* Updated copyright year in the [`LICENSE`](../LICENSE) file.
* Updated copyright year in the [`mkdocs.yml`](../mkdocs.yml) file.
* Updated main branch version and current version in the [`README.md`](../README.md) file.
* Updated version in the [`pyproject.toml`](../pyproject.toml) file.
* Updated `__date__` in the [`rdflib/__init__.py`](../rdflib/__init__.py) file.
* Updated [`CHANGELOG.md`](../CHANGELOG.md) entry for the release with admin tools as described in [`admin/README.md`](../admin/README.md).

Once the PR is merged, switch to the main branch, build the release and upload it to PyPI:

```bash
# Clean up any previous builds
rm -vf dist/*

# Build artifacts
poetry build

# Verify package metadata
bsdtar -xvf dist/rdflib-*.whl -O '*/METADATA' | view -
bsdtar -xvf dist/rdflib-*.tar.gz -O '*/PKG-INFO' | view -

# Check that the built wheel and sdist works correctly:
## Ensure pipx is installed but not within RDFLib's environment
pipx run --no-cache --spec "$(readlink -f dist/rdflib*.whl)" rdfpipe --version
pipx run --no-cache --spec "$(readlink -f dist/rdflib*.whl)" rdfpipe https://github.com/RDFLib/rdflib/raw/main/test/data/defined_namespaces/rdfs.ttl
pipx run --no-cache --spec "$(readlink -f dist/rdflib*.tar.gz)" rdfpipe --version
pipx run --no-cache --spec "$(readlink -f dist/rdflib*.tar.gz)" rdfpipe https://github.com/RDFLib/rdflib/raw/main/test/data/defined_namespaces/rdfs.ttl

# Dry run publishing
poetry publish --repository=testpypi --dry-run
poetry publish --dry-run

# Publish to TestPyPI
## ensure you are authed as per https://pypi.org/help/#apitoken and https://github.com/python-poetry/poetry/issues/6320
poetry publish --repository=testpypi

# Publish to PyPI
poetry publish
## poetry publish -u __token__ -p pypi-<REDACTED>
```

Once this is done, create a release tag from [GitHub releases](https://github.com/RDFLib/rdflib/releases/new). For a release of version 6.3.1 the tag should be `6.3.1` (without a "v" prefix), and the release title should be "RDFLib 6.3.1". The release notes for the latest version be added to the release description. The artifacts built with `poetry build` should be uploaded to the release as release artifacts.

The resulting release will be available at https://github.com/RDFLib/rdflib/releases/tag/6.3.1

Once this is done, announce the release at the following locations:

* Twitter: Just make a tweet from your own account linking to the latest release.
* RDFLib mailing list.
* RDFLib Gitter / matrix.org chat room.

Once this is all done, create another post-release pull request with the following changes:

* Set the just released version in `docker/latest/requirements.in` and run `task docker:prepare` to update the `docker/latest/requirements.txt` file.
* Set the version in the `pyproject.toml` file to the next minor release with a `a0` suffix to indicate alpha 0.

[GraphDB by Ontotext](https://graphdb.ontotext.com/) is an enterprise-ready semantic graph database that supports a number of W3C standards.

Ontotext provides GraphDB in two editions:

- **GraphDB Free** — free to use, with limits on concurrency, scalability, and available connectors.
- **GraphDB Enterprise** — commercial edition that removes the Free edition’s limits and adds enterprise features and additional connectors.

The simplest way to get started with GraphDB is to run it in a container. The [ontotext/graphdb](https://hub.docker.com/r/ontotext/graphdb/) image on Docker Hub includes instructions on how to run and operate the database.

!!! note

    From GraphDB 11 onwards, all editions (including GraphDB Free) require a license. You can register for one on the [GraphDB website](https://www.ontotext.com/products/graphdb/).

GraphDB supports programmatic access via a set of REST APIs. It implements the [RDF4J REST API](https://rdf4j.org/documentation/reference/rest-api/), which allows you to manage repositories, perform SPARQL queries and updates, and manage RDF data. If you only need these capabilities, see the [RDFLib RDF4J](rdf4j.md) documentation for details on using the RDFLib RDF4J Store or Client.

GraphDB also provides its own set of administrative REST APIs that extend the RDF4J REST API with GraphDB-specific features such as security management, monitoring, cluster management, and more.
See the GraphDB documentation at https://graphdb.ontotext.com/documentation/.

RDFLib provides a dedicated GraphDB client that implements these GraphDB-specific REST APIs. This documentation covers the usage of the GraphDB client and its features.

## GraphDB Client

The GraphDB Client extends the [RDF4J Client](rdf4j.md#rdf4j-client) and provides additional functionality from the GraphDB-provided REST APIs.

To use RDFLib with the GraphDB Client, first install RDFLib with the optional `graphdb` extra:

```shell
pip install rdflib[graphdb]
```

### Creating a client instance

The [`GraphDBClient`][rdflib.contrib.graphdb.client.GraphDBClient] class is the main entry point for interacting with the GraphDB REST API. To create an instance, pass the base URL of the GraphDB server to the constructor and optionally a username and password tuple for basic authentication.

The preferred way to create a client instance is to use Python's context manager syntax (`with` statement).

```python
from rdflib.contrib.graphdb.client import GraphDBClient

with GraphDBClient("http://localhost:7200/", auth=("admin", "root")) as client:
    # Perform your operations here.
    ...
```

Alternatively, you can create a client instance and manually close it when you are done using it.

```python
from rdflib.contrib.graphdb.client import GraphDBClient

client = GraphDBClient("http://localhost:7200/", auth=("admin", "root"))
try:
    # Perform your operations here.
    ...
finally:
    client.close()
```

With the GraphDB Client, you can also make authenticated requests by obtaining a GDB token. The following code is an example of how to retrieve a token using your username and password, and then use the token to authenticate subsequent requests.

```python
from rdflib.contrib.graphdb.client import GraphDBClient

# Create a client instance (basic authentication is optional for login)
with GraphDBClient("http://localhost:7200/") as client:
    # Authenticate and obtain a GDB token
    user = client.login("admin", "root")

    # The AuthenticatedUser object contains user details
    print(f"Logged in as: {user.username}")
    print(f"Authorities: {user.authorities}")

    # The user.token is a string (e.g., "GDB abc123...") that can be used
    # to authenticate another client instance without using credentials.
    token = user.token

# Use the token with another client instance
with GraphDBClient("http://localhost:7200/", auth=token) as token_client:
    # Perform your operations here.
    repos = token_client.graphdb_repositories.list()
```

### HTTP client configuration

The GraphDB Client extends the RDF4J Client. See [RDF4J Client's HTTP client configuration](rdf4j.md#http-client-configuration) section for details on configuring the underlying HTTP client.

### Repository Management

GraphDB provides two ways to manage repositories:

1.  **RDF4J Repository Manager** ([`GraphDBClient.repositories`][rdflib.contrib.graphdb.client.GraphDBClient.repositories]): Implements the standard RDF4J protocol. Use this for basic operations like retrieving a repository instance for querying. For more information, see [Working with a Repository](rdf4j.md#working-with-a-repository).
2.  **GraphDB Repository Management** ([`GraphDBClient.graphdb_repositories`][rdflib.contrib.graphdb.client.GraphDBClient.graphdb_repositories]): Implements the GraphDB REST API. Use this for administrative tasks such as listing, creating, editing, and validating repositories.

#### Listing repositories

To list all repositories with their configuration and status:

```python
repos = client.graphdb_repositories.list()
for repo in repos:
    print(f"{repo.id} ({repo.type}) - {repo.state}")
```

#### Getting repository configuration

To retrieve a repository's configuration:

```python
# Get configuration as JSON (default)
config = client.graphdb_repositories.get("my-repo")
print(f"Repository ID: {config.id}")

# Get configuration as RDF Graph
graph = client.graphdb_repositories.get("my-repo", content_type="text/turtle")
```

#### Creating a repository

To create a repository, provide the configuration as a Turtle string:

```python
config = """
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#>.
@prefix rep: <http://www.openrdf.org/config/repository#>.
@prefix sr: <http://www.openrdf.org/config/repository/sail#>.
@prefix sail: <http://www.openrdf.org/config/sail#>.
@prefix graphdb: <http://www.ontotext.com/config/graphdb#>.

[] a rep:Repository ;
    rep:repositoryID "my-repo" ;
    rdfs:label "" ;
    rep:repositoryImpl [
        rep:repositoryType "graphdb:SailRepository" ;
        sr:sailImpl [
            sail:sailType "graphdb:Sail" ;

            graphdb:read-only "false" ;

            # Inference and Validation
            graphdb:ruleset "empty" ;
            graphdb:disable-sameAs "true" ;
            graphdb:check-for-inconsistencies "false" ;

            # Indexing
            graphdb:entity-id-size "32" ;
            graphdb:enable-context-index "false" ;
            graphdb:enablePredicateList "true" ;
            graphdb:enable-fts-index "false" ;
            graphdb:fts-indexes ("default" "iri") ;
            graphdb:fts-string-literals-index "default" ;
            graphdb:fts-iris-index "none" ;

            # Queries and Updates
            graphdb:query-timeout "0" ;
            graphdb:throw-QueryEvaluationException-on-timeout "false" ;
            graphdb:query-limit-results "0" ;

            # Settable in the file but otherwise hidden in the UI and in the RDF4J console
            graphdb:base-URL "http://example.org/owlim#" ;
            graphdb:defaultNS "" ;
            graphdb:imports "" ;
            graphdb:repository-type "file-repository" ;
            graphdb:storage-folder "storage" ;
            graphdb:entity-index-size "10000000" ;
            graphdb:in-memory-literal-properties "true" ;
            graphdb:enable-literal-index "true" ;
        ]
    ].

"""
client.graphdb_repositories.create(config)
```

#### Editing a repository

To edit an existing repository's configuration:

```python
from rdflib.contrib.graphdb.models import RepositoryConfigBeanCreate

existing = client.graphdb_repositories.get("my-repo")
updated_config = RepositoryConfigBeanCreate(
    id=existing.id,
    title="Updated Repository",
    type=existing.type,
    sesameType=existing.sesameType,
    location=existing.location,
    params=existing.params,
)
client.graphdb_repositories.edit("my-repo", updated_config)
```

#### Restarting a repository

To restart a repository:

```python
client.graphdb_repositories.restart("my-repo")

# With sync option
client.graphdb_repositories.restart("my-repo", sync=True)
```

#### Getting repository size

To get information about a repository's size:

```python
size_info = client.graphdb_repositories.size("my-repo")
print(f"Total statements: {size_info.total}")
print(f"Explicit statements: {size_info.explicit}")
print(f"Inferred statements: {size_info.inferred}")
```

#### Deleting a repository

To delete a repository:

```python
client.graphdb_repositories.delete("my-repo")
```

### Working with a Repository

To interact with a specific repository (e.g., to run queries or manage data), retrieve a [`Repository`][rdflib.contrib.graphdb.client.Repository] instance. This returns a GraphDB-specific `Repository` object that extends the [standard RDF4J Repository](rdf4j.md#working-with-a-repository) with additional methods for health checks, data file imports, and fine-grained access control.

```python
repo = client.repositories.get("my-repo")
```

#### Repository health check

To check if a repository is healthy and accessible:

```python
is_healthy = repo.health()
print(f"Repository is healthy: {is_healthy}")

# With custom timeout (in seconds)
is_healthy = repo.health(timeout=10)
```

#### Server-side Import

GraphDB allows you to import files that are already present on the server (in the `imports` directory).

##### Listing available files

To list files available for import:

```python
files = repo.get_server_import_files()
for f in files:
    print(f.name)
```

##### Importing files

To import a file:

```python
from rdflib.contrib.graphdb.models import ServerImportBody

import_body = ServerImportBody(fileNames=["data.ttl"])
repo.import_server_import_file(import_body)
```

##### Cancelling an import

To cancel an ongoing import operation:

```python
repo.cancel_server_import_file("data.ttl")
```

#### Fine-Grained Access Control

GraphDB supports Fine-Grained Access Control (FGAC) to restrict access to specific data. You can manage ACLs using the `repo.acl` property.

##### Listing ACL rules

To list existing ACL rules:

```python
rules = repo.acl.list(scope="statement")
for rule in rules:
    print(f"{rule.policy}: {rule.role} can {rule.operation}")

# Filter by scope, operation, or role
statement_rules = repo.acl.list(scope="statement", operation="read")
```

##### Adding ACL rules

To add a new ACL rule:

```python
from rdflib.contrib.graphdb.models import StatementAccessControlEntry

rule = StatementAccessControlEntry(
    scope="statement",
    policy="allow",
    role="user",
    operation="read",
    subject="<http://example.com/public>",
    predicate="*",
    object="*",
    graph="*",
)
repo.acl.add([rule])
```

##### Setting ACL rules

To replace all existing ACL rules:

!!! warning

    This operation replaces all existing ACL rules with the provided ones.

```python
repo.acl.set([rule])
```

##### Deleting ACL rules

To delete specific ACL rules:

!!! warning

    This operation cannot be undone.

```python
rules = repo.acl.list()
repo.acl.delete(rules[:1])
```

### Security Management

GraphDB provides security features including user authentication, role-based access control, and free access mode. You can manage security settings via `client.security` and users via `client.users`.

#### Managing security status

##### Checking and enabling security

To check if security is enabled and toggle it:

```python
# Check if security is enabled
if not client.security.enabled:
    client.security.enabled = True
```

##### Managing free access mode

Free access mode allows unauthenticated users to access repositories with specific permissions:

```python
from rdflib.contrib.graphdb.models import FreeAccessSettings

# Check free access settings
free_access = client.security.get_free_access_details()
print(f"Free access enabled: {free_access.enabled}")

# Configure free access
settings = FreeAccessSettings(enabled=True)
client.security.set_free_access_details(settings)
```

#### User Management

##### Listing users

To list all users:

```python
users = client.users.list()
for user in users:
    print(user.username)
```

##### Getting a user

To get a specific user:

```python
user = client.users.get("admin")
print(f"Username: {user.username}")
```

##### Creating a user

To create a new user:

```python
from rdflib.contrib.graphdb.models import UserCreate

new_user = UserCreate(
    username="newuser",
    password="password123",
    grantedAuthorities=["ROLE_USER"],
)
client.users.create("newuser", new_user)
```

##### Updating a user

To update user properties:

```python
from rdflib.contrib.graphdb.models import UserUpdate

update = UserUpdate(password="new_password")
client.users.update("newuser", update)
```

To fully replace a user's configuration (requires fetching the existing user first to get `dateCreated`):

```python
from rdflib.contrib.graphdb.models import User

existing_user = client.users.get("newuser")
updated_user = User(
    username="newuser",
    password="newpass123",
    dateCreated=existing_user.dateCreated,
    grantedAuthorities=["ROLE_USER"],
)
client.users.overwrite("newuser", updated_user)
```

##### Getting user roles

To retrieve custom roles assigned to a user:

```python
roles = client.users.custom_roles("someuser")
print(roles)
```

##### Deleting a user

To delete a user:

!!! warning

    This operation cannot be undone.

```python
client.users.delete("newuser")
```

### Monitoring and Administration

GraphDB provides monitoring endpoints for system health, performance metrics, and backup/restore operations. Access these via `client.monitoring` and `client.recovery`.

#### Monitoring

##### Infrastructure statistics

To get system-level statistics such as CPU load, memory usage, and thread count:

```python
stats = client.monitoring.infrastructure
print(f"CPU Load: {stats.cpuLoad}")
print(f"Thread Count: {stats.threadCount}")
print(f"Heap Used: {stats.heapMemoryUsage.used}")
```

##### Structures statistics

To get cache hit/miss statistics for internal data structures:

```python
structures = client.monitoring.structures
print(f"Cache hit: {structures.cacheHit}")
print(f"Cache miss: {structures.cacheMiss}")
```

##### Repository statistics

To get statistics for a specific repository:

```python
repo_stats = client.monitoring.get_repo_stats("my-repo")
print(f"Number of slow queries: {repo_stats.queries.slow}")
```

##### Cluster monitoring

To get cluster-level metrics (GraphDB Enterprise only):

```python
cluster_metrics = client.monitoring.cluster()
print(cluster_metrics)
```

##### Backup monitoring

To track ongoing backup operations:

```python
backup_status = client.monitoring.backup()
if backup_status:
    print(f"Backup in progress: {backup_status.operation}")
else:
    print("No backup in progress")
```

#### Recovery Management

GraphDB supports backup and restore operations for both local and cloud storage.

##### Local backup

To create a local backup:

```python
# Stream to file (recommended for large backups)
backup_path = client.recovery.backup(
    repositories=["my-repo"],
    dest="./backup.tar"
)
print(f"Backup created at: {backup_path}")

# Include system data (users, saved queries, etc.)
backup_path = client.recovery.backup(
    repositories=["my-repo"],
    backup_system_data=True,
    dest="./backup/"
)

# Return as bytes (for small backups or testing)
backup_bytes = client.recovery.backup(repositories=["my-repo"])
```

##### Local restore

To restore from a local backup:

```python
# From file path
client.recovery.restore(backup="./backup.tar")

# Restore specific repositories
client.recovery.restore(
    backup="./backup.tar",
    repositories=["my-repo"],
    restore_system_data=True
)

# Remove repositories not in backup
client.recovery.restore(
    backup="./backup.tar",
    remove_stale_repositories=True
)
```

##### Cloud backup

To back up directly to cloud storage:

```python
client.recovery.cloud_backup(bucket_uri="s3://my-bucket/graphdb-backups/")

# With authentication file for cloud provider
client.recovery.cloud_backup(
    bucket_uri="s3://my-bucket/graphdb-backups/",
    authentication_file="./aws-credentials.json"
)
```

##### Cloud restore

To restore from a cloud backup:

```python
client.recovery.cloud_restore(
    bucket_uri="s3://my-bucket/graphdb-backups/backup-2026-01-16-10-30-00.tar"
)
```

#### Cluster Management

!!! note

    Cluster management is only available in GraphDB Enterprise.

GraphDB Enterprise supports high-availability clustering. You can manage cluster configuration, nodes, and secondary mode via [`GraphDBClient.cluster`][rdflib.contrib.graphdb.client.ClusterGroupManagement].

##### Getting cluster configuration

To retrieve the current cluster configuration, use the [`get_config`][rdflib.contrib.graphdb.client.ClusterGroupManagement.get_config] method:

```python
config = client.cluster.get_config()
print(f"Nodes: {config.nodes}")
print(f"Heartbeat interval: {config.heartbeatInterval}")
```

##### Creating a cluster

To create a new cluster, provide a list of node addresses:

```python
client.cluster.create_config(
    nodes=["node1:7300", "node2:7300", "node3:7300"],
    election_min_timeout=3000,
    heartbeat_interval=1000
)
```

##### Updating cluster configuration

To update an existing cluster's configuration:

```python
updated_config = client.cluster.update_config(
    heartbeat_interval=2000,
    transaction_log_maximum_size_gb=10
)
```

##### Managing cluster nodes

You can add, remove, or replace nodes in an existing cluster:

```python
# Add new nodes
client.cluster.add_nodes(["node4:7300"])

# Remove nodes
client.cluster.remove_nodes(["node1:7300"])

# Replace nodes in one operation
client.cluster.replace_nodes(
    add_nodes=["node5:7300"],
    remove_nodes=["node2:7300"]
)
```

##### Checking cluster and node status

To get the status of the current node or the entire cluster group:

```python
# Get status of the current node
node_status = client.cluster.node_status()
print(f"Node state: {node_status.nodeState}")

# Get status of all nodes in the cluster
group_status = client.cluster.group_status()
for node in group_status:
    print(f"{node.address}: {node.nodeState}")
```

##### Managing secondary mode

You can configure a cluster to operate in secondary mode, replicating data from a primary cluster:

!!! warning

    Enabling secondary mode will delete all data on the secondary cluster and replicate the state of the primary cluster.

```python
# Add a tag to identify the primary cluster
client.cluster.add_tag("production-cluster")

# Enable secondary mode on another cluster
client.cluster.enable_secondary_mode(
    primary_node="primary-node1:7300",
    tag="production-cluster"
)

# Disable secondary mode
client.cluster.disable_secondary_mode()

# Remove the tag from the primary cluster
client.cluster.delete_tag("production-cluster")
```

##### Deleting a cluster

To delete the cluster configuration:

```python
# Delete cluster (requires all nodes to be reachable)
result = client.cluster.delete_config()

# Force delete (only deletes on reachable nodes)
result = client.cluster.delete_config(force=True)
```

##### Truncating the transaction log

To free up storage space by clearing the transaction log:

```python
client.cluster.truncate_log()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/write_nquads.R
\name{read_nquads}
\alias{read_nquads}
\title{read an nquads file}
\usage{
read_nquads(file, ...)
}
\arguments{
\item{file}{path to nquads file}

\item{...}{additional arguments to \code{\link[=rdf_parse]{rdf_parse()}}}
}
\value{
an rdf object.  See \code{\link[=rdf_parse]{rdf_parse()}}
}
\description{
read an nquads file
}
\examples{
tmp <- tempfile(fileext = ".nq")
library(datasets)
write_nquads(iris, tmp)
read_nquads(tmp)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rdf_has_bdb.R
\name{rdf_has_bdb}
\alias{rdf_has_bdb}
\title{Check for BDB support}
\usage{
rdf_has_bdb()
}
\value{
TRUE if BDB support is detected, false otherwise
}
\description{
Detect whether Berkeley Database for disk-based storage of RDF graphs
is available.  Disk-based storage requires redland package
to be installed from source with support for the Berkeley DB
(libdb-dev on Ubuntu, berkeley-db on homebrew), otherwise \code{rdf()} will
fall back to in-memory storage with a warning.
}
\examples{
rdf_has_bdb()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rdf_add.R
\name{rdf_add}
\alias{rdf_add}
\title{Add RDF Triples}
\usage{
rdf_add(
  rdf,
  subject,
  predicate,
  object,
  subjectType = as.character(NA),
  objectType = as.character(NA),
  datatype_uri = as.character(NA)
)
}
\arguments{
\item{rdf}{an rdf object}

\item{subject}{character string containing the subject}

\item{predicate}{character string containing the predicate}

\item{object}{character string containing the object}

\item{subjectType}{the Node type of the subject, i.e. "uri", "blank"}

\item{objectType}{the Node type of the object, i.e. "literal", "uri", "blank"}

\item{datatype_uri}{the datatype URI to associate with a object literal value}
}
\value{
Silently returns the updated RDF graph (rdf object).
Since the rdf object simply contains external pointers
to the model object in C code, note that the input object is modified
directly, so you need not assign the output of rdf_add() to anything.
}
\description{
add a triple (subject, predicate, object) to the RDF graph
}
\details{
\code{rdf_add()} will automatically 'duck type' nodes (if looks like a duck...).
That is, strings that look like URIs will be declared as URIs. (See
\href{https://en.wikipedia.org/wiki/Uniform_Resource_Identifier}{URI}).
Predicate should always be a URI (e.g. URL or  a \code{prefix:string}),
cannot be blank or literal.  Subjects that look like strings will be
treated as \href{https://en.wikipedia.org/wiki/Blank_node}{Blank Nodes} (i.e.
will be prefixed with \verb{_:}).  An empty subject, \code{""}, will create a
blank node with random name.  Objects that look like URIs will be
typed as resource nodes, otherwise as literals.  An empty object \code{""}
will be treated as blank node.  Set \code{subjectType} or \code{objectType}
explicitly to override this behavior, e.g. to treat an object URI
as a literal string.  NAs are also treated as blank nodes in subject
or object  See examples for details.
}
\examples{
rdf <- rdf()
rdf_add(rdf, 
    subject="http://www.dajobe.org/",
    predicate="http://purl.org/dc/elements/1.1/language",
    object="en")
    
## non-URI string in subject indicates a blank subject
## (prefixes to "_:b0")
rdf_add(rdf, "b0", "http://schema.org/jobTitle", "Professor") 

## identically a blank subject.  
## Note rdf is unchanged when we add the same triple twice.
rdf_add(rdf, "b0", "http://schema.org/jobTitle", "Professor", 
        subjectType = "blank") 
        
## blank node with empty string creates a default blank node id
rdf_add(rdf, "", "http://schema.org/jobTitle", "Professor")   
                    

## Subject and Object both recognized as URI resources:
rdf_add(rdf, 
        "https://orcid.org/0000-0002-1642-628X",
        "http://schema.org/homepage", 
        "http://carlboettiger.info")  

 ## Force object to be literal, not URI resource        
rdf_add(rdf, 
        "https://orcid.org/0000-0002-1642-628X",
        "http://schema.org/homepage", 
        "http://carlboettiger.info",
        objectType = "literal")  
        

}
\references{
\url{https://en.wikipedia.org/wiki/Uniform_Resource_Identifier}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rdf_query.R
\name{rdf_query}
\alias{rdf_query}
\title{Perform a SPARQL Query}
\usage{
rdf_query(rdf, query, data.frame = TRUE, ...)
}
\arguments{
\item{rdf}{an rdf object (e.g. from \code{\link{rdf_parse}})}

\item{query}{a SPARQL query, as text string}

\item{data.frame}{logical, should the results be returned as a data.frame?}

\item{...}{additional arguments to a redland initialize-Query}
}
\value{
a data.frame of all query results (default.)  Columns will
be named according to variable names in the SPARQL query. Returned
object values will be coerced to match the corresponding R type
to any associated datatype URI, if provided. If a column would
result in mixed classes (e.g. strings and numerics), all types
in the column will be coerced to character strings. If \code{data.frame}
is false, results will be returned as a list with each element
typed by its data URI.
}
\description{
Perform a SPARQL Query
}
\examples{
doc <- system.file("extdata", "dc.rdf", package="redland")

sparql <-
'PREFIX dc: <http://purl.org/dc/elements/1.1/>
 SELECT ?a ?c
 WHERE { ?a dc:creator ?c . }'

rdf <- rdf_parse(doc)
rdf_query(rdf, sparql)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rdf.R
\docType{package}
\name{rdflib-package}
\alias{rdflib}
\alias{rdflib-package}
\title{rdflib: Tools to Manipulate and Query Semantic Data}
\description{
The Resource Description Framework, or RDF is a widely used
data representation model that forms the cornerstone of the
Semantic Web. 'RDF' represents data as a graph rather than
the familiar data table or rectangle of relational databases.
}
\details{
It has three main goals:

\itemize{
\item Easily read, write, and convert between all major RDF serialization formats
\item Support SPARQL queries to extract data from an RDF graph into a data.frame
\item Support JSON-LD format as a first-class citizen in RDF manipulations
}

For more information, see the Wikipedia pages for RDF, SPARQL, and JSON-LD:

\itemize{
\item \url{https://en.wikipedia.org/wiki/Resource_Description_Framework}
\item \url{https://en.wikipedia.org/wiki/SPARQL}
\item \url{https://en.wikipedia.org/wiki/JSON-LD}
}

To learn more about rdflib, start with the vignettes:
\code{browseVignettes(package = "rdflib")}

Configurations via \code{options()}

\code{rdf_print_format}:
\itemize{
\item NULL or "nquads" (default)
\item any valid serializer name: e.g. "rdfxml", "jsonld", "turtle",  "ntriples"
}

\code{rdf_base_uri}:
\itemize{
\item Default base URI to use (when serializing JSON-LD only at this time)
default is "localhost://"
}

\code{rdf_max_print}:
\itemize{
\item maximum number of lines to print from rdf, default 10
}
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/ropensci/rdflib}
  \item Report bugs at \url{https://github.com/ropensci/rdflib/issues}
}

}
\author{
\strong{Maintainer}: Carl Boettiger \email{cboettig@gmail.com} (\href{https://orcid.org/0000-0002-1642-628X}{ORCID}) [copyright holder]

Other contributors:
\itemize{
  \item Bryce Mecum (\href{https://orcid.org/0000-0002-0381-3766}{ORCID}) [reviewer]
  \item Anna Krystalli (\href{https://orcid.org/0000-0002-2378-4915}{ORCID}) [reviewer]
  \item Viktor Senderov \email{vsenderov@gmail.com} (\href{https://orcid.org/0000-0003-3340-5963}{ORCID}) [contributor]
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rdf_serialize.R
\name{rdf_serialize}
\alias{rdf_serialize}
\title{Serialize an RDF Document}
\usage{
rdf_serialize(
  rdf,
  doc = NULL,
  format = c("guess", "rdfxml", "nquads", "ntriples", "turtle", "jsonld"),
  namespace = NULL,
  prefix = names(namespace),
  base = getOption("rdf_base_uri", "localhost://"),
  ...
)
}
\arguments{
\item{rdf}{an existing rdf triplestore to extend with triples from
the parsed file.  Default will create a new rdf object.}

\item{doc}{file path to write out to. If null, will write to character.}

\item{format}{rdf serialization format of the doc,
one of "rdfxml", "nquads", "ntriples", "turtle"
or "jsonld". If not provided, will try to guess based
on file extension and fall back on rdfxml.}

\item{namespace}{a named character containing the prefix to namespace bindings. \code{names(namespace)} are the prefixes, whereas \code{namespace} are the namespaces}

\item{prefix}{(optional) for backward compatibility. See \code{namespace}. It contains the matching prefixes to the namespaces in \code{namespace} and is set automatically if you provide \code{namespace} as a named character vector.}

\item{base}{the base URI to assume for any relative URIs (blank nodes)}

\item{...}{additional arguments to \code{redland::serializeToFile}}
}
\value{
rdf_serialize returns the output file path \code{doc} invisibly.
This makes it easier to use rdf_serialize in pipe chains with
\code{\link{rdf_parse}}.
}
\description{
Serialize an RDF Document
}
\examples{
infile <- system.file("extdata", "dc.rdf", package="redland")
out <- tempfile("file", fileext = ".rdf")

some_rdf <- rdf_parse(infile)
rdf_add(some_rdf,
    subject = "http://www.dajobe.org/dave-beckett",
    predicate = "http://www.w3.org/1999/02/22-rdf-syntax-ns#type",
    object = "http://xmlns.com/foaf/0.1/Person")
rdf_serialize(some_rdf, out)

## With a namespace
rdf_serialize(some_rdf,
          out,
          format = "turtle",
          namespace = c(dc = "http://purl.org/dc/elements/1.1/",
          foaf = "http://xmlns.com/foaf/0.1/")
          )

readLines(out)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_rdf.R
\name{as_rdf}
\alias{as_rdf}
\title{Coerce an object into RDF}
\usage{
as_rdf(
  x,
  rdf = NULL,
  prefix = NULL,
  base = getOption("rdf_base_uri", "localhost://"),
  context = NULL,
  key_column = NULL
)
}
\arguments{
\item{x}{an object to coerce into RDF (list, list-like, or data.frame)}

\item{rdf}{An existing rdf object, (by default a new object will be initialized)}

\item{prefix}{A default vocabulary (URI prefix) to assume for all predicates}

\item{base}{A base URI to assume for blank subject nodes}

\item{context}{a named list mapping any string to a URI}

\item{key_column}{name of a column which should be
treated as the primary key in a table. must be unique}
}
\description{
Coerce an object into RDF
}
\examples{
as_rdf(mtcars)
as_rdf(list(repo = "rdflib", owner = list("id", "ropensci")))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rdf_free.R
\name{rdf_free}
\alias{rdf_free}
\title{Free Memory Associated with RDF object}
\usage{
rdf_free(rdf, rm = TRUE)
}
\arguments{
\item{rdf}{an rdf object}

\item{rm}{logical, default TRUE. Remove pointer from parent.frame()?
Usually a good idea since referring to a pointer after it has been
removed can crash R.}
}
\description{
Free Memory Associated with RDF object
}
\details{
Free all pointers associated with an rdf object.
Frees memory associated with the storage, world, and model
objects.
}
\examples{
rdf <- rdf()
rdf_free(rdf)
rm(rdf)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rdf_methods.R
\name{c.rdf}
\alias{c.rdf}
\title{Concatenate rdf Objects.}
\usage{
\method{c}{rdf}(...)
}
\arguments{
\item{...}{objects to be concatenated}
}
\description{
All subsequent rdf objects will be appended to the first rdf object
Note: this does not free memory from any of the individual rdf objects
Note: It is generally better to avoid the use of this function by passing
an existing rdf object to and rdf_parse or rdf_add objects. Multiple active
rdf objects can cause problems when using disk-based storage backends.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rdf_parse.R
\name{rdf_parse}
\alias{rdf_parse}
\title{Parse RDF Files}
\usage{
rdf_parse(
  doc,
  format = c("guess", "rdfxml", "nquads", "ntriples", "turtle", "jsonld"),
  rdf = NULL,
  base = getOption("rdf_base_uri", "localhost://"),
  ...
)
}
\arguments{
\item{doc}{path, URL, or literal string of the rdf document to parse}

\item{format}{rdf serialization format of the doc,
one of "rdfxml", "nquads", "ntriples", "turtle"
or "jsonld". If not provided, will try to guess based
on file extension and fall back on rdfxml.}

\item{rdf}{an existing rdf triplestore to extend with triples from
the parsed file.  Default will create a new rdf object.}

\item{base}{the base URI to assume for any relative URIs (blank nodes)}

\item{...}{additional parameters (not implemented)}
}
\value{
an rdf object, containing the redland world
and model objects
}
\description{
Parse RDF Files
}
\examples{
doc <- system.file("extdata", "dc.rdf", package="redland")
rdf <- rdf_parse(doc)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/write_nquads.R
\name{write_nquads}
\alias{write_nquads}
\title{write object out as nquads}
\usage{
write_nquads(x, file, ...)
}
\arguments{
\item{x}{an object that can be represented as nquads}

\item{file}{output filename}

\item{...}{additional parameters, see examples}
}
\description{
write object out as nquads
}
\examples{
tmp <- tempfile(fileext = ".nq")
library(datasets)

## convert data.frame to nquads
write_nquads(iris, tmp)
rdf <- read_nquads(tmp)

## or starting a native rdf object
write_nquads(rdf, tempfile(fileext = ".nq"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rdf.R
\name{rdf}
\alias{rdf}
\title{Initialize an \code{rdf} Object}
\usage{
rdf(
  storage = c("memory", "BDB", "sqlite", "postgres", "mysql", "virtuoso"),
  host = NULL,
  port = NULL,
  user = NULL,
  password = NULL,
  database = NULL,
  charset = NULL,
  dir = NULL,
  dsn = "Local Virtuoso",
  name = "rdflib",
  new_db = FALSE,
  fallback = TRUE
)
}
\arguments{
\item{storage}{Storage backend to use; see details}

\item{host}{host address for mysql, postgres, or virtuoso storage}

\item{port}{port for mysql (mysql storage defaults to mysql standard port, 3306)
or postgres (postgres storage defaults to postgres standard port, 4321)}

\item{user}{user name for postgres, mysql, or virtuoso}

\item{password}{password for postgres, mysql, or virtuoso}

\item{database}{name of the database to be created/used}

\item{charset}{charset for virtuoso database, if desired}

\item{dir}{directory of where to write sqlite or berkeley database.}

\item{dsn}{Virtuoso dsn, either "Local Virtuoso" or "Remote Virtuoso"}

\item{name}{name for the storage object created. Default is usually fine.}

\item{new_db}{logical, default FALSE. Create new database or connect to existing?}

\item{fallback}{logical, default TRUE. If requested storage system cannot initialize,
should \code{rdf()} fall back on memory (default) or throw an error (fallback=FALSE)?}
}
\value{
an rdf object
}
\description{
Initialize an \code{rdf} Object
}
\details{
an rdf Object is a list of class 'rdf', consisting of
three pointers to external C objects managed by the redland library.
These are the \code{world} object: basically a top-level pointer for
all RDF models, and a \code{model} object: a collection of RDF statements,
and a \code{storage} object, indicating how these statements are stored.

\code{rdflib} defaults to an in-memory hash-based storage structure.
which should be best for most use cases. For very large triplestores,
disk-based storage will be necessary. Enabling external storage devices
will require additional libraries and custom compiling. See the storage
vignette for details.
}
\examples{
x <- rdf()

}
